/****************************
Medit Unix text screen editor
-------------------------------
Amendment History:

Dec 1994 4.0  Detach from curses to fix scrolling inefficiency.
         4.1  ,,
         4.2  Changed termios structures to termio, to ensure
              compatability with Sequent.
         4.3  Screen optimisation.
         4.4  ,,
         4.5  ,,
         4.6  ,,
         4.7  Blank out status line when scrolling.
         4.8  More optimisation/bug fixes.
         4.9  ,,
         4.10 Start/End File & PageUp/Down redrew screen uneccessarily, fixed
              this; and minor bug fixes. First 4.x stable version.
         4.11 Take out redundant code & tidy up.
         4.12 More efficient sideways scrolling (20 cols at a time).
         4.13 i)  More efficient showline variant (showline_right).
              ii) Make Backspace & Forwards_delete safer by not allowing them
                  to go past end of line.
         4.14 Undo change no 4.13(ii) as it stopped joining of lines!
         4.15 More optimisation.
         4.16 Implement new non-blocking x_getch() and x_buffch().
         4.17 Improve efficiency of screen updating.
         4.18 ,,
         4.19 Minor bug fixes. This version has been distributed.
         4.20 Minor bug fixes & improvements -
              1. Include Backspace=CHR(127) in default keymapping.
              2. Max Line len. now 2000.
         4.21 1. Minor bug fix - TAB ended input in get_string() even when this
                 was not wanted.
              2. Minor bug fix - when pressing help funkey from command line,
                 command line got screwed up.
              3. Screen update optimisation on copy/move block - only redraw
                 screen from cursor position down.
              4. Make find locate occurances within a line, and move cursor.
              5. Fix find-replace bug.
         4.22 1. Fix potential bug in load_key_map().
              2. Fix find() bug, which only appeared on the Sun.

         4.23 Implement QUERY option on replace.

         4.24 1. Catch signals - especially the fatal ones.
              2. Bug fix - replace all got into infinite replace.
              3. Various bug fixes due to vx not being updated.
              4. Bug fix - Command line jumpy cursor.
              5. Bug fix - Find at end of line spuriously succeeded.
              6. Bug fix - jumpy cursor in text after replace.
              7. Various minor replace bug fixes.
              8. Flush keyboard buffer on initialisation.
              9. Do not truncate trailing spaces.

         4.25 1. Bug fix - hangs or line corruption after replace-query.
              2. Bug Fix - replace various 'while (ERR == (c = x_getch()))'
                 which should be 'c = decode_key()'. Caused spruios junk chars
                 to be inserted.
              3. Handle SIGPWR (power failure) more gracefully.

         4.26 1. Take note of TERM, MEDITKEYPATH and argument zero path to
                 construct a default keymap filename -
                 "$MEDITKEYPATH/medit.key.$TERM", otherwise use default map.
              2. Success in using terminfo key escapes in default keymapping.
              3. Fix obscured "default key map" warning.

         4.27 1. Allow medit to run with no filename parameter.

Outstanding known bugs:

1. A file with a single newline char in it, will be converted to an empty file.
   Note that this also occurs in vi.

2. All null characters (ASCII 0) are stripped from the text. Note that this
   also happens with vi.

******************************/
/*---------------
MEDIT VERSION NO
----------------*/
char medit_text[15] = "Medit V4.27   ";
char usage[] = "Usage: medit [-t] [-k <keymapfile>] [<textfile>]\n";

#include <stdio.h>
#include <curses.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <termio.h>
#include <signal.h>
#include <setjmp.h>
#include <stdlib.h>

#define BADSIG -1
/*
#define SIG_IGNX 1
#define SIG_DFLX 0
*/
#define LINEMAXLEN 2001
#define FORWARDS TRUE
#define BACKWARDS FALSE
#define FROM_DELETEBLOCK TRUE
#define NOT_FROM_DELETEBLOCK FALSE
#define STR_SAME 0
#define STR_NOT_SAME 1
#define SAVE_FILE FALSE
#define SAVE_BLOCK TRUE
#define CASE_SENS TRUE
#define NOT_CASE_SENS FALSE
#define ALL_STATUS 0
#define INS_OVR_ONLY 2
#define DELETE_OLD TRUE
#define LEAVE_OLD FALSE
#define UNIX_OK 0
#define LOAD_FILE 0
#define PASTE_FILE 1
#define FUNKEYS_ALLOWED TRUE
#define NO_FUNKEYS FALSE
#define SAVE_COMMAND TRUE
#define NO_SAVE FALSE
#define SQUASH_SPACES TRUE
#define NO_SQUASH FALSE
#define CTRL_ALLOWED TRUE
#define NO_CTRL FALSE
#define TAB_ENDS_INPUT TRUE
#define NO_TAB_ENDS_INPUT FALSE
#define FROM_MOVEBLOCK TRUE
#define NOT_FROM_MOVEBLOCK FALSE
#define TRUNC_TRAILING_SPACES FALSE
#define LEAVE_TRAILING_SPACES TRUE
#define NO_VX_CHECK TRUE
#define VX_CHECK FALSE
#define SAME_LEVEL (TRUE)
#define NEXT_LEVEL (FALSE)
#define FOREVER (TRUE)
#define PAUSE (TRUE)
#define NO_PAUSE (FALSE)
#define HILITE (TRUE)
#define NO_HILITE (FALSE)

/* ASCII codes */

#define CHAR_NULL  0
#define CHAR_TAB   9
#define CHAR_SPACE 32
#define CHAR_DEL   127
#define CHAR_255   255

/* Standard key escape sequences */

#define KB_BACKSPACE  8
#define KB_TAB        9
#define KB_NEWLINE  10
#define KB_DEL      127

#define KB_CTRL_A  1
#define KB_CTRL_B  2
#define KB_CTRL_C  3
#define KB_CTRL_D  4
#define KB_CTRL_E  5
#define KB_CTRL_F  6
#define KB_CTRL_G  7
#define KB_CTRL_H  8
#define KB_CTRL_I  9
#define KB_CTRL_J  10
#define KB_CTRL_K  11
#define KB_CTRL_L  12
#define KB_CTRL_M  13
#define KB_CTRL_N  14
#define KB_CTRL_O  15
#define KB_CTRL_P  16
#define KB_CTRL_Q  17
#define KB_CTRL_R  18
#define KB_CTRL_S  19
#define KB_CTRL_T  20
#define KB_CTRL_U  21
#define KB_CTRL_V  22
#define KB_CTRL_W  23
#define KB_CTRL_X  24
#define KB_CTRL_Y  25
#define KB_CTRL_Z  26

struct line_rec
  {
  struct line_rec *next_line;
  struct line_rec *prev_line;
  int in_block;
  char text[2]; /* N.B. This item is actually VARIABLE LENGTH. Therefore    */
  };            /* an instance of a line_rec CANNOT be copied as a whole -  */
                /* the new instance must be created with the malloc() call  */
                /* specifying how big the new instance must be (including   */
                /* the REAL size of the variable length text item); then    */
                /* each of the individual items must be copied seperately.  */

struct line_rec *line_table_start;
struct line_rec *line_table_top;
struct line_rec *line_table_next;
struct line_rec *line_table_end;
struct line_rec *line_table_bot;
struct line_rec *line_table_cur;
struct line_rec *line_table_block_start;
struct line_rec *line_table_block_end;

struct line_rec *buff_table_start;
struct line_rec *buff_table_end;


/* This is a multi-dimensional table for storing key escape sequences */

struct esc_node
  {
  unsigned int esc_char;
  struct esc_node * next_node;
  struct esc_node * next_level;
  int func_code;
  };

/* This points to the start of the key escape sequence table */
struct esc_node * esc_table = NULL;

/* This is the current position in the key escape sequence table */
struct esc_node * esc_current = NULL;


/* This is the table that matches key descriptions to Medit function
   descriptions. The key descriptions will have been typed in by the user
   doing the key mapping.
*/

#define KEY_DESC_LEN 14
#define FUNC_DESC_LEN 21
struct key_desc_node
  {
  char key_desc[KEY_DESC_LEN+1];
  char func_desc[FUNC_DESC_LEN+1];
  struct key_desc_node * next_node;
  };

/* Pointer to the first entry in the table */
struct key_desc_node * key_desc_table = NULL;
struct key_desc_node * key_desc_last  = NULL;
struct key_desc_node * key_desc_ptr   = NULL;
struct key_desc_node * prev_key_desc_table = NULL;


/*
   BACKSPACE & NEWLINE & TAB MUST HAVE COMMAND CODES = ASCII CODES
   SO THAT THEY DEFAULT TO THE CORRECT VALUE IF UN-MAPPED
*/
#define BACKSPACE     8
#define TAB           9
#define NEW_LINE     10

#define COMMAND_LINE 1001
#define CLEAR_LINE 1002
#define PAGEUP     1003
#define PAGEDOWN   1004
#define START_LINE 1005
#define END_LINE   1006
#define HELP       1007
#define MARK_BLOCK 1008
#define COPY_BLOCK 1009
#define MOVE_BLOCK 1010
#define DEL_LINE   1011
#define FIND_FOR   1012
#define FIND_BACK  1013
#define START_FILE 1014
#define END_FILE   1015
#define CLEAR_MARK 1016
#define DEL_BLOCK  1017
#define PASTE_BLOCK 1018
#define CONTROL_CHAR 1019
#define TOGGLE_INS   1020
#define DEL_CHAR     1021
#define CURSOR_UP    1022
#define CURSOR_DOWN  1023
#define CURSOR_LEFT  1024
#define CURSOR_RIGHT 1025
#define NO_COMMAND 0

char func_desc[][21] =
  {
  "Backwards Del. Char.", /*    8 */
  "Tab                 ", /*    9 */
  "New Line.           ", /*   10 */
  "Help.               ", /* 1007 */
  "Command Line.       ", /* 1001 */
  "Clear Line.         ", /* 1002 */
  "Delete Line.        ", /* 1011 */
  "Find Forwards.      ", /* 1012 */
  "Find Backwards.     ", /* 1013 */
  "Mark Block.         ", /* 1008 */
  "Copy Block.         ", /* 1009 */
  "Move Block.         ", /* 1010 */
  "Cut Block to Buffer.", /* 1017 */
  "Clear Block Markers.", /* 1016 */
  "Paste from Buffer.  ", /* 1018 */
  "Start of File.      ", /* 1014 */
  "End of File.        ", /* 1015 */
  "Insert Control Char.", /* 1019 */
  "Toggle Insert Mode. ", /* 1020 */
  "Forwards Del. Char. ", /* 1021 */
  "Start of Line.      ", /* 1005 */
  "End of Line.        ", /* 1006 */
  "Pageup.             ", /* 1003 */
  "Pagedown.           ", /* 1004 */
  "Cursor Up.          ", /* 1022 */
  "Cursor Down.        ", /* 1023 */
  "Cursor Left.        ", /* 1024 */
  "Cursor Right.       ", /* 1025 */
  "Invalid Command.    "  /* 0    */
  };

int func_code[] =
  {
  BACKSPACE,
  TAB,
  NEW_LINE,
  HELP,
  COMMAND_LINE,
  CLEAR_LINE,
  DEL_LINE,
  FIND_FOR,
  FIND_BACK,
  MARK_BLOCK,
  COPY_BLOCK,
  MOVE_BLOCK,
  DEL_BLOCK,
  CLEAR_MARK,
  PASTE_BLOCK,
  START_FILE,
  END_FILE,
  CONTROL_CHAR,
  TOGGLE_INS,
  DEL_CHAR,
  START_LINE,
  END_LINE,
  PAGEUP,
  PAGEDOWN,
  CURSOR_UP,
  CURSOR_DOWN,
  CURSOR_LEFT,
  CURSOR_RIGHT,
  NO_COMMAND
  };


WINDOW * winptr = (WINDOW *) 0;
WINDOW * statusptr = (WINDOW *) 1;
WINDOW * helpptr = (WINDOW *) 2;

struct termio old_term;
struct termio new_term;

FILE *fptr;

char **g_argv;
int g_argc;

int i;
int c = 0;
int last_c = 0;

/* these were BOTH the current position in the text AND
   the current position of the cursor on the screen,
   but now they are ONLY the position in the text.
*/
int cx = 0;
int cy = 0;
int new_cy;
int use_new_cy = FALSE;

/* current x position in command line
   -1 tells help() & ctrl_char_help() that ccx is not valid
*/

int ccx = -1;

/*
   these are the current position of the cursor on the screen,
*/
int sx = 0;
int cur_sx = 0;
int sy = 0;


/*
  this is the horizontal offset of the virtual screen,
  e.g. if VX = 10 then the left edge of the screen starts 10 chars into
  each line of text.
*/
int vx = 0;

int COLS = 80;
int LINES = 24;
int COMMAND_MAX = 72;
int file_empty = TRUE;
int message_pending = FALSE;
int tabs_in_line = 0;
int text_had_nulls;
int text_had_tabs;
int block_marked = FALSE;
int read_failed;
int save_failed;
int save_abandoned;
int case_sens = FALSE;
int insert_mode = TRUE;
int line_needs_updating = FALSE;
int total_lines = 0;
int no_of_chars;
int total_chars = 0;
int file_needs_saving = FALSE;
int highlight_tabs = FALSE;
int keymap = FALSE;
int scrolling = FALSE;
int scrollingup = FALSE;
int scrollingdown = FALSE;
int blk_cmd_mapped;
int copyblk_mapped = FALSE;
int deleteblk_mapped = FALSE;
int pasteblk_mapped = FALSE;
int moveblk_mapped = FALSE;
int markblk_mapped = FALSE;
int scr_buffer_i = 0;
int panic_signal;
int power_failure = FALSE;
static jmp_buf jmpbuf;

long win_standout = 0;
long win_reverse = 0;
long win_bold = 0;
long win_underline = 0;

long status_standout = 0;
long status_reverse = 0;
long status_bold = 0;
long status_underline = 0;

long help_standout = 0;
long help_reverse = 0;
long help_bold = 0;
long help_underline = 0;

char errortext[LINEMAXLEN];
char mess_text[82];
char findstr[82];
char find_text[31];
char filename[LINEMAXLEN];
char tempfile[L_tmpnam];
char keymapname[LINEMAXLEN];
char pathname[LINEMAXLEN];
char termtype[LINEMAXLEN];
char file_text[LINEMAXLEN];
char prev_command[LINEMAXLEN];
char line_buffer[LINEMAXLEN];
/* char temp_str[LINEMAXLEN]; */

#define SCR_BUFFER_LEN 4096

char scr_buffer[SCR_BUFFER_LEN];

int  tabstop[81];

void initialise();
void finish(int errcode);
char getline(FILE *stream,char s[],int line_max);
void showscreen();
void showscreen_cur();
void pagedown(int y);
void pageup(int y);
void up();
void down();
void left();
void right();
void updateline(int leave_trailing_spaces);
void save(int save_block);
void insert_line();
void insert_line_end();
void delete_line();
void split_line();
void command_line();
int gotoline(int line);
int find(char str[],int direction,int case_sens);
void markblock();
void copyblock(int from_deleteblock);
void pasteblock(int from_moveblock);
void deleteblock(int from_moveblock);
void empty_buffer();
void highlight_block(int on_off);
void showscreency(int y);
void next_mess(char message[]);
void remove_tabs();
struct line_rec * replace_line(struct line_rec *line_table_ptr,char line_buffer[]);
void insert_line_text(struct line_rec *line_table_ptr,char line_buffer[]);
struct line_rec * join_ptr(struct line_rec *line_table_ptr);
void replace(char find_str[],char replace_str[],char suffix_str[]);
char * replace_string(int *len,char *line,char *search,char *replace,int *cx);
void help();
void readfile(char readfname[],int load_or_paste);
char * upper_case(char *source);
void restore_status(int update_what,int y);
int disp_mess();
void insert_char(int c);
void backspace();
void delete_char();
void new_line();
void zap_spaces(char *source);
int in_str(char line[],char *pattern,int *cx,int direction);
char * str_str(char *string,char *pattern);
int rename_file(char *old, char *new);
char * str_tok(
         char * target,
         char * source,
         char * delimiter,
         int squash_spaces,
         int * quoted);
void shift_left_str(char * string, int position, int max_len);
void shift_right_str(char * string, int position, int max_len);
void ins_char_str(char string[], int position, char c, int max_len);
int x_getch();
void x_waddch(WINDOW * winptr,int c,int no_vx_check);
/* void x_winsch(WINDOW * winptr,int c, int no_vx_check); */
void x_waddstr(char * string,int no_vx_check,int init_sx);
void n_waddch(WINDOW * winptr,int c,int no_vx_check,int init_cx,int cx);
void n_waddstr(char * string,int no_vx_check,int init_cx);
void control_char();
void ctrl_char_help();
void winattr(WINDOW * window);
void moveblock();
void dup_block();
void display(char * mess, int pause);
void showline(struct line_rec * line_table_ptr, char * text, int ly, int cx, int hilite_block);
void showline_right(struct line_rec * line_table_ptr, char * text, int ly, int cx, int hilite_block);
void show_str(char * text,int ly,int cx);
void load_buffer();
unsigned int decode_key();
void insert_esc(unsigned int c, int same_level, int f);
int find_esc_char(unsigned int c);
void insert_seq(unsigned int sequence[],int f);
void insert_key_desc(char kdesc[], int func);
void default_key_map();
int map_func(int func);
void map_functions();
void load_key_map();
int get_string(WINDOW * window,
                char command_str[],
                int funkeys_allowed,
                int screeny,
                int screenx,
                int save_command,
                int ctrl_allowed,
                int tab_ends_input);
void calc_sx_ch(int c);
int  calc_sx_str(char * string, int cx);
void down_no_upd();
void up_no_upd();
void show_bot_line();
void showblock(int hilite);
void recalc_screen_cur();
void recalc_screen_cy(int y);
void flush_scr_buffer();
void x_buffch(unsigned int c);
void catchsigs();
void panic_stations(int panic_signal);
void panic(int sig);
static void setsig(int sig, void (*fcn)());

extern void setup();
extern void clreol();
extern void clrscr();
extern void clrbot();
extern void cmove(int y, int x);
extern void ins_mode();
extern void ovr_mode();
extern void enter_bold();
extern void enter_inv();
extern void enter_under();
extern void exit_under();
extern void enter_blink();
extern void normal();
extern void delchar();
extern void enter_ca();
extern void exit_ca();
extern int get_key_str(unsigned char * p,char k[]);
extern int key_backspace_str(unsigned char * p);
extern int key_help_str(unsigned char * p);
extern int key_home_str(unsigned char * p);
extern int key_end_str(unsigned char * p);
extern int key_nextpage_str(unsigned char * p);
extern int key_prevpage_str(unsigned char * p);
extern int key_left_str(unsigned char * p);
extern int key_right_str(unsigned char * p);
extern int key_up_str(unsigned char * p);
extern int key_down_str(unsigned char * p);
extern int getCOLS();
extern int getLINES();
extern void deline();
extern void insline();

/*************************/
main(argc,argv)

int argc;
char **argv;
{
unsigned int c;

g_argv = argv;
g_argc = argc;

initialise();

/*
-------------------------
The main edit mode loop
--------------------------
*/

load_buffer();

if ((panic_signal = setjmp(jmpbuf)) != 0)
  /* A longjmp has just been executed ( from panic() ), so ... */
  panic_stations(panic_signal);
else
  /* Setjmp is returning from initialising the jmp point,
     so no need to panic, just catch the signals.
  */
  catchsigs();

disp_mess();
strcpy(mess_text,"");
c = decode_key();

while (FOREVER)
  {
  switch (c)
    {
    case COMMAND_LINE:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        command_line();
        load_buffer();
        break;

    case CONTROL_CHAR:
        control_char();
        break;

    case TOGGLE_INS:
        if (insert_mode)
          insert_mode = FALSE;
        else
          insert_mode = TRUE;
        restore_status(INS_OVR_ONLY,LINES-1);
        break;

    case FIND_FOR:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        if (find(findstr,FORWARDS,case_sens))
          {
          cur_sx = calc_sx_str(line_table_cur->text,cx);

          if (  (cur_sx > (vx + COLS - 2 - (int)strlen(findstr)))
             || (cur_sx < vx)
             || (  (cur_sx < (COLS - 2 - (int)strlen(findstr)))
                && (vx > 0)
                )
             )
            {
            vx = cur_sx - COLS + 2 + (int)strlen(findstr);
            if (vx < 0) vx = 0;
            use_new_cy = FALSE;
            }

          showscreency(LINES/4);

          next_mess(find_text);
          }
        else
          next_mess("String not found.");
        load_buffer();
        break;

    case FIND_BACK:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        if (find(findstr,BACKWARDS,case_sens))
          {
          cur_sx = calc_sx_str(line_table_cur->text,cx);

          if (  (cur_sx > (vx + COLS - 2 - (int)strlen(findstr)))
             || (cur_sx < vx)
             || (  (cur_sx < (COLS - 2 - (int)strlen(findstr)))
                && (vx > 0)
                )
             )
            {
            vx = cur_sx - COLS + 2 + (int)strlen(findstr);
            if (vx < 0) vx = 0;
            use_new_cy = FALSE;
            }

          showscreency(LINES/4*3);

          next_mess(find_text);
          }
        else
          next_mess("String not found.");
        load_buffer();
        break;

    case HELP:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        help();
        load_buffer();
        break;

    case DEL_LINE:
        delete_line();
        load_buffer();
        break;

    case CLEAR_LINE:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        line_buffer[cx] = '\0';
        updateline(LEAVE_TRAILING_SPACES);
        showline(line_table_cur,line_buffer,cy,cx,HILITE);
        line_needs_updating = TRUE;
        cmove(cy,cur_sx-vx);
        break;

    case BACKSPACE:
        backspace();
        break;

    case DEL_CHAR:
        delete_char();
        break;

    case NEW_LINE:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        new_line();
        load_buffer();
        break;

    case START_LINE:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        cx = 0;
        cur_sx = calc_sx_str(line_table_cur->text,cx);

        if (cur_sx < vx)
          {
          vx = cur_sx;
          showscreen();
          }
        else
          cmove(cy,cur_sx-vx);

        load_buffer();
        break;

    case END_LINE:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        cx = strlen(line_table_cur->text);
        cur_sx = calc_sx_str(line_table_cur->text,cx);

        if (cur_sx > (vx + COLS - 1))
          {
          vx = cur_sx - COLS + 1;
/*        vx = cur_sx - COLS + 21; */
          showscreen();
          }
        else
          cmove(cy,cur_sx-vx);

        load_buffer();
        break;

    case START_FILE:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        if (line_table_cur->prev_line != NULL)
          {
          cx = cy = vx = 0;
          line_table_top = line_table_start;
          showscreen();
          }
        load_buffer();
        break;

    case END_FILE:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        if (line_table_cur->next_line != NULL)
          {
          line_table_cur = line_table_end;
          line_table_top = line_table_end;
          cx = vx = 0;
          cy = LINES-2;
          showscreency(LINES-2);
          }
        load_buffer();
        break;

    case PAGEDOWN:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        if (line_table_bot->next_line != NULL)
          {
          pagedown(LINES-2);
          showscreen();
          }
        load_buffer();
        break;

    case PAGEUP:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        if (line_table_top->prev_line != NULL)
          {
          pageup(LINES-2);
          showscreen();
          }
        load_buffer();
        break;

    case CURSOR_UP:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        up();
        load_buffer();
        break;

    case CURSOR_DOWN:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        down();
        load_buffer();
        break;

    case CURSOR_LEFT:
        left();
        break;

    case CURSOR_RIGHT:
        right();
        break;

    case MARK_BLOCK:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        markblock();
        load_buffer();
        break;

    case PASTE_BLOCK:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        pasteblock(NOT_FROM_MOVEBLOCK);
        load_buffer();
        break;

    case MOVE_BLOCK:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        moveblock();
        load_buffer();
        break;

    case COPY_BLOCK:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        dup_block();
        load_buffer();
        break;

    case DEL_BLOCK:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        deleteblock(NOT_FROM_MOVEBLOCK);
        load_buffer();
        break;

    case CLEAR_MARK:
        if (line_needs_updating)
          updateline(LEAVE_TRAILING_SPACES);
        showblock(NO_HILITE);
        highlight_block(FALSE);
        block_marked = FALSE;
        line_table_block_start = line_table_block_end = NULL;
        next_mess("Block markers cleared.");
        load_buffer();
        break;

    case NO_COMMAND:
        next_mess("Keystroke ignored.");
        break;

    default:
        if (  (  (c >= CHAR_SPACE)
              && (c < CHAR_DEL)
              )
           || (c == TAB)
           )
        insert_char(c);
        break;
    }

  if (scrolling)
    if (  (  (c != CURSOR_UP)
          && (scrollingup)
          )
       || (  (c != CURSOR_DOWN)
          && (scrollingdown)
          )
       )
      {
      scrolling = FALSE;
      scrollingup = FALSE;
      scrollingdown = FALSE;
      new_term.c_cc[4] = 1; /* MIN  */
      new_term.c_cc[5] = 0; /* TIME */
      if (ioctl(0,TCSETAF,(struct termio *) &new_term))
        next_mess("Unable to setup terminal: ioctl() failed.");
      restore_status(ALL_STATUS,LINES-1);
      }

  if (message_pending == TRUE)
    message_pending = disp_mess();

  c = decode_key();

  }
}

/*************************/
static void setsig(int sig, void (*fcn)())
{
int s;

s = (int) signal(sig,SIG_IGN);

if (s == BADSIG)
  next_mess("Signal trapping error!");
else if (s == (int)SIG_IGN)
  return;
else if (s != (int)SIG_DFL)
  next_mess("Signal already caught!");

/*
switch(s)
  {
  case BADSIG:
    next_mess("Signal trapping error!");
  case SIG_IGNX:
    return;
  case SIG_DFLX:
    break;
  default:
    next_mess("Signal already caught!");
  }
*/

if ((int)signal(sig,fcn) == BADSIG)
  next_mess("Signal trapping error!");
}

/*************************/
void catchsigs()
{
setsig(SIGHUP,panic);
setsig(SIGINT,panic);
setsig(SIGQUIT,panic);
setsig(SIGILL,panic);
setsig(SIGTRAP,SIG_IGN);
setsig(SIGIOT,panic);
setsig(SIGEMT,panic);
setsig(SIGFPE,panic);
setsig(SIGBUS,panic);
setsig(SIGSEGV,panic);
setsig(SIGSYS,panic);
setsig(SIGPIPE,panic);
setsig(SIGALRM,SIG_IGN);
setsig(SIGTERM,panic);
setsig(SIGCLD,SIG_IGN);
setsig(SIGPWR,panic);
}

/*************************/
void panic(int sig)
{
if (sig != SIGPWR)
  {
  clrscr();
  flush_scr_buffer();
  longjmp(jmpbuf,sig);
  fprintf(stderr,"\n\nSignal interupt received, but longjmp did not work.\n\n");
  panic_stations(sig);
  }
else
  {
  power_failure = TRUE;
  message_pending = TRUE;
  }
}

/*************************/
void panic_stations(int panic_signal)
{
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;
struct stat fstat_buffer;
char temp_str[LINEMAXLEN];
int read_only = FALSE;
int ok_to_write = FALSE;
mode_t mode;

if ((int)signal(panic_signal,SIG_IGN) == BADSIG)
  fprintf(stderr,"\n\nSignal trapping error in signal trapping function.\n\n");

save_abandoned = FALSE;
save_failed = FALSE;
message_pending = FALSE;


fptr = fopen(tempfile,"w");
if (fptr == NULL)
  sprintf(errortext,"%s on file '%s'",strerror(errno),tempfile);
else
  {
  line_table_ptr = line_table_start;
  while (  (line_table_ptr != NULL)
        && (! save_failed)
        )
    {
    if (fputs(line_table_ptr->text,fptr) == EOF)
      {
      sprintf(errortext,"%s on file '%s'",strerror(errno),tempfile);
      next_mess(errortext);
      save_failed = TRUE;
      break;
      }
    if (!file_empty)
      {
      if (fputc('\n',fptr) == EOF)
        {
        sprintf(errortext,"%s on file '%s'",strerror(errno),tempfile);
        next_mess(errortext);
        save_failed = TRUE;
        break;
        }
      line_table_next = line_table_ptr->next_line;
      line_table_ptr = line_table_next;
      }
    }
  }
fclose(fptr);

switch(panic_signal)
  {
  case SIGHUP:
    sprintf(temp_str,"\n\nMedit interrupted with SIGHUP signal.\n");
    break;
  case SIGINT:
    sprintf(temp_str,"\n\nMedit interrupted with SIGINT signal.\n");
    break;
  case SIGQUIT:
    sprintf(temp_str,"\n\nMedit interrupted with SIGQUIT signal.\n");
    break;
  case SIGILL:
    sprintf(temp_str,"\n\nMedit interrupted with SIGILL signal.\n");
    break;
  case SIGTRAP:
    sprintf(temp_str,"\n\nMedit interrupted with SIGTRAP signal.\n");
    break;
  case SIGIOT:
    sprintf(temp_str,"\n\nMedit interrupted with SIGIOT signal.\n");
    break;
  case SIGEMT:
    sprintf(temp_str,"\n\nMedit interrupted with SIGEMT signal.\n");
    break;
  case SIGFPE:
    sprintf(temp_str,"\n\nMedit interrupted with SIGFPE signal.\n");
    break;
  case SIGKILL:
    sprintf(temp_str,"\n\nMedit interrupted with SIGKILL signal.\n");
    break;
  case SIGBUS:
    sprintf(temp_str,"\n\nMedit has crashed with a bus error (SIGBUS signal).\n");
    break;
  case SIGSEGV:
    sprintf(temp_str,"\n\nMedit has crashed with a segmentation error (SIGSEGV signal).\n");
    break;
  case SIGSYS:
    sprintf(temp_str,"\n\nMedit interrupted with SIGSYS signal.\n");
    break;
  case SIGPIPE:
    sprintf(temp_str,"\n\nMedit interrupted with SIGPIPE signal.\n");
    break;
  case SIGALRM:
    sprintf(temp_str,"\n\nMedit interrupted with SIGALRM signal.\n");
    break;
  case SIGTERM:
    sprintf(temp_str,"\n\nMedit interrupted with SIGTERM signal.\n");
    break;
  case SIGUSR1:
    sprintf(temp_str,"\n\nMedit interrupted with SIGUSR1 signal.\n");
    break;
  case SIGUSR2:
    sprintf(temp_str,"\n\nMedit interrupted with SIGUSR2 signal.\n");
    break;
  case SIGCLD:
    sprintf(temp_str,"\n\nMedit interrupted with SIGCLD signal.\n");
    break;
  case SIGPWR:
    sprintf(temp_str,"\n\nMedit interrupted with SIGPWR signal.\n");
    break;
  }

cmove(0,0);
clrscr();
fprintf(stderr,temp_str);
if (!save_failed)
  {
  fprintf(stderr,"You may be able to recover your edited text from the file:\n");
  fprintf(stderr,"'%s'\n\n",tempfile);
  fprintf(stderr,"Please delete this temporary file after you have finished with it.\n");
  }
exit_ca();
if (ioctl(0,TCSETAF,(struct termio *) &old_term))
  fprintf(stderr,"Unable to RESET initial terminal characteristics.\n");
exit(1);
}

/******************************************************************************
THE KEYBOARD ESCAPE SEQUENCE TABLE & ASSOCIATED CODE.
-----------------------------------------------------
This is one of the most tricky bits of the program.

The escape sequence table is a mutli-dimensionsal table implemented as a
linked-list.

Each entry has two pointers -
    1. one to the "next level" (next char in escape sequence).
    2. one to the "next node" (next alternative to the current escape char).

For example, take the following example escape sequences, entered into the
table in sequence.
    1.  Esc H p  - mapped to function 1.
    2.  Esc H x  - mapped to function 2.
    3.  127      - mapped to function 3.

These would look like this in the escape char table:

--------------------      --------------------
| Next Node:  -----+----->| Next Node:  NULL |
| Next Level: -----+----  | Next Level: NULL |
| Char:       Esc  |   |  | Char:       127  |
| Function:        |   |  | Function:   3    |
--------------------   |  --------------------
                       |
-------------------- <--
| Next Node:  NULL |
| Next Level: -----+----
| Char:       H    |   |
| Function:        |   |
--------------------   |
                       |
-------------------- <--  --------------------
| Next Node:  -----+----->| Next Node:  NULL |
| Next Level: NULL |      | Next Level: NULL |
| Char:       p    |      | Char:       x    |
| Function:   1    |      | Function:   2    |
--------------------      --------------------

An entry contains a terminating character for a sequence if the "Next Level"
pointer is NULL. The Function field will then contain the Medit function that
the sequence is mapped to.

An escape sequence cannot consist of a subset of another sequence.

Variables:
----------
  esc_current    is the current position in the escape table (while
                 waiting for the next escape char in the sequence).

  esc_table      is the start of the escape table.

Subroutines:
------------
  find_esc_char  sees if the specified char is valid given the current point
                 in the escape table, and if it is, then it advances the
                 esc_current pointer.

  insert_esc     inserts a single char into the table at the current point.

  insert_seq     calls insert_esc to insert a complete escape sequence into the
                 escape table.

Inserting escape characters into the table.
-------------------------------------------
Inserting a new escape sequence into the table is not trivial, because:

a) some of the escape chars will probably already be in the table, but these
   chars must not be entered again, only the completely "new" chars must be
   entered into the table. To use the example above, when inserting the 2nd
   escape sequence, only the "p" gets entered into the table.

b) it has to be figured out whether to attach the new escape chars to the "Next
   Node" pointer, or to the "Next Level" pointer. Maybe I could make this
   easier on myself, by having the find_esc_char routine be more consistent in
   how it leaves the esc_current pointer.

Notes.
------
No routines exist to delete escape sequences from the table, this is mainly
because it was not needed for the current version. However, note that I think
it would be a difficult task, and that any changes made to this effect should
be done with care & trepidation.

*****************************************************************************/
void insert_seq(unsigned int sequence[],int f)
{
int i,j;
char temp_str[82];
char temp_hex[82];
struct esc_node * esc_prev;

if (sequence[0] == 0)
  {
  next_mess("Empty key escape sequence in keymap file.");
  return;
  }

esc_current = NULL;
i = -1;
do{
  esc_prev = esc_current;
  i++;
  } while (find_esc_char(sequence[i]));

if (sequence[i] == 0)
  {
  strcpy(temp_str,"Key esc ");
  j = 0;
  while (j < i)
    {
    sprintf(temp_hex,"%X ",sequence[j]);
    strcat(temp_str,temp_hex);
    j++;
    }
  strcat(temp_str,"more than once in keymap file.");
  next_mess(temp_str);
  }
else
  {
  if (esc_current != esc_prev)
    {
    if (sequence[i+1] == 0)
      insert_esc(sequence[i],SAME_LEVEL,f);
    else
      insert_esc(sequence[i],SAME_LEVEL,0);
    i++;
    }

  while (sequence[i] != 0)
    {
    if (sequence[i+1] == 0)
      insert_esc(sequence[i],NEXT_LEVEL,f);
    else
      insert_esc(sequence[i],NEXT_LEVEL,0);
    i++;
    }
  }
}

/*****************************************************************************/
int find_esc_char(unsigned int c)
{
struct esc_node * esc_ptr;
struct esc_node * esc_prev;

esc_prev = esc_current;

if (esc_current == NULL)
  esc_prev = esc_ptr = esc_table;
else
  esc_ptr = esc_current->next_level;

while (esc_ptr != NULL)
  {
  if (esc_ptr->esc_char == c)
    {
    esc_current = esc_ptr;
    return TRUE;
    }
  esc_prev = esc_ptr;
  esc_ptr = esc_ptr->next_node;
  }

esc_current = esc_prev;
return FALSE;
}

/*****************************************************************************/
void insert_esc(unsigned int c, int same_level, int f)
{
struct esc_node * esc_ptr;
struct esc_node * esc_next;

esc_ptr = (struct esc_node *) malloc(sizeof (struct esc_node));
if (esc_ptr == NULL)
  next_mess("Not enough memory to add escape sequence to table.");
else
  {
  if (esc_table == NULL)
    esc_table = esc_ptr;

  esc_ptr->esc_char = c;
  esc_ptr->func_code = f;

  esc_ptr->next_node = NULL;
  esc_ptr->next_level = NULL;

  if (esc_current == NULL)
    esc_current = esc_table;
  else
    {
    if (same_level)
      esc_current->next_node = esc_ptr;
    else
      esc_current->next_level = esc_ptr;
    esc_current = esc_ptr;
    }
  }
}

/*****************************************************************************/
void insert_key_desc(char kdesc[], int func)
{
int i;
struct key_desc_node * key_desc_ptr = NULL;

i = 0;
while ((func_code[i] != func) && (func_code[i] != NO_COMMAND)) i+=1;

if (func_code[i] != NO_COMMAND)
  {
  key_desc_ptr = (struct key_desc_node *) malloc(sizeof (struct key_desc_node));
  if (key_desc_ptr == NULL)
    next_mess("Not enough memory to add keymap entry to table.");
  else
    {
    strncpy(key_desc_ptr->key_desc,kdesc,KEY_DESC_LEN);
    strncpy(key_desc_ptr->func_desc,func_desc[i],FUNC_DESC_LEN);
    key_desc_ptr->next_node = NULL;

    if (key_desc_table == NULL)
      key_desc_table = key_desc_last = key_desc_ptr;
    else
      {
      key_desc_last->next_node = key_desc_ptr;
      key_desc_last = key_desc_ptr;
      }
    }
  }
else
  next_mess("Invalid function code in key map file.");
}

/**************************/
void flush_scr_buffer()
{
if (scr_buffer_i > 0)
  write(1,scr_buffer,scr_buffer_i);
scr_buffer_i =0;
}

/**************************/
void x_buffch(unsigned int c)
{
scr_buffer[scr_buffer_i] = (unsigned char) c;
scr_buffer_i += 1;
if (scr_buffer_i >= SCR_BUFFER_LEN)
  flush_scr_buffer();
}

/**************************/
int x_getch()
{
int c;
unsigned char x;

flush_scr_buffer();

if (read(0, &x, 1) > 0)
  c = x;
else
  c = ERR;
return c;
}

/**************************/
unsigned int decode_key()
{
int c;
int finished = FALSE;

c = x_getch();

while (  (c != ERR)
      && (!finished)
      )
  {
  if (!find_esc_char(c))
    {
    esc_current = NULL;
    finished = TRUE;
    }
  else
    {
    if (esc_current->next_level == NULL)
      {
      c = esc_current->func_code;
      esc_current = NULL;
      finished = TRUE;
      }
    }

  if (!finished)
    c = x_getch();
  }

return c;
}

/*************************/
int map_func(int func)
{
int i,di,ki,ic,f,c,field,error;
char temp_str[LINEMAXLEN];
char temp_str2[LINEMAXLEN];
struct key_desc_node * key_desc_ptr = NULL;
unsigned int kcode[81];
char kdesc[14];

i = 0;
f = 0;

while(i < LINES-1)
  {
  cmove(i,0);
  clreol();
  i += 1;
  }
i = 0;

while ((func_code[f] != func) && (func_code[f] != NO_COMMAND)) f+=1;
strcpy(kdesc,"");

enter_bold();
enter_inv();
i += 1;
cmove(i,0);
clreol();
sprintf(temp_str,"           Map Medit '%s' function to a key            ",func_desc[f]);
x_waddstr(temp_str,NO_VX_CHECK,0);

normal();
enter_bold();
i += 2;
cmove(i,0);
clreol();
sprintf(temp_str,"Press TAB to move between items, press RETURN when finished:");
x_waddstr(temp_str,NO_VX_CHECK,0);

i += 2;
di = i;
cmove(i,0);
clreol();
sprintf(temp_str,"Description of key ");
x_waddstr(temp_str,NO_VX_CHECK,0);

cmove(i,20);

i += 2;
ki = i;
cmove(i,0);
clreol();
sprintf(temp_str,"Press the key on the terminal ");
x_waddstr(temp_str,NO_VX_CHECK,0);

cmove(i,31);

normal();

strcpy(kdesc,"");

error = FALSE;
COMMAND_MAX = 13;
cmove(di,20);
field = 1;
ic = 0;
c = '\0';
while (c != NEW_LINE)
  {
  if (field == 1)
    c = get_string(helpptr,kdesc,NO_FUNKEYS,di,20,NO_SAVE,NO_CTRL,TAB_ENDS_INPUT);
  else
    {
    while((c = x_getch()) == ERR);
    ic = 0;
    while (  (c != NEW_LINE)
          && (c != TAB)
          && (ic < 80)
          )
      {
      cmove(ki,31);
      enter_inv();
      enter_under();
      enter_bold();
      x_waddstr("     ",NO_VX_CHECK,0);
      normal();
      cmove(ki,31);

      kcode[ic] = (unsigned int) c;

      ic++;
      while((c = x_getch()) == ERR);
      }
    kcode[ic] = 0;
    }

  if (c == NEW_LINE)
    if (  (strcmp(kdesc,"") != STR_SAME)
       && (ic == 0)
       )
      {
      c = '\0';
      next_mess("Key has a description but no escape sequence.");
      error = TRUE;
      message_pending = disp_mess();
      if (field == 1)
        cmove(di,20);
      else if (field == 2)
        cmove(ki,31);
      }
    else if (error)
      {
      next_mess("");
      message_pending = disp_mess();
      error = FALSE;
      }

  if (c == TAB)
    if (field == 1)
      {
      field = 2;
      ic = 0;
      cmove(ki,31);
      clreol();
      cmove(ki,31);
      }
    else
      {
      field = 1;
      cmove(di,20);
      }
  }
normal();
help_reverse = 0;

COMMAND_MAX = 72;

if (ic != 0)
  {
  sprintf(temp_str,"%X \"%s\" ",func,kdesc);
  i = 0;
  while(kcode[i] != 0)
    {
    sprintf(temp_str2,"%X ",kcode[i]);
    strcat(temp_str,temp_str2);
    i+=1;
    }
  insert_line_text(line_table_cur,temp_str);
  }

if (ic == 0)
  return TRUE;
else
  return FALSE;
}

/*****************************************************************************/
void map_functions()
{
int i,finished;
char temp_str[82];
struct esc_node * prev_esc_table = NULL;

if (!file_empty)
  {
  cmove(LINES-1,9);
  clreol();
  x_waddstr("File not empty. Do you want to continue (Y or N)?",NO_VX_CHECK,0);
  cmove(LINES-1,59);
  strcpy(temp_str,"");
  get_string(statusptr,temp_str,NO_FUNKEYS,LINES-1,59,NO_SAVE,NO_CTRL,NO_TAB_ENDS_INPUT);
  upper_case(temp_str);
  zap_spaces(temp_str);
  if (  (strcmp(temp_str,"YES") != STR_SAME)
     && (strcmp(temp_str,"Y") != STR_SAME)
     )
    return;
  cmove(LINES-1,9);
  clreol();
  }

if (line_needs_updating)
  updateline(LEAVE_TRAILING_SPACES);

prev_esc_table = esc_table;
prev_key_desc_table = key_desc_table;

i = 0;
while (func_code[i] != NO_COMMAND)
  {
  finished = FALSE;
  while (!finished)
    finished = map_func(func_code[i]);
  /* map_func(func_code[i]); */
  i+=1;
  }
finished = FALSE;
while (!finished)
  finished = map_func(func_code[i]);

showscreen();
load_buffer();
}

/*****************************************************************************/
void load_key_map()
{
unsigned int esc_char[81];
char first_token[LINEMAXLEN];
char second_token[LINEMAXLEN];
int  quoted;
char next_token[LINEMAXLEN];
char *keymap_ptr;
char line_buffer[LINEMAXLEN];
char temp_str[82];
int * quoted_ptr;
char quote_str[2];
struct stat fstat_buffer;
int fcode;
int * int_ptr;
int c,i;
int command_func_mapped = FALSE;
int ctrl_c_mapped = FALSE;
int finished = FALSE;

quoted_ptr = (int *) &quoted;
quote_str[0] = '\0';
quote_str[1] = '\0';

fptr = fopen(keymapname,"r");

if (fptr == NULL)
  {
  strcpy(keymapname,"medit.key");
  fptr = fopen(keymapname,"r");
  }

if (fptr == NULL)
  {
  default_key_map();
  if (keymap)
    next_mess("Missing keymap file - press CTRL-C then enter 'help'.");
  else
    next_mess("Default keymap - press CTRL-C then enter 'help'.");
  }
else
  {
  /* check that "keymapname" is an ordinary file */
  stat(keymapname,&fstat_buffer);
  if ((fstat_buffer.st_mode & S_IFDIR) == S_IFDIR)
    {
    sprintf(line_buffer,"%s is a directory.",keymapname);
    strcpy(errortext,line_buffer);
    finish(1);
    }
  else if ((fstat_buffer.st_mode & S_IFREG) != S_IFREG)
    {
    sprintf(line_buffer,"%s is not an ordinary file.",keymapname);
    strcpy(errortext,line_buffer);
    finish(1);
    }

  c = getline(fptr,line_buffer,LINEMAXLEN);
  if (c == EOF)
    {
    default_key_map();
    next_mess("Key map file empty - press CTRL-C then enter 'help'.");
    }
  else
    while (c != EOF)
      {
      keymap_ptr = (char *) &line_buffer;
  
      /* start parsing by skipping leading spaces */
      while (*keymap_ptr == ' ') keymap_ptr++;
  
      if (  (strcmp(line_buffer,"") != STR_SAME)
         && (*keymap_ptr != '\0')
         )
        {
        quote_str[0] = '\0';
        keymap_ptr = str_tok(first_token,keymap_ptr,quote_str,SQUASH_SPACES,NULL);
        quote_str[0] = '"';
        quoted = FALSE;
        keymap_ptr = str_tok(second_token,keymap_ptr,quote_str,NO_SQUASH,quoted_ptr);
        if (*keymap_ptr == '"') keymap_ptr++;
  
        quote_str[0] = '\0';

        int_ptr = (int *) &fcode;
 
        if (sscanf(first_token,"%X",int_ptr) != 1)
          {
          strcpy(errortext,"Invalid function code in key map file.");
          finish(1);
          }
        else
          if (strcmp(second_token,"") != STR_SAME)
            {
            while ((int)strlen(second_token) < 13)
              {
              strcpy(temp_str," ");
              strcat(temp_str,second_token);
              strcpy(second_token,temp_str);
              }

            insert_key_desc(second_token,fcode);
            }

        if (*keymap_ptr == '\0')
          {
          strcpy(errortext,"Key escape codes missing from key map file.");
          finish(1);
          }

        while (*keymap_ptr == ' ') keymap_ptr++;
        i = 0;
        while (  (*keymap_ptr != '\0')
              && (i < 80)
              )
          {
          keymap_ptr = str_tok(next_token,keymap_ptr,quote_str,SQUASH_SPACES,NULL);
          while (*keymap_ptr == ' ') keymap_ptr++;
          int_ptr = (int *) &esc_char[i];
          if (sscanf(next_token,"%X",int_ptr) != 1)
            {
            strcpy(errortext,"Invalid key escape code in key map file.");
            finish(1);
            }
          i += 1;
          }
        if (i > 79)
          {
          sprintf(errortext,"More than 80 key escape codes in '%s' mapping.",second_token);
          finish(1);
          }

        esc_char[i] = 0;
        insert_seq(esc_char,fcode);
  
        if (fcode == COMMAND_LINE)
          command_func_mapped = TRUE;

        if (fcode == PASTE_BLOCK)
          pasteblk_mapped = TRUE;

        if (fcode == DEL_BLOCK)
          deleteblk_mapped = TRUE;

        if (fcode == MOVE_BLOCK)
          moveblk_mapped = TRUE;

        if (fcode == COPY_BLOCK)
          copyblk_mapped = TRUE;

        if (fcode == MARK_BLOCK)
          markblk_mapped = TRUE;

        if (esc_char[0] == KB_CTRL_C)
          ctrl_c_mapped = TRUE;
        }

      c = getline(fptr,line_buffer,LINEMAXLEN);
      }

  if (!command_func_mapped)
  /* do them a favour - map COMMAND_LINE function if they havn't */
    {
    if (!ctrl_c_mapped)
      {
      esc_char[0] = KB_CTRL_C;
      esc_char[1] = 0;
      insert_seq(esc_char,COMMAND_LINE);
      insert_key_desc("       Ctrl-C",COMMAND_LINE);
      }
    else
      {
      esc_current = NULL;
      if (find_esc_char(KB_CTRL_C))
        {
        esc_current->next_level = NULL;
        esc_current->func_code = COMMAND_LINE;
        }
      else
        /* if they have already mapped CTRL-C we can not override this */
        {
        strcpy(errortext,"'Command Line' function not mapped in key map file.");
        finish(1);
        }
      }
    next_mess("'Command Line' function has been mapped to Ctrl-C.");
    }
  }
if (  (copyblk_mapped)
   && (deleteblk_mapped)
   && (pasteblk_mapped)
   && (moveblk_mapped)
   && (markblk_mapped)
   )
  blk_cmd_mapped = TRUE;
else
  blk_cmd_mapped = FALSE;

esc_current = NULL;
}

/*****************************************************************************/
void default_key_map()
{
int i,j,k;
unsigned char * p;
unsigned int esc_char[LINEMAXLEN];
char temp_str[LINEMAXLEN];
unsigned char temp_esc[LINEMAXLEN];
char key[][7] = 
  {"kbs","khlp","khome","kend","knp","kpp","kcub1","kcuf1","kcuu1","kcud1",""};
int func_code[] =
  {
  BACKSPACE,
  HELP,
  START_LINE,
  END_LINE,
  PAGEDOWN,
  PAGEUP,
  CURSOR_LEFT,
  CURSOR_RIGHT,
  CURSOR_UP,
  CURSOR_DOWN,
  };


/* attempt to automatically use real cursor keys in default key map */
k = 0;
while (key[k][0] != '\0')
  {
  i = 0;
  j = 0;
  if (get_key_str((unsigned char *) &temp_esc,key[k]))
    {
    while ((temp_esc[j] != '\0') && (i < LINEMAXLEN))
      {
      esc_char[i] = (unsigned int) temp_esc[j];
      j += 1;
      i += 1;
      }
    if (i <= LINEMAXLEN)
      {
      esc_char[i] = (unsigned int) temp_esc[j];
      insert_seq(esc_char,func_code[k]);
      }
    }
  k += 1;
  }

esc_char[0] = KB_CTRL_Y;
esc_char[1] = 0;
insert_seq(esc_char,DEL_LINE);
insert_key_desc("       Ctrl-Y",DEL_LINE);

esc_char[0] = KB_CTRL_Z;
esc_char[1] = 0;
insert_seq(esc_char,CONTROL_CHAR);
insert_key_desc("       Ctrl-Z",CONTROL_CHAR);

esc_char[0] = KB_CTRL_F;
esc_char[1] = 0;
insert_seq(esc_char,FIND_FOR);
insert_key_desc("       Ctrl-F",FIND_FOR);

esc_char[0] = KB_CTRL_T;
esc_char[1] = 0;
insert_seq(esc_char,START_FILE);
insert_key_desc("       Ctrl-T",START_FILE);

esc_char[0] = KB_CTRL_E;
esc_char[1] = 0;
insert_seq(esc_char,END_FILE);
insert_key_desc("       Ctrl-E",END_FILE);

esc_char[0] = KB_CTRL_B;
esc_char[1] = 0;
insert_seq(esc_char,MARK_BLOCK);
insert_key_desc("       Ctrl-B",MARK_BLOCK);

esc_char[0] = KB_CTRL_C;
esc_char[1] = 0;
insert_seq(esc_char,COMMAND_LINE);
insert_key_desc("       Ctrl-C",COMMAND_LINE);

esc_char[0] = KB_CTRL_V;
esc_char[1] = 0;
insert_seq(esc_char,PASTE_BLOCK);
insert_key_desc("       Ctrl-V",PASTE_BLOCK);

esc_char[0] = KB_CTRL_G;
esc_char[1] = 0;
insert_seq(esc_char,COPY_BLOCK);
insert_key_desc("       Ctrl-G",COPY_BLOCK);

esc_char[0] = KB_CTRL_K;
esc_char[1] = 0;
insert_seq(esc_char,MOVE_BLOCK);
insert_key_desc("       Ctrl-K",MOVE_BLOCK);

esc_char[0] = KB_BACKSPACE;
esc_char[1] = 0;
insert_seq(esc_char,BACKSPACE);

esc_char[0] = KB_DEL;
esc_char[1] = 0;
insert_seq(esc_char,BACKSPACE);

esc_char[0] = KB_NEWLINE;
esc_char[1] = 0;
insert_seq(esc_char,NEW_LINE);

esc_char[0] = KB_CTRL_W;
esc_char[1] = 0;
insert_seq(esc_char,TOGGLE_INS);
insert_key_desc("       Ctrl-W",TOGGLE_INS);

esc_char[0] = KB_CTRL_P;
esc_char[1] = 0;
insert_seq(esc_char,PAGEDOWN);
insert_key_desc("       Ctrl-P",PAGEDOWN);

esc_char[0] = KB_CTRL_O;
esc_char[1] = 0;
insert_seq(esc_char,PAGEUP);
insert_key_desc("       Ctrl-O",PAGEUP);

esc_char[0] = KB_CTRL_U;
esc_char[1] = 0;
insert_seq(esc_char,CURSOR_UP);
insert_key_desc("       Ctrl-U",CURSOR_UP);

esc_char[0] = KB_CTRL_D;
esc_char[1] = 0;
insert_seq(esc_char,CURSOR_DOWN);
insert_key_desc("       Ctrl-D",CURSOR_DOWN);

esc_char[0] = KB_CTRL_L;
esc_char[1] = 0;
insert_seq(esc_char,CURSOR_LEFT);
insert_key_desc("       Ctrl-L",CURSOR_LEFT);

esc_char[0] = KB_CTRL_R;
esc_char[1] = 0;
insert_seq(esc_char,CURSOR_RIGHT);
insert_key_desc("       Ctrl-R",CURSOR_RIGHT);

esc_char[0] = KB_CTRL_N;
esc_char[1] = 0;
insert_seq(esc_char,CLEAR_MARK);
insert_key_desc("       Ctrl-N",CLEAR_MARK);

esc_char[0] = KB_CTRL_X;
esc_char[1] = 0;
insert_seq(esc_char,DEL_BLOCK);
insert_key_desc("       Ctrl-X",DEL_BLOCK);

esc_current = NULL;
next_mess("Default keymap - press Ctrl-C then enter \"help\" for keys.");

copyblk_mapped = TRUE;
deleteblk_mapped = TRUE;
pasteblk_mapped = TRUE;
moveblk_mapped = TRUE;
markblk_mapped = TRUE;
}

/**************************/
void load_buffer()
{
strcpy(line_buffer,line_table_cur->text);
}

/*************************/
void show_str(char * text,int ly,int cx)
{
  cmove(ly,cx);
  n_waddstr(text,NO_VX_CHECK,cx);
  clreol();
}

/**************************/
int get_string(WINDOW * window,
                char command_str[],
                int funkeys_allowed,
                int screeny,
                int screenx,
                int save_command,
                int ctrl_allowed,
                int tab_ends_input)
{
int i,c;
int finished,ins_mode_copy,status_sx;
char * buffer_ptr;

i = 0;

ccx = screenx;

/* strcpy(command_str,""); */
ins_mode_copy = insert_mode;

finished = FALSE;

c = decode_key();

while (  (c != NEW_LINE)
      && (! finished)
      && (  (c != TAB)
         || (!tab_ends_input)
         )
      )
  {
/* add char to command line */
  if (  (  (!ctrl_allowed)
        && (  (  (c >= CHAR_SPACE)
              && (c < CHAR_DEL)
              )
           || (c == TAB)
           )
        )
     || (  (ctrl_allowed)
        && (c >= CHAR_NULL)
        && (c <= CHAR_255)
        )
     )
    {
    if (insert_mode)
      {
      if ((int)strlen(command_str) < COMMAND_MAX)
        {
        ins_char_str(command_str,i,c,COMMAND_MAX);
        i = i + 1;
        }
      }
    else
      {
      if (i < COMMAND_MAX)
        {
        if (command_str[i] == '\0')
          ins_char_str(command_str,i,c,COMMAND_MAX);
        else
          command_str[i] = c;
        i = i + 1;
        }
      }
    show_str(command_str,screeny,screenx);
    status_sx = calc_sx_str(command_str,i);
    ccx = status_sx + screenx;
    cmove(screeny,ccx);
    }

/* Insert - switch insert/overtype mode */

  else if (c == TOGGLE_INS)
    {
    if (insert_mode)
      insert_mode = FALSE;
    else
      insert_mode = TRUE;
    }

/* cursor up - previous command */
  else if (c == CURSOR_UP)
    {
    if (  (strcmp(prev_command,"") != STR_SAME)
       && (save_command)
       )
      {
      i = 0;
      strcpy(command_str,prev_command);
      show_str(command_str,screeny,screenx);
      status_sx = calc_sx_str(command_str,i);
      ccx = status_sx + screenx;
      cmove(screeny,ccx);
      }
    }

/* cursor down - empty command */
  else if (c == CURSOR_DOWN)
    {
    if (save_command)
      {
      i = 0;
      command_str[i] = '\0';
      show_str(command_str,screeny,screenx);
      status_sx = calc_sx_str(command_str,i);
      ccx = status_sx + screenx;
      cmove(screeny,ccx);
      }
    }

/* backspace */
  else if (  (c == BACKSPACE)
          || (  (c == DEL_LINE)
             && (ctrl_allowed)
             )
          )
    {
    if (i > 0)
      {
      shift_left_str(command_str,i,COMMAND_MAX);
      i = i - 1;
      show_str(command_str,screeny,screenx);
      status_sx = calc_sx_str(command_str,i);
      ccx = status_sx + screenx;
      cmove(screeny,ccx);
      }
    }

/* FORWARDS delete */
  else if (c == DEL_CHAR)
    {
    shift_left_str(command_str,i+1,COMMAND_MAX);
    show_str(command_str,screeny,screenx);
    status_sx = calc_sx_str(command_str,i);
    ccx = status_sx + screenx;
    cmove(screeny,ccx);
    }

/* cursor left */
  else if (c == CURSOR_LEFT)
    {
    if (i > 0)
      {
      i = i - 1;
      status_sx = calc_sx_str(command_str,i);
      ccx = status_sx + screenx;
      cmove(screeny,ccx);
      }
    }
/* cursor right */
  else if (c == CURSOR_RIGHT)
    {
/*  if (i < COMMAND_MAX) */
    if (i < (int) strlen(command_str))
      {
      i = i + 1;
      status_sx = calc_sx_str(command_str,i);
      ccx = status_sx + screenx;
      cmove(screeny,ccx);
      }
    }
/* help */
  else if (  (c == HELP)
          && (ctrl_allowed)
          )
    ctrl_char_help();
  else if (  (funkeys_allowed)
          && (c == HELP)
          )
    help();

  cmove(screeny,ccx);
  if (! finished)
    c = decode_key();
  }

/* truncate trailing spaces */
buffer_ptr = command_str;
if (*buffer_ptr != '\0')
  {
  while (*buffer_ptr != '\0')
    buffer_ptr++;
  buffer_ptr--;
  while (*buffer_ptr == ' ')
    buffer_ptr--;
  buffer_ptr++;
  *buffer_ptr = '\0';
  }

if (  (strcmp(command_str,"") != STR_SAME)
   && (save_command)
   )
  strcpy(prev_command,command_str);
ccx = -1;
insert_mode = ins_mode_copy;
return c;
}

/**************************/
void moveblock()
{
struct line_rec *line_table_ptr;
struct line_rec *line_table_prev;
int top_of_file = TRUE;

if (!block_marked)
  next_mess("No block marked.");
else
  {
  line_table_ptr = line_table_cur;
  line_table_prev = line_table_cur;
  up_no_upd();
  if (line_table_prev != line_table_cur)
    {
    line_table_prev = line_table_cur;
    down_no_upd();
    top_of_file = FALSE;
    }
  if (line_table_ptr->in_block)
    next_mess("Destination must be outside block.");
  else
    {
    deleteblock(FROM_MOVEBLOCK);
    line_table_cur = line_table_ptr;
    pasteblock(FROM_MOVEBLOCK);
    showblock(NO_HILITE);
    highlight_block(FALSE);
    block_marked = FALSE;
    line_table_block_start = line_table_block_end = NULL;
/*  showscreency((LINES-1)/2); */
    if (!top_of_file)
      {
      line_table_cur = line_table_prev;
      down_no_upd();
      showscreency(cy);
      }
    else
      showscreen();
    restore_status(ALL_STATUS,LINES-1);
    next_mess("Block moved.");
    file_needs_saving = TRUE;
    }
  }
}

/**************************/
void dup_block()
{
struct line_rec *line_table_ptr;

if (!block_marked)
  next_mess("No block marked.");
else
  {
  line_table_ptr = line_table_cur;

  copyblock(FROM_DELETEBLOCK);
  line_table_cur = line_table_ptr;
  pasteblock(FROM_MOVEBLOCK);
  showblock(NO_HILITE);
  highlight_block(FALSE);
  block_marked = FALSE;
  line_table_block_start = line_table_block_end = NULL;
/* showscreency((LINES-1)/2); */
  showscreen_cur();
  restore_status(ALL_STATUS,LINES-1);
  next_mess("Block copied.");
  file_needs_saving = TRUE;
  }
}

/**************************/
void command_line()
{
char command_str[LINEMAXLEN];
char command_copy[LINEMAXLEN];
char first_token[LINEMAXLEN];
char second_token[LINEMAXLEN];
int  second_quoted;
char third_token[LINEMAXLEN];
int  third_quoted;
char fourth_token[LINEMAXLEN];
char temp_str[LINEMAXLEN];
char prev_filename[LINEMAXLEN];
char quote_str[2];

char *token_ptr;
char *command_ptr;
char *ahead_ptr;

int intarg = 0;
int i,j;
int find_dir;
int finished;
int ok_to_write;
int ok_to_quit;
int * quoted_ptr;

char c;

strcpy(first_token,"");
strcpy(second_token,"");
strcpy(third_token,"");
strcpy(fourth_token,"");

cmove(LINES-1,0);
normal();
enter_bold();
clreol();
x_waddstr("Command:",NO_VX_CHECK,0);
normal();

cmove(LINES-1,9);

strcpy(command_str,"");
get_string(statusptr,command_str,FUNKEYS_ALLOWED,LINES-1,9,SAVE_COMMAND,NO_CTRL,NO_TAB_ENDS_INPUT);

/* make a copy of the command line string because strtok() tramples on it */
strcpy(command_copy,command_str);
command_ptr = (char *) &command_copy;

/* find the first quotes delimiter */
ahead_ptr = command_ptr;
/* while (strchr("'@`\\|/?[{]}-_=+\"$%^&*()@;:<>,.",*ahead_ptr) == NULL) */
while (strchr("'`\"",*ahead_ptr) == NULL)
  {
  ahead_ptr++;
  }

/* store the first quotes delimiter for use in terminating strings later */
quote_str[0] = *ahead_ptr;
quote_str[1] = '\0';

/* start parsing the command line by skipping leading spaces */
while (  (*command_ptr == ' ')
      || (*command_ptr == '\t')
      )
  command_ptr++;

if (*command_ptr == '!')
  {
  strcpy(first_token,"!");
  command_ptr++;
  strcpy(second_token,command_ptr);
  }
else
  {
  command_ptr = str_tok(first_token,command_ptr,quote_str,SQUASH_SPACES,NULL);
  quoted_ptr = (int *) &second_quoted;
  command_ptr = str_tok(second_token,command_ptr,quote_str,NO_SQUASH,quoted_ptr);
  quoted_ptr = (int *) &third_quoted;
  command_ptr = str_tok(third_token,command_ptr,quote_str,NO_SQUASH,quoted_ptr);
  if (  (third_quoted)
     && (*command_ptr == quote_str[0])
     )
    command_ptr++;
  command_ptr = str_tok(fourth_token,command_ptr,quote_str,SQUASH_SPACES,NULL);

  upper_case(first_token);
  zap_spaces(first_token);
  upper_case(fourth_token);
  zap_spaces(fourth_token);
  }

/* map , M , m : map keys */
  if (strcmp(first_token,"MAP") == STR_SAME)
    {
    map_functions();
    }

/* quit : quit without saving */
  else if (  (strcmp(first_token,"QUIT") == STR_SAME)
          || (strcmp(first_token,"Q") == STR_SAME)
          )
    {
    if (strcmp(filename,"") != STR_SAME)
      {
      if (file_needs_saving)
        {
        cmove(LINES-1,9);
        clreol();
        x_waddstr("File not saved. Do you want to save it (Y or N or Cancel)?",NO_VX_CHECK,0);
        cmove(LINES-1,68);
        strcpy(temp_str,"");
        get_string(statusptr,temp_str,NO_FUNKEYS,LINES-1,68,NO_SAVE,NO_CTRL,NO_TAB_ENDS_INPUT);
        upper_case(temp_str);
        zap_spaces(temp_str);
        if (  (strcmp(temp_str,"YES") == STR_SAME)
           || (strcmp(temp_str,"Y") == STR_SAME)
           )
          ok_to_write = TRUE;
        else
          ok_to_write = FALSE;
        if (  (strcmp(temp_str,"CANCEL") == STR_SAME)
           || (strcmp(temp_str,"C") == STR_SAME)
           )
          ok_to_quit = FALSE;
        else
          ok_to_quit = TRUE;
        }
      else
        {
        ok_to_write = FALSE;
        ok_to_quit = TRUE;
        }
      if (  (ok_to_write)
         && (ok_to_quit)
         )
        {
        save(SAVE_FILE);
        if (  (! save_failed)
           && (! save_abandoned)
           )
          {
          sprintf(errortext," %s saved ok.",filename);
          cmove(LINES-1,0);
          clreol();
         /* wrefresh(statusptr); */
          finish(0);
          }
        else if (! message_pending)
          next_mess("File not saved.");
        }
      else if (  (!ok_to_write)
              && (ok_to_quit)
              )
        {
        if (file_needs_saving)
          sprintf(errortext," %s NOT SAVED.",filename);
        else
          sprintf(errortext," quitted.");
        cmove(LINES-1,0);
        clreol();
        finish(0);
        }
      else
        next_mess("Quit cancelled.");
      }
    else
      {
      if (file_needs_saving)
        {
        cmove(LINES-1,9);
        clreol();
        x_waddstr("File not saved. Do you want to cancel the quit (Y or N)?",NO_VX_CHECK,0);
        cmove(LINES-1,66);
        strcpy(temp_str,"");
        get_string(statusptr,temp_str,NO_FUNKEYS,LINES-1,66,NO_SAVE,NO_CTRL,NO_TAB_ENDS_INPUT);
        upper_case(temp_str);
        zap_spaces(temp_str);
        if (  (strcmp(temp_str,"YES") == STR_SAME)
           || (strcmp(temp_str,"Y") == STR_SAME)
           )
          ok_to_quit = FALSE;
        else
          ok_to_quit = TRUE;
        }
      else
        ok_to_quit = TRUE;
      if (ok_to_quit)
        {
        if (file_needs_saving)
          sprintf(errortext," %s NOT SAVED.",filename);
        else
          sprintf(errortext," quitted.");
        cmove(LINES-1,0);
        clreol();
        /* wrefresh(statusptr); */
        finish(0);
        }
      else
        next_mess("Quit cancelled.");
      }
    }

/* exit : save & exit */
  else if (  (strcmp(first_token,"EXIT") == STR_SAME)
          || (strcmp(first_token,"E") == STR_SAME)
          )
    {
    if (  (strcmp(filename,"") == STR_SAME)
       || (strcmp(second_token,"") != STR_SAME)
       )
     {  
     if (strcmp(second_token,"") != STR_SAME)
      {
      if ((fptr = fopen(second_token,"r")) == NULL)
        ok_to_write = TRUE;
      else
        {
        fclose(fptr);
        cmove(LINES-1,9);
        clreol();
        x_waddstr("File exists. Do you want to overwrite it (Y or N)?",NO_VX_CHECK,0);
        cmove(LINES-1,60);
        strcpy(temp_str,"");
        get_string(statusptr,temp_str,NO_FUNKEYS,LINES-1,60,NO_SAVE,NO_CTRL,NO_TAB_ENDS_INPUT);
        upper_case(temp_str);
        zap_spaces(temp_str);
        if (  (strcmp(temp_str,"YES") == STR_SAME)
           || (strcmp(temp_str,"Y") == STR_SAME)
           )
          ok_to_write = TRUE;
        else
          ok_to_write = FALSE;
        }
      if (ok_to_write)
        {
        strcpy(prev_filename,filename);
        strncpy(filename,second_token,LINEMAXLEN-1);

        save(SAVE_FILE);
        if (  (!save_failed)
           && (!save_abandoned)
           )
          {
          sprintf(errortext," %s saved ok.",filename);
          cmove(LINES-1,0);
          clreol();
          finish(0);
          }

        else 
          {
          strcpy(filename,prev_filename);
          if (! message_pending)
            next_mess("File not saved.");
          }
        }
      else
        next_mess("File not saved.");
      }
      else
        next_mess("Filename must be specified.");
     }
    else
     {
      save(SAVE_FILE);
      if (  (! save_failed)
         && (! save_abandoned)
         )
        {
        sprintf(errortext," %s saved ok.",filename);
        cmove(LINES-1,0);
        clreol();
        finish(0);
        }
      else if (! message_pending)
              next_mess("File not saved.");
     }
    }

/* save , s , S : Save the file */
  else if (  (strcmp(first_token,"SAVE") == STR_SAME)
          || (strcmp(first_token,"S") == STR_SAME)
          )
    {
    if (strcmp(second_token,"") != STR_SAME)
      {
      if ((fptr = fopen(second_token,"r")) == NULL)
        ok_to_write = TRUE;
      else
        {
        fclose(fptr);
        cmove(LINES-1,9);
        clreol();
        x_waddstr("File exists. Do you want to overwrite it (Y or N)?",NO_VX_CHECK,0);
        cmove(LINES-1,60);
        strcpy(temp_str,"");
        get_string(statusptr,temp_str,NO_FUNKEYS,LINES-1,60,NO_SAVE,NO_CTRL,NO_TAB_ENDS_INPUT);
        upper_case(temp_str);
        zap_spaces(temp_str);
        if (  (strcmp(temp_str,"YES") == STR_SAME)
           || (strcmp(temp_str,"Y") == STR_SAME)
           )
          ok_to_write = TRUE;
        else
          ok_to_write = FALSE;
        }
      if (ok_to_write)
        {
        strcpy(prev_filename,filename);
        strncpy(filename,second_token,LINEMAXLEN-1);

        save(SAVE_FILE);
        if (  (!save_failed)
           && (!save_abandoned)
           )
          {
          next_mess("File saved ok.");
          sprintf(file_text,"File: %s",filename);
          }
 
        else
          {
          strcpy(filename,prev_filename);
          if (! message_pending)
            next_mess("File not saved.");
          }
        }
      else
        next_mess("File not saved.");
      }
    else
      {
      if (strcmp(filename,"") == STR_SAME)
        next_mess("Filename must be specified.");
      else
        {
        save(SAVE_FILE);
        if (  (!save_failed)
           && (!save_abandoned)
           )
           next_mess("File saved ok.");
        else if (! message_pending)
                next_mess("File not saved.");
        }
      }
    }

/* write to a new file name */
  else if (  (strcmp(first_token,"WRITE") == STR_SAME)
          || (strcmp(first_token,"W") == STR_SAME)
          )
    {
    if (strcmp(second_token,"") != STR_SAME)
      {
      if ((fptr = fopen(second_token,"r")) == NULL)
        ok_to_write = TRUE;
      else
        {
        fclose(fptr);
        cmove(LINES-1,9);
        clreol();
        x_waddstr("File exists. Do you want to overwrite it (Y or N)?",NO_VX_CHECK,0);
        cmove(LINES-1,60);
        strcpy(temp_str,"");
        get_string(statusptr,temp_str,NO_FUNKEYS,LINES-1,60,NO_SAVE,NO_CTRL,NO_TAB_ENDS_INPUT);
        upper_case(temp_str);
        zap_spaces(temp_str);
        if (  (strcmp(temp_str,"YES") == STR_SAME)
           || (strcmp(temp_str,"Y") == STR_SAME)
           )
          ok_to_write = TRUE;
        else
          ok_to_write = FALSE;
        }
      if (ok_to_write)
        {
        strcpy(prev_filename,filename);
        strncpy(filename,second_token,LINEMAXLEN-1);

        if (block_marked)
          {
          save(SAVE_BLOCK);
          if (  (!save_failed)
             && (!save_abandoned)
             )
            {
            next_mess("Block saved ok.");
            strcpy(filename,prev_filename);
            }

          else
            {
            strcpy(filename,prev_filename);
            if (! message_pending)
              next_mess("Block not written.");
            }
          }
        else
          {
          save(SAVE_FILE);
          if (  (!save_failed)
             && (!save_abandoned)
             )
            {
            next_mess("File saved ok.");
            sprintf(file_text,"File: %s",filename);
            }

          else
            {
            strcpy(filename,prev_filename);
            if (! message_pending)
              next_mess("File not written.");
            }
          }
        }
      else
        next_mess("File not written.");
      }
    else
      next_mess("Filename must be specified.");
    }

/* execute unix command */
  else if (strcmp(first_token,"!") == STR_SAME)
    {
    if (strcmp(second_token,"") != STR_SAME)
      {
      /* flushinp(); */

      if (ioctl(0,TCSETAF,(struct termio *) &old_term))
        next_mess("Unable to RESET initial terminal characteristics.");
      strcpy(temp_str,"tput clear;");
      strcat(temp_str,second_token);
      strcat(temp_str,";echo \"** Press any key to return to Medit **\"");
      strcpy(second_token,temp_str);

      system(second_token);

      if (ioctl(0,TCSETAF,(struct termio *) &new_term))
        next_mess("Unable to setup terminal: ioctl() failed.");

      c = decode_key();
      /* while((c = x_getch()) == ERR); */

      if (strlen(mess_text) == 0)
        next_mess(find_text);
      showscreen();
      restore_status(ALL_STATUS,LINES-1);
      }
    }

/* paste from a file */
  else if (  (strcmp(first_token,"FILE") == STR_SAME)
          || (strcmp(first_token,"F")    == STR_SAME)
          )
    {
    if (strcmp(second_token,"") != STR_SAME)
      {
      readfile(second_token,PASTE_FILE);
      }
    }

/* mark block */
  else if (  (strcmp(first_token,"BLOCK") == STR_SAME)
          || (strcmp(first_token,"B")    == STR_SAME)
          )
    {
    markblock();
    }

/* paste block */
  else if (  (strcmp(first_token,"PASTE") == STR_SAME)
          || (strcmp(first_token,"P")    == STR_SAME)
          )
    {
    pasteblock(NOT_FROM_MOVEBLOCK);
    }

/* move block */
  else if (  (strcmp(first_token,"MOVE") == STR_SAME)
          || (strcmp(first_token,"M")    == STR_SAME)
          )
    {
    moveblock();
    }

/* delete block */
  else if (  (strcmp(first_token,"DELETE") == STR_SAME)
          || (strcmp(first_token,"D")    == STR_SAME)
          )
    {
    deleteblock(NOT_FROM_MOVEBLOCK);
    }

/* a number , eg 999 : goto that line */
  else if (sscanf(first_token,"%d",&intarg) == 1)
    {
    if (intarg != 0)
      gotoline(intarg);
    }


/* help , H , h : show help screen */
  else if (  (strcmp(first_token,"HELP") == STR_SAME)
          || (strcmp(first_token,"H") == STR_SAME)
          )
    {
    help();
    }


/* replace 'string' command */
  else if  (  (strcmp(first_token,"REPLACE") == STR_SAME)
           || (strcmp(first_token,"R") == STR_SAME)
           )
    {
    if (strcmp(second_token,"") == STR_SAME)
      next_mess("String to replace must not be empty.");
    else if (!second_quoted)
      next_mess("String to replace must be in quotes.");
    else if (!third_quoted)
      next_mess("Replacement string must be in quotes.");
    else
      {
      if (  (strcmp(fourth_token,"ALL") == STR_SAME)
         || (strcmp(fourth_token,"A") == STR_SAME)
         || (strcmp(fourth_token,"QUERY") == STR_SAME)
         || (strcmp(fourth_token,"Q") == STR_SAME)
         || (strcmp(fourth_token,"") == STR_SAME)
         )
        {
        strncpy(findstr,second_token,80);
        strcpy(find_text,quote_str);
        strncat(find_text,findstr,22);
        strcat(find_text,quote_str);

        replace(findstr,third_token,fourth_token);
        }
      else
        next_mess("Replace option not recognised.");
      }
    }


/* 'string' , "string" : find next line containing string */

  else if (  (strcmp(second_token,"") != STR_SAME)
          && (  (strcmp(first_token,"") == STR_SAME)
             || (strcmp(first_token,"CASE") == STR_SAME)
             || (strcmp(first_token,"C") == STR_SAME)
             )
          && (strcmp(third_token,"") == STR_SAME)
          && (strcmp(fourth_token,"") == STR_SAME)
          )
    {
    if (!second_quoted)
      next_mess("String to find must be in quotes.");
    else
      {
      find_dir = FORWARDS;

      if (  (strcmp(first_token,"CASE") == STR_SAME)
         || (strcmp(first_token,"C") == STR_SAME)
         )
        case_sens = TRUE;
      else
        case_sens = FALSE;

      strncpy(findstr,second_token,80);
      if (case_sens)
        {
        strcpy(find_text,"Case ");
        strcat(find_text,quote_str);
        }
      else
        strcpy(find_text,quote_str);
      strncat(find_text,findstr,22);
      strcat(find_text,quote_str);

      if (find(findstr,find_dir,case_sens))
        {
        if (find_dir == FORWARDS)
          {
          cur_sx = calc_sx_str(line_table_cur->text,cx);

          if (  (cur_sx > (vx + COLS - 2 - (int)strlen(findstr)))
             || (cur_sx < vx)
             || (  (cur_sx < (COLS - 2 - (int)strlen(findstr)))
                && (vx > 0)
                )
             )
            {
            vx = cur_sx - COLS + 2 + (int)strlen(findstr);
            if (vx < 0) vx = 0;
            use_new_cy = FALSE;
            }

          showscreency(LINES/4);
          }
        else
          {
          cur_sx = calc_sx_str(line_table_cur->text,cx);

          if (  (cur_sx > (vx + COLS - 2 - (int)strlen(findstr)))
             || (cur_sx < vx)
             || (  (cur_sx < (COLS - 2 - (int)strlen(findstr)))
                && (vx > 0)
                )
             )
            {
            vx = cur_sx - COLS + 2 + (int)strlen(findstr);
            if (vx < 0) vx = 0;
            use_new_cy = FALSE;
            }

          showscreency(LINES/4*3);
          }
        next_mess(find_text);
        }
      else
        next_mess("String not found.");
      }
    }
/* copy block */
  else if (  (strcmp(first_token,"COPY") == STR_SAME)
          || (strcmp(first_token,"C")    == STR_SAME)
          )
    {
    dup_block();
    }

  else
    {
    if (strcmp(command_str,"") == STR_SAME)
      next_mess("No command entered.");
    else
      /* unrecognised command */
      next_mess("Command not recognised.");
    }

restore_status(ALL_STATUS,LINES-1);
}

/*************************/
void control_char()
{
char command_str[LINEMAXLEN];
char first_token[LINEMAXLEN];
char * command_ptr;
int intarg = 0;

command_ptr = (char *) &command_str;

cmove(LINES-1,0);
enter_bold();
clreol();
x_waddstr("Insert control character: ",NO_VX_CHECK,0);
normal();

cmove(LINES-1,26);
clreol();

strcpy(command_str,"");
get_string(statusptr,command_str,NO_FUNKEYS,LINES-1,26,NO_SAVE,CTRL_ALLOWED,NO_TAB_ENDS_INPUT);

/* start parsing the command line by skipping leading spaces */
while (*command_ptr == ' ') command_ptr++;

if (*command_ptr < ' ')
  {
  if (  (*command_ptr != '\0')
     && (*command_ptr != NEW_LINE)
     )
    insert_char(*command_ptr);
  else
    next_mess("No control char inserted.");
  }
else
  {
  command_ptr = str_tok(first_token,command_ptr," ",SQUASH_SPACES,NULL);
  zap_spaces(first_token);
  if (sscanf(first_token,"%d",&intarg) == 1)
    {
    if (  (intarg > 0)
       && (intarg <= 255)
       && (intarg != NEW_LINE)
       )
      {
      if (intarg != '\0')
        insert_char(intarg);
      else
        next_mess("No control char inserted.");
      }
    else
      next_mess("Invalid character code.");
    }
  else
    next_mess("Invalid character code.");
  }
restore_status(ALL_STATUS,LINES-1);
}

/*************************/
char * str_tok(
  char * target,
  char * source,
  char * delimiter,
  int squash_spaces,
  int * quoted)
{
if (squash_spaces)
  while (  (*source != *delimiter)
        && (*source != '\0')
        && (  (*source == ' ')
           || (*source == '\t')
           )
        )
    source++;

if (  (*delimiter != '\0')
   && (*source == *delimiter)
   && (quoted != NULL)
   )
  {
  source++;
  *quoted = TRUE;
  }

if (squash_spaces)
  while (  (*source != *delimiter)
        && (*source != '\0')
        && (*source != ' ')
        && (*source != '\t')
        )
    {
    *target = *source;
    source++;
    target++;
    }
else
  while (  (*source != *delimiter)
        && (*source != '\0')
        )
    {
    *target = *source;
    source++;
    target++;
    }

if (squash_spaces)
  while (  (*source != *delimiter)
        && (*source != '\0')
        && (  (*source == ' ')
           || (*source == '\t')
           )
        )
    source++;

*target = '\0';
return source;
}

/**************************/
void shift_left_str(char * string, int position,int max_len)
/* shift chars (may be a subset) in a string to the left */
{
int count;
char * dest;
char * origin;

if (  (position <= max_len)
   && (position > 0)
   )
  {
  count = 0;
  origin = string;
  while (count < position)
    {
    origin++;
    count += 1;
    }

  dest = origin;
  dest--;

  while (*dest != '\0')
    {
    *dest = *origin;
    dest++;
    origin++;
    count += 1;
    }
  }
}

/**************************/
void shift_right_str(char * string,int position,int max_len)
/* shift chars (may be a subset) in a string to the right */
{
int count,total;
char * dest;
char * origin;

/* find the position of the substring to shifted right */
count = 0;
origin = string;

while (  (count < position)
      && (count < max_len)
      )
  {
  if (*origin == '\0')
    {
    *origin = ' ';
    origin++;
    *origin = '\0';
    }
  else
    origin++;
  count += 1;
  }
total = count;

/* now count the number of chars in the substring that have to be shifted */
count = 0;
dest = origin;
while (*dest != '\0')
  {
  dest++;
  count += 1;
  }
total = total + count;

/* if the end of the substring ( = null) is less than the max length of the
   allocated space, then the null is to be shifted right with the substring,
   and the destination of the substring is one after the null.

   if, hwoever the '\0' terminater is at the max length of the allocated space,
   then the null must be left where it is intact, and the destination of the
   substring is the one BEFORE the null.
*/

if (total == max_len)
  {
  dest--;
  count -= 1;
  }
else
  {
  dest++;
  count += 1;
  }

/* on a move right in situ, the chars have to be moved at the right end first,
   working BACKWARDS/leftwards towards the start of the substring.
*/
origin = dest;
origin--;

while (count > 0)
  {
  *dest = *origin;
  *origin = ' ';
  dest--;
  origin--;
  count -= 1;
  }

}

/**************************/
void ins_char_str(char string[], int position, char c, int max_len)
/* insert char into string */
{
shift_right_str(string,position,max_len);
string[position] = c;
}

/**************************/
void winattr(WINDOW * window)
{
if (window == winptr)
  {
  normal();
  if (win_reverse == A_REVERSE)
    enter_inv();

  if (win_bold == A_BOLD)
    enter_bold();

  if (win_underline == A_UNDERLINE)
    enter_under();
  }

if (window == statusptr)
  {
  normal();
  if (status_reverse == A_REVERSE)
    enter_inv();

  if (status_bold == A_BOLD)
    enter_bold();

  if (status_underline == A_UNDERLINE)
    enter_under();
  }

if (window == helpptr)
  {
  normal();
  if (help_reverse == A_REVERSE)
    enter_inv();

  if (help_bold == A_BOLD)
    enter_bold();

  if (help_underline == A_UNDERLINE)
    enter_under();
  }
}

/**************************/
void calc_sx_ch(int c)
{
int i;

  if (  (  (c >= ' ')
        && (c <= '\176')
        )
     )
    sx = sx + 1;
  else
    if (c == TAB)
      {
      i = 1;
      while (  (tabstop[i] <= sx)
            && (i < 80)
            && (tabstop[i] < 32000)
            )
        i = i + 1;

      if (  (i < 80)
         && (tabstop[i] < 32000)
         )
        while (sx < tabstop[i])
          sx = sx + 1;
      }
    else
      sx = sx + 1;
}

/**************************/
int calc_sx_str(char * string,int cx)
{
int c = 0;
sx = 0;
while (  (*string != '\0')
      && (c < cx)
      )
  {
  calc_sx_ch((unsigned char)*string);
  string++;
  c += 1;
  }
return sx;
}

/**************************/
void x_waddch(WINDOW * winptr,int c,int no_vx_check)
{
int i;

  if (  (  (c >= ' ')
        && (c <= '\176')
        )
     )
    {
    if (  (no_vx_check)
       || (sx >= vx)
       )
      x_buffch(c);
    sx = sx + 1;
    }
  else
    if (c == TAB)
      {
      if (highlight_tabs)
        {
        enter_inv();
        enter_under();
        enter_bold();
        }
      i = 1;
      while (  (tabstop[i] <= sx)
            && (i < 80)
            && (tabstop[i] < 32000)
            )
        i = i + 1;
      if (  (i < 80)
         && (tabstop[i] < 32000)
         )
        while (sx < tabstop[i])
          {
          if (  (no_vx_check)
             || (sx >= vx)
             )
            x_buffch(' ');
          sx = sx + 1;
          }
      if (highlight_tabs)
        winattr(winptr);
      }
    else
      {
      enter_inv();
      enter_under();
      enter_bold();
      if (c > '\176')
        {
        if (  (no_vx_check)
           || (sx >= vx)
           )
          x_buffch('?');
        sx = sx + 1;
        }
      else
        {
        if (c == '\0')
          {
          if (  (no_vx_check)
             || (sx >= vx)
             )
            x_buffch(' ');
          sx = sx + 1;
          }
        else
          {
          if (  (no_vx_check)
             || (sx >= vx)
             )
            x_buffch(c+'A'-1);
          sx = sx + 1;
          }
        }
      winattr(winptr);
      }
}

/**************************/
void x_waddstr(char * string,int no_vx_check,int init_sx)
{
sx = init_sx;

ovr_mode();

while (  (*string != '\0')
      && (sx < (vx + COLS - 1))
      )
  {
  x_waddch(winptr,(unsigned char)*string,no_vx_check);
  string++;
  }
}

/**************************/
void n_waddch(WINDOW * winptr,int c,int no_vx_check,int init_cx,int cx)
{
int i;
int n = 0;

  if (  (  (c >= ' ')
        && (c <= '\176')
        )
     )
    {
    if (  (no_vx_check)
       || (  (sx >= vx)
          && (cx >= init_cx)
          )
       )
      x_buffch(c);
    sx = sx + 1;
    }
  else
    if (c == TAB)
      {
      if (highlight_tabs)
        {
        enter_inv();
        enter_under();
        enter_bold();
        }
      i = 1;
      while (  (tabstop[i] <= sx)
            && (i < 80)
            && (tabstop[i] < 32000)
            )
        i = i + 1;
      if (  (i < 80)
         && (tabstop[i] < 32000)
         )
        while (sx < tabstop[i])
          {
          if (  (no_vx_check)
             || (  (sx >= vx)
                && (cx >= init_cx)
                )
             )
            x_buffch(' ');
          sx = sx + 1;
          }
      if (highlight_tabs)
        winattr(winptr);
      }
    else
      {
      enter_inv();
      enter_under();
      enter_bold();
      if (c > '\176')
        {
        if (  (no_vx_check)
           || (  (sx >= vx)
              && (cx >= init_cx)
              )
           )
          x_buffch('?');
        sx = sx + 1;
        }
      else
        {
        if (c == '\0')
          {
          if (  (no_vx_check)
             || (  (sx >= vx)
                && (cx >= init_cx)
                )
             )
            x_buffch(' ');
          sx = sx + 1;
          }
        else
          {
          if (  (no_vx_check)
             || (  (sx >= vx)
                && (cx >= init_cx)
                )
             )
            x_buffch(c+'A'-1);
          sx = sx + 1;
          }
        }
      winattr(winptr);
      }
}

/**************************/
void n_waddstr(char * string,int no_vx_check,int init_cx)
{
int i=0;

sx = 0;

ovr_mode();

while (  (*string != '\0')
      && (sx < (vx + COLS - 1))
      )
  {
  n_waddch(winptr,(unsigned char)*string,no_vx_check,init_cx,i);
  string++;
  i++;
  }
}

/**************************/
/*
void x_winsch(WINDOW * winptr,int c,int no_vx_check)
{
ins_mode();

  if (  (  (c >= ' ')
        && (c <= '\176')
        )
     )
    {
    if (  (no_vx_check)
       || (sx >= vx)
       )
      x_buffch(c);
    sx = sx + 1;
    }
  else
    if (c == TAB)
      {
      if (highlight_tabs)
        {
        enter_inv();
        enter_under();
        enter_bold();
        }
      i = 1;
      while (  (tabstop[i] <= sx)
            && (i < 80)
            && (tabstop[i] < 32000)
            )
        i = i + 1;
      if (  (i < 80)
         && (tabstop[i] < 32000)
         )
        while (sx < tabstop[i])
          {
          if (  (no_vx_check)
             || (sx >= vx)
             )
            x_buffch(' ');
          sx = sx + 1;
          }
      if (highlight_tabs)
        winattr(winptr);
      }
    else
      {
      enter_inv();
      enter_under();
      enter_bold();
      if (c > '\176')
        {
        if (  (no_vx_check)
           || (sx >= vx)
           )
          x_buffch('?');
        sx = sx + 1;
        }
      else
        {
        if (c == '\0')
          {
          if (  (no_vx_check)
             || (sx >= vx)
             )
            x_buffch(' ');
          sx = sx + 1;
          }
        else
          {
          if (  (no_vx_check)
             || (sx >= vx)
             )
            x_buffch(c+'A'-1);
          sx = sx + 1;
          }
        }
      winattr(winptr);
      }

ovr_mode();
}
*/

/**************************************/
char * str_str(char *string,char *pattern)
{
int len;
len = strlen(pattern);

while (  (strncmp(string,pattern,len) != STR_SAME)
      && (*string != '\0')
      )
  string++;

if (*string == '\0')
  return NULL;
else
  return string;
}

/**************************************/
void zap_spaces(char *source)
{
char *dest;

dest = source;
while (*dest != '\0')
  {
  if (  (*source != ' ')
     && (*source != '\t')
     )
    {
    *dest = *source;
    source++;
    dest++;
    }
  else
    source++;
  }
}

/**************************************/
void new_line()
{
if (cx == 0)
  {
  file_needs_saving = TRUE;
  if (cy < (LINES-2))
    {
    insert_line();
    insline();
    restore_status(ALL_STATUS,LINES-1);
    cy = cy + 1;
    recalc_screen_cy(cy);
    cmove(cy,0);
    }
  else
    {
    insert_line();
    if (line_table_top->next_line != NULL)
      line_table_top = line_table_top->next_line;
    recalc_screen_cur();
    cmove(0,0);
    deline();
    cmove(LINES-3,0);
    insline();
    cx = 0;
    cur_sx = calc_sx_str(line_table_cur->text,cx);
    recalc_screen_cur();
    restore_status(ALL_STATUS,LINES-1);
    }
  }
else
  {
  if (cx > ((int)(strlen(line_table_cur->text) - 1)))
    {
    if (line_table_cur != line_table_end)
      {
      if (cy < (LINES-2))
        {
        down_no_upd();
        insert_line();
        up_no_upd();
        cy += 1;
        recalc_screen_cy(cy);
        cx = 0;
        cur_sx = calc_sx_str(line_table_cur->text,cx);
        cmove(cy,0);
        insline();
        restore_status(ALL_STATUS,LINES-1);
        file_needs_saving = TRUE;
        }
      else
/* its the last line on the screen */
        {
        down_no_upd();
        insert_line();
        if (line_table_cur->prev_line != NULL)
          line_table_cur = line_table_cur->prev_line;
        recalc_screen_cy(cy);
        cmove(0,0);
        deline();
        cmove(LINES-2,0);
        insline();
        cx = 0;
        cur_sx = calc_sx_str(line_table_cur->text,cx);
        cmove(cy,cur_sx-vx);
        }
      }
    else
/* its the last line in the file */
      {
      file_needs_saving = TRUE;
      if (cy < (LINES-2))
        {
        insert_line_end();
        cy += 1;
        recalc_screen_cy(cy);
        cx = 0;
        cur_sx = calc_sx_str(line_table_cur->text,cx);
        cmove(cy,0);
        insline();
        restore_status(ALL_STATUS,LINES-1);
        }
      else
/* its the last line in the file & screen */
        {
        insert_line_end();
        down_no_upd();
        cmove(0,0);
        deline();
        cmove(LINES-2,0);
        insline();
        cx = 0;
        if (line_table_top->next_line != NULL)
          line_table_top = line_table_top->next_line;
        recalc_screen_cy(cy);
        cur_sx = calc_sx_str(line_table_cur->text,cx);
        cmove(cy,cur_sx-vx);
        }
      }
    }
  else
/* its a split line */
    {
    if (cy < (LINES-2))
      {
      split_line();
      clreol();
      cy = cy + 1;
      recalc_screen_cur();
      cx = 0;
      cur_sx = calc_sx_str(line_table_cur->text,cx);
      cmove(cy,0);
      insline();
      showline(line_table_cur,line_table_cur->text,cy,0,HILITE);
      cmove(cy,cur_sx-vx);
      restore_status(ALL_STATUS,LINES-1);
      }
    else
/* its the last line on the screen */
      {
      split_line();
      clreol();
      if (line_table_top->next_line != NULL)
        line_table_top = line_table_top->next_line;
      recalc_screen_cur();
      cmove(0,0);
      deline();
      cmove(LINES-2,0);
      showline(line_table_cur,line_table_cur->text,cy,0,HILITE);
      cx = 0;
      cur_sx = calc_sx_str(line_table_cur->text,cx);
      cmove(cy,cur_sx-vx);
      restore_status(ALL_STATUS,LINES-1);
      }
    file_needs_saving = TRUE;
    }
  }
cur_sx = calc_sx_str(line_table_cur->text,cx);

if (cur_sx < vx)
  {
  vx = cur_sx - (COLS/3);
/*  vx = cur_sx - 19; */
  if (vx < 0) vx = 0;
  showscreen();
  }
}

/**************************************/
void delete_char()
/* text only - not command line */
{
  if (cx > (int) strlen(line_buffer))
    cx = (int) strlen(line_buffer);

 if (cx < ((int)strlen(line_buffer)))
   /* then cursor is before end of line */
   {
   shift_left_str(line_buffer,cx+1,LINEMAXLEN);
/* showline(line_table_cur,line_buffer,cy,cx,HILITE); */
   cur_sx = calc_sx_str(line_buffer,cx);
/* delchar(); */
   showline_right(line_table_cur,line_buffer,cy,cx,HILITE);
   cmove(cy,cur_sx-vx);
   file_needs_saving = TRUE;
   line_needs_updating = TRUE;
   }
 else
   if (line_table_cur->next_line != NULL)
     {
     if (line_needs_updating)
       updateline(LEAVE_TRAILING_SPACES);
     cx = strlen(line_table_cur->text);
     line_table_cur = join_ptr(line_table_cur->next_line);
     showline(line_table_cur,line_table_cur->text,cy,cx,HILITE);
     cur_sx = calc_sx_str(line_table_cur->text,cx);
     if (cy < (LINES-2))
       {
       cmove(cy+1,0);
       deline();
       if (line_table_bot->next_line != NULL)
         {
         line_table_bot = line_table_bot->next_line;
         show_bot_line();
         restore_status(ALL_STATUS,LINES-1);
         }
       else
         {
         cmove(LINES-2,0);
         insline();
         }
       }
     cmove(cy,cur_sx-vx);
     file_needs_saving = TRUE;
     load_buffer();
     }

 if (  (line_table_cur->prev_line == NULL)
    && (line_table_cur->next_line == NULL)
    && ((int)strlen(line_buffer) == 0)
    && (! file_empty)
    )
   {
   file_empty = TRUE;
   restore_status(ALL_STATUS,LINES-1);
   }
}

/**************************************/
void backspace()
/* text only - not command line */
{
struct line_rec *line_table_temp;

 if (cx > (int) strlen(line_buffer))
   cx = (int) strlen(line_buffer);

 if (cx > 0)
     /* then cursor is after beginning of line */
     {
     shift_left_str(line_buffer,cx,LINEMAXLEN);
     cx = cx - 1;
     cur_sx = calc_sx_str(line_buffer,cx);
     if (cur_sx < vx)
       {
       vx = cur_sx - (COLS/3);
/*     vx = cur_sx - 19; */
       if (vx < 0) vx = 0;
       updateline(LEAVE_TRAILING_SPACES);
       showscreen();
       }
     else
       {
       cmove(cy,cur_sx-vx);
/*     delchar(); */
       showline_right(line_table_cur,line_buffer,cy,cx,HILITE);
       cmove(cy,cur_sx-vx);
       line_needs_updating = TRUE;
       }
     file_needs_saving = TRUE;
     }
 else
     {
     if (line_needs_updating)
       updateline(LEAVE_TRAILING_SPACES);

     line_table_temp = line_table_cur->prev_line;
     if (line_table_temp != NULL)
       {
       cx = strlen(line_table_temp->text);

       line_table_cur = join_ptr(line_table_cur);

       if (cy == 0)
         {
         if (line_table_top->prev_line != NULL)
           {
           line_table_top = line_table_top->prev_line;
           line_table_cur = line_table_top;
           cmove(cy,0);
           clreol();
           cur_sx = calc_sx_str(line_table_cur->text,cx);
           showline(line_table_cur,line_table_cur->text,cy,cx,HILITE);
           recalc_screen_cur();
           cmove(cy,cur_sx-vx);
           }
         }
       else
         {
         cmove(LINES-1,0);
         clreol();
         cmove(cy,0);
         deline();
         cy = cy - 1;
         cur_sx = calc_sx_str(line_table_cur->text,cx);
         showline(line_table_cur,line_table_cur->text,cy,cx,HILITE);
         recalc_screen_cy(cy);
         show_bot_line();
         restore_status(ALL_STATUS,LINES-1);
         }

       if (cur_sx > (vx + COLS - 1))
         {
         vx = cur_sx - COLS + (COLS/3);
/*       vx = cur_sx - COLS + 21; */
         showscreen();
         }
       file_needs_saving = TRUE;
       load_buffer();
       }
     }

 if (  (line_table_cur->prev_line == NULL)
    && (line_table_cur->next_line == NULL)
    && ((int)strlen(line_buffer) == 0)
    && (! file_empty)
    )
   {
   file_empty = TRUE;
   restore_status(ALL_STATUS,LINES-1);
   }
}

/**************************************/
void insert_char(int c)
{
  int x;

  if (cx > (int) strlen(line_buffer))
    cx = (int) strlen(line_buffer);

  if (cur_sx > (vx + COLS - 1))
    {
    vx = cur_sx - COLS + (COLS/3);
/*  vx = cur_sx - COLS + 21; */
    showscreen();
    }
  else
    if (cur_sx < vx)
      {
      vx = cur_sx - (COLS/3);
/*    vx = cur_sx - 19; */
      if (vx < 0) vx = 0;
      showscreen();
      }

  if (line_table_cur->in_block == TRUE)
    enter_inv();
  else
    normal();

  if (insert_mode)
    {
    if ( (int)strlen(line_buffer) >= (LINEMAXLEN - 1) )
      next_mess("Line is maximum length.");
    else
      {
      ins_char_str(line_buffer,cx,c,LINEMAXLEN);
      cx+=1;
      cur_sx = calc_sx_str(line_buffer,cx);
      if (cur_sx > (vx + COLS - 1))
        {
        vx = cur_sx - COLS + (COLS/3);
/*      vx = cur_sx - COLS + 21; */
        updateline(LEAVE_TRAILING_SPACES);
        showscreen();
        }
      else
        {
/*      showline(line_table_cur,line_buffer,cy,cx,HILITE); */
        cx -= 1;
        cur_sx = calc_sx_str(line_buffer,cx);
        cmove(cy,cur_sx-vx);
        showline_right(line_table_cur,line_buffer,cy,cx,HILITE);
        cx += 1;
        cur_sx = calc_sx_str(line_buffer,cx);
        cmove(cy,cur_sx-vx);
        }
      }
    }
  else
    {
    if (cx >= ((int)strlen(line_buffer)))
      ins_char_str(line_buffer,cx,c,LINEMAXLEN);
    else
      line_buffer[cx] = c;
    cx+=1;
    cur_sx = calc_sx_str(line_buffer,cx);
    if (cur_sx > (vx + COLS - 1))
      {
      vx = cur_sx - COLS + (COLS/3);
/*    vx = cur_sx - COLS + 21; */
      updateline(LEAVE_TRAILING_SPACES);
      showscreen();
      }
    else
      {
      showline(line_table_cur,line_buffer,cy,cx,HILITE);
      cmove(cy,cur_sx-vx);
      }
    }

  if (file_empty)
    {
    file_empty = FALSE;
    restore_status(ALL_STATUS,LINES-1);
    }

  line_needs_updating = TRUE;
  file_needs_saving = TRUE;
  }

/**************************************/
void restore_status(int update_what,int y)
{
if (!message_pending)
  if (!scrolling)
    {
    if (update_what == ALL_STATUS)
      {
      cmove(y,0);
      clreol();
      normal();
      enter_bold();
      x_waddstr(medit_text,NO_VX_CHECK,0);
      }

    if (COLS > 15)
      {
      cmove(y,15);
      normal();
      enter_bold();
      if (insert_mode)
        x_waddstr("Ins  ",NO_VX_CHECK,15);
      else
        x_waddstr("Ovr  ",NO_VX_CHECK,15);
      }

    if (update_what == ALL_STATUS)
      {
      if (COLS > 20)
        {
        cmove(y,20);
        clreol();
        if (strcmp(mess_text,"") != STR_SAME)
          {
          normal();
          enter_blink();
          x_waddstr(mess_text,NO_VX_CHECK,20);
          winattr(statusptr);
          }
        else
          {
          if (power_failure)
            {
            normal();
            enter_blink();
            x_waddstr("POWER FAILURE- save and exit!",NO_VX_CHECK,20);
            }
          else if (file_empty)
            {
            normal();
            enter_bold();
            x_waddstr("Empty file.",NO_VX_CHECK,20);
            }
          else if (block_marked)
            {
            normal();
            enter_blink();
            x_waddstr("Block marked.",NO_VX_CHECK,20);
            winattr(statusptr);
            }
          else
            {
            normal();
            enter_bold();
            x_waddstr(find_text,NO_VX_CHECK,20);
            }
          }
        }
      }

    if (  (update_what == ALL_STATUS)
       && (((int)strlen(mess_text)) < 30)
       )
      {
      if (COLS > 50)
        {
        cmove(y,50);
        normal();
        enter_bold();
        clreol();
        x_waddstr(file_text,NO_VX_CHECK,50);
        }
      }
    cmove(y,COLS-1);
    clrbot();
    }
normal();
cmove(cy,cur_sx-vx);
}

/**************************************/
char * replace_string(int *len,char *line,char *search,char *replace,int *cx)
{
int i,count;
char * new_line;
char * dest;
char * orig;

count = (int) strlen(replace) - (int) strlen(search);

*len = (int) strlen(line);
if (count > 0)
  *len += count;

if (*len < LINEMAXLEN)
  {
  new_line = (char *) malloc(*len + 2);

  if (new_line != NULL)
    {
    strcpy(new_line,line);

    i = 0;
    if (count < 0)
      while (i != count)
        {
        shift_left_str(new_line,*cx + 1,*len + 1);
        i -= 1;
        }
    else if (count > 0)
      while (i != count)
        {
        shift_right_str(new_line,*cx,*len + 1);
        i += 1;
        }

    dest = new_line;
    dest += *cx;
    orig = replace;

    while (*orig != '\0')
      {
      *dest = *orig;
      dest++;
      orig++;
      }
    }
  }

*cx += (int) strlen(replace);
return new_line;
}

/**********************/
void replace(char find_str[],char replace_str[],char suffix_str[])
{
struct line_rec *line_table_oldcur;
struct line_rec *line_table_prev;
char temp_str[82];
int occurances = 0;

char *new_text;
int new_len;
int old_len;
int finished = FALSE;
int replaced = FALSE;
int found_one = FALSE;
int temp_cx;
int temp_sx;
int prev_cx;
int prev_cy;
int old_cx;
int old_cy;
int replace_all = FALSE;
int replace_query = FALSE;
char reply_str[82];
int find_len;
int ok_to_replace = FALSE;

next_mess("");
old_cx = cx;
old_cy = cy;
line_table_oldcur = line_table_cur;

temp_cx = cx;
prev_cx = cx;
prev_cy = cy;
line_table_prev = line_table_cur;

if (  (strcmp(suffix_str,"A") == STR_SAME)
   || (strcmp(suffix_str,"ALL") == STR_SAME)
   )
  replace_all = TRUE;

if (  (strcmp(suffix_str,"Q") == STR_SAME)
   || (strcmp(suffix_str,"QUERY") == STR_SAME)
   )
  replace_query = TRUE;

if (  (replace_all)
   || (replace_query)
   )
  {
  if (block_marked == TRUE)
    line_table_cur = line_table_block_start;
  else
    line_table_cur = line_table_start;

  cx = 0;
  recalc_screen_cy(0);

  line_table_prev = line_table_cur;
  cx = -1;
  if (!find(find_str,FORWARDS,CASE_SENS))
    {
    cx = 0;
    line_table_cur = line_table_prev;
    finished = TRUE;
    }
  else
    if (  (block_marked == TRUE)
       && (line_table_cur->in_block != TRUE)
       )
      finished = TRUE;
    else
      {
      if (replace_query)
        {
        cur_sx = calc_sx_str(line_table_cur->text,temp_cx);

        if (  (cur_sx > (vx + COLS - 2 - (int)strlen(replace_str)))
           || (cur_sx < vx)
           )
          {
          vx = cur_sx - (COLS/4) + 2 + (int)strlen(findstr);
/*        vx = cur_sx - (COLS/4); */
          if (vx < 0) vx = 0;
          }
        use_new_cy = FALSE;
        showscreency(LINES/4);
        }
      }
  }
else
  {
  new_text = (char *) &line_table_cur->text[cx];
  find_len = (int) strlen(find_str);

  if (STR_SAME != strncmp(new_text,find_str,find_len))
    finished = TRUE;
  }

while (!finished)
  {
  temp_cx = cx;
  prev_cx = cx;
  prev_cy = new_cy;

  found_one = TRUE;
  replaced = FALSE;
  ok_to_replace = FALSE;
  if (!replace_query)
    ok_to_replace = TRUE;
  else
    {
    cmove(cy,cur_sx-vx);
/*  cmove(new_cy,temp_cx); */
    enter_inv();
    enter_bold();
    enter_under();
    x_waddstr(find_str,VX_CHECK,cur_sx);
    normal();
    cmove(LINES-1,9);
    clreol();
    x_waddstr("Do you want to replace this string (Y,N or Cancel)?",NO_VX_CHECK,0);
    cmove(LINES-1,61);
    strcpy(temp_str,"");
    get_string(statusptr,temp_str,NO_FUNKEYS,LINES-1,61,NO_SAVE,NO_CTRL,NO_TAB_ENDS_INPUT);
    upper_case(temp_str);
    zap_spaces(temp_str);
    cmove(new_cy,temp_cx);
    if (  (strcmp(temp_str,"YES") == STR_SAME)
       || (strcmp(temp_str,"Y") == STR_SAME)
       )
      ok_to_replace = TRUE;
    else if (  (strcmp(temp_str,"Cancel") == STR_SAME)
            || (strcmp(temp_str,"C") == STR_SAME)
            )
      {
      showline(line_table_cur,line_table_cur->text,new_cy,temp_cx,HILITE);
      finished = TRUE;
      }
    else
      showline(line_table_cur,line_table_cur->text,new_cy,temp_cx,HILITE);
    }
  if (!ok_to_replace)
    {
    temp_cx += (int) strlen(find_str);
    }
  else
    {
    old_len = strlen(line_table_cur->text);
    new_text = replace_string(&new_len,line_table_cur->text,find_str,replace_str,&temp_cx);

    if (new_len >= LINEMAXLEN)
      {
      free(new_text);
      next_mess("Line too long.");
      break;
      }
    else if (new_text == NULL)
      {
      next_mess("Not enough memory available.");
      break;
      }
    else
      {
      if (new_len != old_len)
        {
        if (line_table_cur != line_table_oldcur)
          line_table_cur = replace_line(line_table_cur,new_text);
        else
          {
          line_table_cur = replace_line(line_table_cur,new_text);
          line_table_oldcur = line_table_cur;
          }
        }
      else
        strcpy(line_table_cur->text,new_text);

      free(new_text);
      new_text = NULL;

      if (replace_query)
        showline(line_table_cur,line_table_cur->text,new_cy,temp_cx,HILITE);

      occurances = occurances + 1;
      }
    }
  line_table_prev = line_table_cur;
  if (!finished)
    {
    if (  (replace_all)
       || (replace_query)
       )
      {
      cx = temp_cx - 1;
      if (find(find_str,FORWARDS,CASE_SENS))
        {
        if (  (block_marked == TRUE)
           && (line_table_cur->in_block != TRUE)
           )
          {
          line_table_cur = line_table_prev;
          finished = TRUE;
          }
        else
          {
          temp_cx = cx;
          if (replace_query)
            {
            cur_sx = calc_sx_str(line_table_cur->text,temp_cx);

            if (  (cur_sx > (vx + COLS - 2 - (int)strlen(replace_str)))
               || (cur_sx < vx)
               )
              {
              vx = cur_sx - (COLS/4) + 2 + (int)strlen(findstr);
  /*          vx = cur_sx - (COLS/4); */
              if (vx < 0) vx = 0;
              use_new_cy = FALSE;
              showscreency(LINES/4);
              }
            else
              {
              if (new_cy > (LINES-2))
                showscreency(LINES/4);
              else
                showscreency(new_cy);
              }

/*        cmove(cy,cur_sx-vx); */
/*        showscreency(0); */
            }
          }
        }
      else
        {
        line_table_cur = line_table_prev;
        finished = TRUE;
        }
      }
    else
      {
      line_table_cur = line_table_prev;
      finished = TRUE;
      }
    }

  }

if (!found_one)
  {
  next_mess("String not found.");
  line_table_cur = line_table_oldcur;
  cx = old_cx;
  cy = old_cy;
  recalc_screen_cy(cy);
  }
else
  {
  file_needs_saving = TRUE;
  if (mess_text[0] == '\0')
    {
    sprintf(temp_str,"Replaced %d occurances of string.",occurances);
    next_mess(temp_str);
    }

  if (replace_all)
    {
    line_table_cur = line_table_oldcur;
    cx = old_cx;
    cy = old_cy;
    use_new_cy = FALSE;
    showscreency(cy);
    }
  else
    if (!replace_query)
      {
      cx = old_cx;
      showline(line_table_cur,line_table_cur->text,cy,cx,HILITE);
      }
    else
      {
      cx = prev_cx;
      cy = prev_cy;
      line_table_cur = line_table_prev;
      }
  }

use_new_cy = FALSE;
}

/**********************/
void deleteblock(int from_moveblock)
{
struct line_rec *line_table_ptr;
struct line_rec *table_next;
struct line_rec *table_prev;
struct line_rec *line_table_before;
struct line_rec *line_table_after;
int c;


if (block_marked == FALSE)
  next_mess("No block marked.");
else
  {
  highlight_block(FALSE);
  copyblock(FROM_DELETEBLOCK);
  table_prev = NULL;
  line_table_before = line_table_block_start->prev_line;
  line_table_ptr = line_table_block_start;

  while (table_prev != line_table_block_end)
    {
    if (line_table_top == line_table_ptr)
      line_table_top = NULL;
    if (line_table_start == line_table_ptr)
      line_table_start = NULL;
    if (line_table_end == line_table_ptr)
      line_table_end = NULL;
    table_prev = line_table_ptr;
    table_next = line_table_ptr->next_line;

    if (  (line_table_before == NULL)
       && (table_next == NULL)
       )
           /* we are about to delete the last line of the file
              but we don't want to do that, just NULL the text */
      {
      strcpy(line_table_ptr->text,"");
      line_table_ptr->in_block = FALSE;
      if (! file_empty)
        {
        file_empty = TRUE;
        restore_status(ALL_STATUS,LINES-1);
        }
      line_table_ptr->next_line = NULL;
      line_table_ptr->prev_line = NULL;
      line_table_after = NULL;
      }
    else
      {
      free(line_table_ptr);
      line_table_ptr = table_next;
      line_table_after = line_table_ptr;
      }
    }

  if (  (line_table_before != NULL)
     && (line_table_after != NULL)
     )
    {
    line_table_before->next_line = line_table_after;
    line_table_after->prev_line = line_table_before;
    line_table_cur = line_table_before;
    }
  else if (  (line_table_before == NULL)
          && (line_table_after != NULL)
          )
    {
    line_table_cur = line_table_after;
    line_table_after->prev_line = NULL;
    }
  else if (  (line_table_before != NULL)
          && (line_table_after == NULL)
          )
    {
    line_table_cur = line_table_before;
    line_table_before->next_line = NULL;
    }
  else if (  (line_table_before == NULL)
          && (line_table_after == NULL)
          )
    {
    line_table_start = line_table_end = line_table_cur = line_table_ptr;
    line_table_top = line_table_ptr;
    }

  if (line_table_start == NULL) line_table_start = line_table_cur;
  if (line_table_end   == NULL) line_table_end   = line_table_cur;
  if (line_table_top   == NULL) line_table_top   = line_table_cur;

  line_table_block_end = NULL;
  line_table_block_start = NULL;
  block_marked = FALSE;

  if (!from_moveblock)
    {
    showscreency((LINES-1)/2);
    next_mess("Block deleted.");
    }

  file_needs_saving = TRUE;
  }
}

/**********************/
void pasteblock(int from_moveblock)
{
struct line_rec *line_table_ptr;
struct line_rec *buff_table_ptr;
struct line_rec *line_table_oldprev;
struct line_rec *line_table_newcur;
struct line_rec *table_next;
struct line_rec *table_prev;

if (buff_table_start == NULL)
  next_mess("Buffer empty.");
else
  {
  line_table_oldprev = line_table_cur->prev_line;
  buff_table_ptr = buff_table_start;

  line_table_ptr =
    (struct line_rec *)
    malloc(sizeof (struct line_rec) + strlen(buff_table_ptr->text));

  if (line_table_ptr != NULL)
    {
    line_table_ptr->prev_line = line_table_oldprev;
    if (line_table_oldprev != NULL)
      line_table_oldprev->next_line = line_table_ptr;
    else
      line_table_start = (line_table_top = line_table_ptr);
    }
  else
    {
    next_mess("Insufficient memory.");
    goto exit;
    }
  while (buff_table_ptr != NULL)
    {
    if (line_table_ptr == NULL)
      {
      next_mess("Insufficient memory.");
      goto cleanup;
      }
    strcpy(line_table_ptr->text,buff_table_ptr->text);
    line_table_ptr->in_block = FALSE;

    table_next = buff_table_ptr->next_line;
    buff_table_ptr = table_next;

    if (buff_table_ptr != NULL)
      table_next =
        (struct line_rec *)
        malloc(sizeof (struct line_rec) + strlen(buff_table_ptr->text));
    else
      table_next =
        (struct line_rec *)
        malloc(sizeof (struct line_rec));

    if (table_next == NULL)
      {
      next_mess("Insufficient memory.");
      goto cleanup;
      }
    table_prev = line_table_ptr;
    line_table_ptr->next_line = table_next;
    line_table_ptr = table_next;
    line_table_ptr->prev_line = table_prev;
    }
  free(line_table_ptr);

cleanup:
  line_table_ptr = table_prev;
  if (file_empty)
    {
    free(line_table_cur);
    line_table_cur = line_table_start;
    line_table_ptr->next_line = NULL;
    line_table_bot = line_table_end = line_table_ptr;
    }
  else
    {
    line_table_ptr->next_line = line_table_cur;
    line_table_cur->prev_line = line_table_ptr;
    }
exit:
  file_needs_saving = TRUE;
  if (!from_moveblock)
    {
    restore_status(ALL_STATUS,LINES-1);
    showscreen_cur();
    next_mess("Block pasted.");
    }
  }
}

/**********************/
void copyblock(from_deleteblock)
{
struct line_rec *line_table_ptr;
struct line_rec *buff_table_ptr;
struct line_rec *table_next;
struct line_rec *buff_table_prev;
struct line_rec *line_table_prev;

if (block_marked == FALSE)
  next_mess("No block marked.");
else
  {
  empty_buffer();
  showblock(NO_HILITE);
  highlight_block(FALSE);
  line_table_prev = NULL;
  line_table_ptr = line_table_block_start;

  buff_table_start =
      (struct line_rec *)
      malloc(sizeof (struct line_rec) + strlen(line_table_ptr->text));

  buff_table_ptr = buff_table_start;
  if (buff_table_ptr != NULL)
    buff_table_ptr->prev_line = NULL;

  while (line_table_prev != line_table_block_end)
    {
    if (buff_table_ptr == NULL)
      {
      next_mess("Insufficient memory.");
      goto exit;
      }
    strcpy(buff_table_ptr->text,line_table_ptr->text);
    line_table_ptr->in_block = FALSE;
    line_table_prev = line_table_ptr;
    table_next = line_table_ptr->next_line;
    line_table_ptr = table_next;

    if (line_table_ptr != NULL)
      table_next =
        (struct line_rec *)
        malloc(sizeof (struct line_rec) + strlen(line_table_ptr->text));
    else
      table_next =
        (struct line_rec *)
        malloc(sizeof (struct line_rec));

    buff_table_prev = buff_table_ptr;
    buff_table_ptr->next_line = table_next;
    buff_table_ptr = table_next;
    buff_table_ptr->prev_line = buff_table_prev;
    }
  free(buff_table_ptr);

exit:
  buff_table_ptr = buff_table_prev;
  buff_table_ptr->next_line = NULL;

  if (!from_deleteblock)
    {
    block_marked = FALSE;
    line_table_block_start = line_table_block_end = NULL;
    next_mess("Block copied.");
    restore_status(ALL_STATUS,LINES-1);
    showscreen();
    }
  }
}

/***********************/
void empty_buffer()
{
struct line_rec *buff_table_ptr;
struct line_rec *table_next;

buff_table_ptr = buff_table_start;
while (buff_table_ptr != NULL)
  {
  table_next = buff_table_ptr->next_line;
  free(buff_table_ptr);
  buff_table_ptr = table_next;
  }
buff_table_end = buff_table_start = NULL;
}

/**********************/
void markblock()
{
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;

if (file_empty)
  next_mess("No text to mark.");
else
if (line_table_block_start == NULL)
/* there are no block markers */
  {
  line_table_block_start = line_table_cur;
  next_mess("One block marker placed.");
  }
else if (line_table_block_end == NULL)
/* there is a block start but no block end */
  {
  if (line_table_cur == line_table_block_start)
  /* there is a block start, block end will be on same line (1 line block) */
    {
    line_table_block_end = line_table_cur;
    block_marked = TRUE;
    highlight_block(TRUE);
    showblock(HILITE);
    next_mess("Block marked.");
    }
  else
  /* There is already a block start & the block end will not be on same line.
     So it is necessary to check that the 2nd marker (block end) is below
     the first (block start). If it is not, then they must be swapped around.
  */
    {
    line_table_ptr = line_table_end;
    while (  (line_table_ptr != line_table_block_start)
          && (line_table_ptr != line_table_cur)
          && (line_table_ptr != NULL)
          )
      {
      line_table_next = line_table_ptr->prev_line;
      line_table_ptr = line_table_next;
      }
    if (line_table_ptr == line_table_cur)
      line_table_block_end = line_table_cur;
    else if (line_table_ptr == line_table_block_start)
      {
      line_table_block_end = line_table_block_start;
      line_table_block_start = line_table_cur;
      }
    block_marked = TRUE;
    highlight_block(TRUE);
    showblock(HILITE);
    next_mess("Block marked.");
    }
  }
else
/* There is already a block start & block end, so NULL them to create
   a new block, with a block start only.
*/
  {
  showblock(NO_HILITE);
  highlight_block(FALSE);
  line_table_block_end = NULL;
  line_table_block_start = line_table_cur;
  block_marked = FALSE;
  next_mess("One block marker placed.");
  }
}

/**********************/
void highlight_block(int on_off)
{
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;
struct line_rec *line_table_prev = NULL;

if (  (block_marked == TRUE)
   && (line_table_block_start != NULL)
   && (line_table_block_end != NULL)
   )
  {
  line_table_ptr = line_table_block_start;
  while (  (line_table_prev != line_table_block_end)
        && (line_table_ptr != NULL)
        )
    {
    line_table_prev = line_table_ptr;
    line_table_ptr->in_block = on_off;
    line_table_next = line_table_ptr->next_line;
    line_table_ptr = line_table_next;
    }
  }
}

/**************************************/
int in_str(char line[],char *pattern,int *cx,int direction)
{
int len;
int found = 1;
int finished = FALSE;
char * start;

len = strlen(pattern);

start = (char *) &(line[*cx]);
if (*start == '\0')
  finished = TRUE;
else
  if (STR_SAME == (found = strncmp(start,pattern,len)))
    finished = TRUE;

while (!finished)
  {
  if (direction == FORWARDS)
    *cx += 1;
  else
    *cx -= 1;

  if (*cx >= 0)
    {
    start = (char *) &(line[*cx]);
    if (*start == '\0')
      finished = TRUE;
    }
  else
    finished = TRUE;

  if (!finished)
    if (STR_SAME == (found = strncmp(start,pattern,len)))
      finished = TRUE;
  }

if (found != 0)
  return FALSE;
else
  return TRUE;
}

/**********************/
int find(char str[],int direction,int case_sens)
{
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;
char temp_line[LINEMAXLEN];
char str_to_find[LINEMAXLEN];
int temp_cx;
int found = FALSE;
int finished = FALSE;
int end_of_line = FALSE;
char * temp_cx_ptr;
int i;

new_cy = cy;

if (strcmp(str,"") != STR_SAME)
  {
  line_table_ptr = line_table_cur;

  if (direction == FORWARDS)
    temp_cx = cx + 1;
  else
    temp_cx = cx - 1;

  if (temp_cx >= 0)
    {
    if (temp_cx >= (int) strlen(line_table_ptr->text))
      end_of_line = TRUE;
    }
  else
    end_of_line = TRUE;

  if (end_of_line)
    {
    if (direction == FORWARDS)
      {
      line_table_ptr = line_table_ptr->next_line;
      temp_cx = 0;
      new_cy += 1;
      }
    else
      {
      line_table_ptr = line_table_ptr->prev_line;
      if (line_table_ptr != NULL)
        temp_cx = strlen(line_table_ptr->text) - 1;
      new_cy -= 1;
      }
    }

  if (line_table_ptr == NULL)
    return FALSE;
  else
    {
    strcpy(temp_line,line_table_ptr->text);
    strcpy(str_to_find,str);
    if (! case_sens)
      {
      upper_case(temp_line);
      upper_case(str_to_find);
      }

    finished = (found = in_str(temp_line,str_to_find,&temp_cx,direction));

    while (!finished)
      {
      if (direction == FORWARDS)
        {
        line_table_next = line_table_ptr->next_line;
        temp_cx = 0;
        new_cy += 1;
        }
      else
        {
        line_table_next = line_table_ptr->prev_line;
        if (line_table_next != NULL)
          temp_cx = strlen(line_table_next->text) - 1;
        new_cy -= 1;
        }
      line_table_ptr = line_table_next;

      if (line_table_ptr != NULL)
        {
        if (! case_sens)
          {
          strcpy(temp_line,line_table_ptr->text);
          upper_case(temp_line);
          }
        else
          strcpy(temp_line,line_table_ptr->text);

        finished = (found = in_str(temp_line,str_to_find,&temp_cx,direction));
        }
      else
        finished = TRUE;
      }

    if (!found)
      {
      use_new_cy = FALSE;
      return FALSE;
      }
    else
      {
      line_table_cur = line_table_ptr;
      cx = temp_cx;
      use_new_cy = TRUE;
      return TRUE;
      }
    }
  }
else
  {
  use_new_cy = FALSE;
  return FALSE;
  }
}

/**********************/
char * upper_case(char *source)
{
char * orig;
orig = source;
while (*source != '\0')
  {
  *source = (char) toupper(*source);
  source++;
  }
return orig;
}

/**********************/
int gotoline(int line)
{
int i = 1;
struct line_rec *line_table_next;
struct line_rec *line_table_ptr;
struct line_rec *line_table_save = line_table_top;

line_table_ptr = line_table_start;
while (  (i < line)
      && (line_table_ptr->next_line != NULL)
      )
  {
  line_table_next = line_table_ptr->next_line;
  line_table_ptr = line_table_next;
  i = i + 1;
  }
if (i == line)
  {
  line_table_top = line_table_cur = line_table_ptr;
  showscreency((LINES-1)/2);
  return TRUE;
  }
else
  {
  next_mess("Line not found.");
  return TRUE;
  }
}

/**************************/
void split_line()
{
struct line_rec *line_table_new;

char prev_text[LINEMAXLEN];
char cur_text[LINEMAXLEN];

int i,j;

strcpy(prev_text,line_table_cur->text);
strcpy(cur_text,line_table_cur->text);

i = cx;
/*
while (  (prev_text[i-1] == ' ')
      && (i > 0)
      )
  i -= 1;
*/
prev_text[i] = '\0';

insert_line_text(line_table_cur,prev_text);

i = cx;
if (i >= ((int)strlen(cur_text)))
  i = strlen(cur_text);
j = 0;
do
  {
  cur_text[j] = cur_text[i];
  i += 1;
  j += 1;
  } while (cur_text[j-1] != '\0');

line_table_cur = replace_line(line_table_cur,cur_text);
}

/**************************/
void delete_line()
{
struct line_rec *line_table_prev;
struct line_rec *line_table_temp;

line_table_prev = line_table_cur->prev_line;
line_table_next = line_table_cur->next_line;
line_table_temp = line_table_cur;

if (  (line_table_cur == line_table_block_start)
   || (line_table_cur == line_table_block_end)
   )
  {
  showblock(NO_HILITE);
  highlight_block(FALSE);
  line_table_block_start = line_table_block_end = NULL;
  block_marked = FALSE;
  next_mess(" ");
  }

if (  (line_table_next != NULL)
   && (line_table_prev != NULL)
   )
  {
  line_table_next->prev_line = line_table_prev;
  line_table_prev->next_line = line_table_next;
  line_table_cur = line_table_next;
  if (line_table_top == line_table_temp) line_table_top = line_table_cur;
  deline();
  if (line_table_bot->next_line != NULL)
    {
    line_table_bot = line_table_bot->next_line;
    show_bot_line();
/*
    cur_sx = sx = cx = vx = 0;
*/
    cur_sx = calc_sx_str(line_table_cur->text,cx);
    cmove(cy,cur_sx-vx);
    restore_status(ALL_STATUS,LINES-1);
    }
  else
    {
    cmove(LINES-2,0);
    insline();
/*
    cur_sx = sx = cx = vx = 0;
    cmove(cy,0);
*/
    cur_sx = calc_sx_str(line_table_cur->text,cx);
    cmove(cy,cur_sx-vx);
    }
  free(line_table_temp);
  file_needs_saving = TRUE;
  }
else
  if (  (line_table_next == NULL)
     && (line_table_prev != NULL)
     )
    {
    line_table_prev->next_line = NULL;
    line_table_end = line_table_prev;
    line_table_cur = line_table_prev;
    cmove(cy,0);
    clreol();
    cy = cy - 1;
/*
    cx = 0;
    cmove(cy,0);
*/
    cur_sx = calc_sx_str(line_table_cur->text,cx);
    cmove(cy,cur_sx-vx);
    if (line_table_top == line_table_temp) line_table_top = line_table_cur;
    free(line_table_temp);
    file_needs_saving = TRUE;
    }
  else
    if (  (line_table_next != NULL)
       && (line_table_prev == NULL)
       )
      {
      line_table_next->prev_line = NULL;
      line_table_start = line_table_next;
      line_table_cur = line_table_next;
      if (line_table_top == line_table_temp) line_table_top = line_table_cur;
      deline();
      if (line_table_bot->next_line != NULL)
        {
        line_table_bot = line_table_bot->next_line;
        show_bot_line();
/*      cur_sx = sx = cx = vx = 0; */
        cur_sx = calc_sx_str(line_table_cur->text,cx);
        cmove(cy,cur_sx-vx);
        restore_status(ALL_STATUS,LINES-1);
        }
      else
        {
        cmove(LINES-2,0);
        insline();
/*
        cur_sx = sx = cx = vx = 0;
        cmove(cy,0);
*/
        cur_sx = calc_sx_str(line_table_cur->text,cx);
        cmove(cy,cur_sx-vx);
        }
      free(line_table_temp);
      file_needs_saving = TRUE;
      }
    else
      if (  (line_table_next == NULL)
         && (line_table_prev == NULL)
         )
      /* we are about to delete the last line in the file, but don't do that.
         Just null the text & set the file_empty flag
      */
        {
        line_table_top = line_table_cur;
        line_table_start = line_table_cur;
        line_table_end = line_table_cur;
        line_table_bot = line_table_cur;
        cmove(cy,0);
        clreol();
        cy = 0;
        cx = 0;
        cur_sx = calc_sx_str(line_table_cur->text,cx);
        cmove(cy,cur_sx-vx);
        strcpy(line_table_cur->text,"");
        line_table_cur->in_block = FALSE;
        file_needs_saving = TRUE;

        if (! file_empty)
          {
          file_empty = TRUE;
          restore_status(ALL_STATUS,LINES-1);
          }
        }
}

/**************************/
void insert_line()
{
struct line_rec *line_table_new;
struct line_rec *line_table_prev;
struct line_rec *line_table_temp;

if (file_empty == TRUE)
  {
  strcpy(line_table_cur->text,"");
  file_empty = FALSE;
  restore_status(ALL_STATUS,LINES-1);
  }
if (file_empty != TRUE)
  {
  line_table_new = (struct line_rec *) malloc(sizeof (struct line_rec));
  if (line_table_new == NULL)
    {
    next_mess("Insufficient memory.");
    goto exit;
    }
  strcpy(line_table_new->text,"");
  line_table_prev = line_table_cur->prev_line;
  if (line_table_prev != NULL)
    line_table_prev->next_line = line_table_new;
  line_table_cur->prev_line = line_table_new;
  line_table_new->next_line = line_table_cur;
  line_table_new->prev_line = line_table_prev;
  if (line_table_new->prev_line == NULL)
    line_table_start = line_table_new;
  if (line_table_top == line_table_cur)
    line_table_top = line_table_new;

  if (line_table_prev != NULL)
    {
    if (  (line_table_prev->in_block == TRUE)
       && (line_table_cur->in_block == TRUE)
       )
      line_table_new->in_block = TRUE;
    else
      line_table_new->in_block = FALSE;
    }
  else
    line_table_new->in_block = FALSE;

  file_needs_saving = TRUE;
  }
exit:
  ;
}

/**********************/
void insert_line_end()
{
struct line_rec *line_table_new;

if (file_empty == TRUE)
  {
  strcpy(line_table_cur->text,"");
  file_empty = FALSE;
  restore_status(ALL_STATUS,LINES-1);
  }
if (file_empty != TRUE)
  {
  line_table_new = (struct line_rec *) malloc(sizeof(struct line_rec));
  if (line_table_new == NULL)
    {
    next_mess("Insufficient memory.");
    goto exit;
    }
  strcpy(line_table_new->text,"");
  line_table_new->in_block = FALSE;
  line_table_end->next_line = line_table_new;
  line_table_new->prev_line = line_table_end;
  line_table_new->next_line = NULL;
  line_table_end = line_table_new;
  line_table_cur = line_table_new;
  file_needs_saving = TRUE;
  cx = 0;
/* vx = 0; */
  }
exit:
  ;
}


/**********************/
struct line_rec * join_ptr(struct line_rec *line_table_ptr)

/*-------------
join the line_table_ptr line to the previous line, if it exists,
and return a pointer to the new line.
-------------*/

{
struct line_rec *line_table_prev;
struct line_rec *line_table_temp;

int total_len,prev_len,cur_len;
char prev_text[LINEMAXLEN];
char cat_text[LINEMAXLEN];
char dummy;

line_table_prev = line_table_ptr->prev_line;

if (line_table_prev != NULL)
  {
  strcpy(prev_text,line_table_prev->text);
  strcpy(cat_text,line_table_ptr->text);

  prev_len = strlen(prev_text);

  cur_len = strlen(cat_text);
  total_len = prev_len + cur_len;

  if (total_len > (LINEMAXLEN-1))
    {
    next_mess("Line too long.");
    return line_table_ptr;
    }
  else
    {
    if (  (block_marked == TRUE)
       && (line_table_prev->in_block != line_table_ptr->in_block)
       )
      {
      showblock(NO_HILITE);
      highlight_block(FALSE);
      line_table_block_start = line_table_block_end = NULL;
      block_marked = FALSE;
      next_mess(find_text);
      }

    strcat(prev_text,cat_text);

    line_table_prev = replace_line(line_table_prev,prev_text);

    line_table_prev->next_line = line_table_ptr->next_line;
    if (line_table_prev->next_line == NULL)
      {
      line_table_bot = line_table_prev;
      line_table_end = line_table_prev;
      }
    else
      {
      line_table_temp = line_table_prev->next_line;
      line_table_temp->prev_line = line_table_prev;
      }

    line_table_prev->in_block = line_table_ptr->in_block;

    if (block_marked == TRUE)
      {
      if (line_table_ptr == line_table_block_start)
        line_table_block_start = line_table_prev;

      if (line_table_ptr == line_table_block_end)
        line_table_block_end = line_table_prev;
      }

    free(line_table_ptr);
    return line_table_prev;
    }
  }
else
  return line_table_ptr;
}

/*************************/
void save(int save_block)
{
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;
struct stat fstat_buffer;
char temp_str[LINEMAXLEN];
int read_only = FALSE;
int ok_to_write = FALSE;
mode_t mode;

save_abandoned = FALSE;
save_failed = FALSE;
message_pending = FALSE;

if (stat(filename,&fstat_buffer) == -1)
  {
  if (errno == ENOENT)
    ok_to_write = TRUE;
  else
    ok_to_write = FALSE;
  }
else
  {
  mode = fstat_buffer.st_mode;
  if ((mode & S_IWUSR) == S_IWUSR)
    ok_to_write = TRUE;
  else
    {
    cmove(LINES-1,9);
    clreol();
    x_waddstr("File is read-only. Do you want to overwrite it(Y or N)? ",NO_VX_CHECK,0);
    cmove(LINES-1,65);
    strcpy(temp_str,"");
    get_string(statusptr,temp_str,NO_FUNKEYS,LINES-1,65,NO_SAVE,NO_CTRL,NO_TAB_ENDS_INPUT);
    upper_case(temp_str);
    zap_spaces(temp_str);
    if (  (strcmp(temp_str,"YES") == STR_SAME)
       || (strcmp(temp_str,"Y") == STR_SAME)
       )
      {
      chmod(filename,(mode | S_IWUSR));
      stat(filename,&fstat_buffer);
      if ((fstat_buffer.st_mode & S_IWUSR) == S_IWUSR)
          /* set the user write permissions bit */
          /* for a discussion of this operation see the readfile() subroutine */
        {
        ok_to_write = TRUE;
        read_only = TRUE;
        }
      else
        {
        ok_to_write = FALSE;
        cmove(LINES-1,9);
        clreol();
        sprintf(errortext,"Cannot get write permission on file '%s'",filename);
        next_mess(errortext);
        save_abandoned = TRUE;
        }
      }
    else
      {
      save_abandoned = TRUE;
      }
    }
  }

if (ok_to_write)
    {
    fptr = fopen(tempfile,"w");
    if (fptr == NULL)
      {
      sprintf(errortext,"%s on file '%s'",strerror(errno),tempfile);
      next_mess(errortext);
      save_failed = TRUE;
      }
    else
      {
      if (save_block)
        line_table_ptr = line_table_block_start;
      else
        line_table_ptr = line_table_start;
      while (  (line_table_ptr != NULL)
            && (! save_failed)
            )
        {
        if (fputs(line_table_ptr->text,fptr) == EOF)
          {
          sprintf(errortext,"%s on file '%s'",strerror(errno),tempfile);
          next_mess(errortext);
          save_failed = TRUE;
          break;
          }
        if (!file_empty)
          if (fputc('\n',fptr) == EOF)
            {
            sprintf(errortext,"%s on file '%s'",strerror(errno),tempfile);
            next_mess(errortext);
            save_failed = TRUE;
            break;
            }
        if (  (save_block)
           && (line_table_ptr == line_table_block_end)
           )
          line_table_ptr = NULL;
        else
          {
          line_table_next = line_table_ptr->next_line;
          line_table_ptr = line_table_next;
          }
        }
      fclose(fptr);
      if (rename_file(tempfile,filename) != UNIX_OK)
        {
        sprintf(errortext,"%s on file '%s'",strerror(errno),filename);
        next_mess(errortext);
        save_failed = TRUE;
        }
      else
        {
        if (save_block)
          {
          showblock(NO_HILITE);
          highlight_block(FALSE);
          block_marked = FALSE;
          line_table_block_start = line_table_block_end = NULL;
          /* showscreen(); */
          }
        else
          {
          file_needs_saving = FALSE;
          if (read_only)
            /* unset the user write permissions bit */
            /* for a discussion of this see the readfile() subroutine */
            chmod(filename,(mode & (~S_IWUSR)));
          }
        }
      }
    }
}

/*************************/
struct line_rec * replace_line(struct line_rec *line_table_ptr,char line_buffer[])
/*----------
replace the line_table_ptr text with line_buffer.
Note that this is now more compilacted with variable text fields.
------------
*/
{
struct line_rec *line_table_new;
struct line_rec *line_table_temp;
int x,c;

line_table_new =
  (struct line_rec *)
     malloc(sizeof (struct line_rec) + strlen(line_buffer));

if (line_table_new == NULL)
  {
  next_mess("Insufficient memory.");
  return line_table_ptr;
  }

strcpy(line_table_new->text,line_buffer);
line_table_new->in_block = line_table_ptr->in_block;

line_table_temp = line_table_ptr->prev_line;
if (line_table_temp != NULL)
  line_table_temp->next_line = line_table_new;
line_table_new->prev_line = line_table_temp;

line_table_temp = line_table_ptr->next_line;
if (line_table_temp != NULL)
  line_table_temp->prev_line = line_table_new;
line_table_new->next_line = line_table_temp;

if (line_table_cur == line_table_ptr) line_table_cur = line_table_new;
if (line_table_start == line_table_ptr) line_table_start = line_table_new;
if (line_table_end == line_table_ptr) line_table_end = line_table_new;
if (line_table_top == line_table_ptr) line_table_top = line_table_new;
if (line_table_bot == line_table_ptr) line_table_bot = line_table_new;
if (line_table_block_start == line_table_ptr)
  line_table_block_start = line_table_new;

if (line_table_block_end == line_table_ptr)  
  line_table_block_end = line_table_new;

free(line_table_ptr);
file_needs_saving = TRUE;
return line_table_new;
}

/**************************/
void insert_line_text(struct line_rec *line_table_ptr,char line_buffer[])
/*----------
insert a new line of text above the line_table_ptr line,
and containing the text in line_buffer.
------------
*/
{
  struct line_rec *line_table_new;
  struct line_rec *line_table_prev;
  int i = 0;

  if (file_empty == TRUE)
    {
    strcpy(line_table_cur->text,"");
    file_empty = FALSE;
    restore_status(ALL_STATUS,LINES-1);
    }

  line_table_new =
    (struct line_rec *) malloc(sizeof (struct line_rec) + strlen(line_buffer));
  if (line_table_new == NULL)
    {
    next_mess("Insufficient memory.");
    goto exit;
    }
  strcpy(line_table_new->text,line_buffer);
  line_table_prev = line_table_ptr->prev_line;
  if (line_table_prev != NULL)
    line_table_prev->next_line = line_table_new;
  line_table_ptr->prev_line = line_table_new;
  line_table_new->next_line = line_table_ptr;
  line_table_new->prev_line = line_table_prev;
  if (line_table_new->prev_line == NULL)
    line_table_start = line_table_new;
  if (line_table_top == line_table_ptr)
    line_table_top = line_table_new;

  if (line_table_prev != NULL)
    {
    if (  (line_table_prev->in_block == TRUE)
       && (line_table_ptr->in_block == TRUE)
       )
      line_table_new->in_block = TRUE;
    else
      line_table_new->in_block = FALSE;
    }
  else
    line_table_new->in_block = FALSE;
  file_needs_saving = TRUE;

exit:
  ;
}

/*************************/
void updateline(int leave_trailing_spaces)
{
char * buffer_ptr;

/* never truncate trailing spaces
if (  (line_buffer[0] != '\0')
   && (! leave_trailing_spaces)
   )
  {
  buffer_ptr = (char *) &line_buffer;
  while (*buffer_ptr != '\0')
    buffer_ptr++;
  buffer_ptr--;
  while (  (*buffer_ptr == ' ')
        || (*buffer_ptr == TAB)
        )
    buffer_ptr--;
  buffer_ptr++;
  *buffer_ptr = '\0';
  }
*/

line_table_cur = replace_line(line_table_cur,line_buffer);

if (  (strcmp(line_table_cur->text,"") == STR_SAME)
   && (line_table_cur->prev_line == NULL)
   && (line_table_cur->next_line == NULL)
   )
  {
/*  strcpy(line_table_cur->text,""); */
  if (! file_empty)
    {
    file_empty = TRUE;
    restore_status(ALL_STATUS,LINES-1);
    }
  }
else
  if (file_empty)
    {
    file_empty = FALSE;
    restore_status(ALL_STATUS,LINES-1);
    }

line_needs_updating = FALSE;
}

/*************************/
void left()
{
if (cx > (int) strlen(line_buffer))
  cx = (int) strlen(line_buffer);

if (cx > 0)
  cx = cx - 1;

cur_sx = calc_sx_str(line_buffer,cx);

if (cur_sx < vx)
  {
  if (line_needs_updating)
    updateline(LEAVE_TRAILING_SPACES);
  vx = cur_sx - (COLS/3);
/*vx = cur_sx - 19; */
  if (vx < 0) vx = 0;
  showscreen();
  load_buffer();
  }
else
  cmove(cy,cur_sx-vx);
}

/*************************/
void right()
{
if (cx > (int) strlen(line_buffer))
  cx = (int) strlen(line_buffer);

if (cx < (int) strlen(line_buffer))
  cx = cx + 1;

cur_sx = calc_sx_str(line_buffer,cx);

if (cur_sx > (vx + COLS - 2))
  {
  if (line_needs_updating)
    updateline(LEAVE_TRAILING_SPACES);
  vx = cur_sx - COLS + (COLS/3);
/*vx = cur_sx - COLS + 21; */
  showscreen();
  load_buffer();
  }
else
  cmove(cy,cur_sx-vx);

}

/*************************/
void up()
{
struct line_rec *line_table_next;

if (cy > 0)
  {
  line_table_cur = line_table_cur->prev_line;
  cy = cy - 1;
  cur_sx = calc_sx_str(line_table_cur->text,cx);
  cmove(cy,cur_sx-vx);
  }
else
  if (line_table_top->prev_line != NULL)
    {
    if (!scrolling)
      {
      scrolling = TRUE;
      scrollingup = TRUE;
      scrollingdown = FALSE;
      new_term.c_cc[4] = 0;
      new_term.c_cc[5] = 3;
      if (ioctl(0,TCSETAF,(struct termio *) &new_term))
        next_mess("Unable to setup terminal: ioctl() failed.");
      }
    line_table_top = line_table_top->prev_line;
    line_table_cur = line_table_top;
    recalc_screen_cur();

    cmove(LINES-1,0);
    clreol();
    cmove(0,0);
    insline();
    showline(line_table_cur,line_table_cur->text,0,0,HILITE);
    cur_sx = calc_sx_str(line_table_cur->text,cx);
    restore_status(ALL_STATUS,LINES-2);
    }
}

/*************************/
void up_no_upd()
{
struct line_rec *line_table_next;

if (line_table_cur != line_table_top)
  line_table_cur = line_table_cur->prev_line;
else
  {
  if (line_table_top->prev_line != NULL)
    {
    line_table_top = line_table_top->prev_line;
    line_table_bot = line_table_bot->prev_line;
    line_table_cur = line_table_top;
    }
  }
}

/*************************/
void down()
{
struct line_rec *line_table_next;

if (line_table_cur->next_line != NULL)
  if (cy < (LINES-2))
    {
    line_table_next = line_table_cur->next_line;
    line_table_cur = line_table_next;
    cy = cy + 1;
    cur_sx = calc_sx_str(line_table_cur->text,cx);
    cmove(cy,cur_sx-vx);
    }
  else
    {
    if (line_table_bot->next_line != NULL)
      {
      if (!scrolling)
        {
        scrolling = TRUE;
        scrollingup = FALSE;
        scrollingdown = TRUE;
        new_term.c_cc[4] = 0;
        new_term.c_cc[5] = 3;
        if (ioctl(0,TCSETAF,(struct termio *) &new_term))
          next_mess("Unable to setup terminal: ioctl() failed.");
        }
      line_table_bot = line_table_bot->next_line;
      line_table_top = line_table_top->next_line;
      line_table_cur = line_table_bot;

      cmove(0,0);
      deline();
      showline(line_table_cur,line_table_cur->text,LINES-2,0,HILITE);
      cur_sx = calc_sx_str(line_table_cur->text,cx);
      restore_status(ALL_STATUS,LINES-1);
      }
    }
}

/************************************************/
void down_no_upd()
{
struct line_rec *line_table_next;

if (line_table_cur->next_line != NULL)
  {
  if (line_table_cur != line_table_bot)
    line_table_cur = line_table_cur->next_line;
  else
    if (line_table_bot->next_line != NULL)
      {
      line_table_bot = line_table_bot->next_line;
      line_table_top = line_table_top->next_line;
      line_table_cur = line_table_bot;
      }
  }
}

/*************************/
void pagedown(int y)
{
int linecount;

linecount = 0;

while (  (line_table_bot->next_line != NULL)
      && (linecount < y)
      )
  {
  line_table_next = line_table_bot->next_line;
  line_table_bot = line_table_next;
  line_table_next = line_table_top->next_line;
  line_table_top = line_table_next;
  line_table_next = line_table_cur->next_line;
  line_table_cur = line_table_next;
  linecount = linecount + 1;
  }
}

/*************************/
void pageup(int y)
{
int linecount;

linecount = 0;

while (  (line_table_top->prev_line != NULL)
      && (linecount < y)
      )
  {
  line_table_next = line_table_bot->prev_line;
  line_table_bot = line_table_next;
  line_table_next = line_table_top->prev_line;
  line_table_top = line_table_next;
  line_table_next = line_table_cur->prev_line;
  line_table_cur = line_table_next;
  linecount = linecount + 1;
  }
}

/*************************/
void showline(struct line_rec * line_table_ptr,
              char * text,
              int ly,
              int cx,
              int hilite_block)
{
  cmove(ly,0);
  sx = 0;
  if (  (line_table_ptr->in_block == TRUE)
     && (hilite_block)
     )
    {
    enter_inv();
    win_reverse = A_REVERSE;
    }
  else
    {
    normal();
    win_reverse = 0;
    }

  if (*text == '\0')
    x_waddstr(" ",VX_CHECK,0);
  else
    x_waddstr(text,VX_CHECK,0);

  clreol();
}

/*************************/
void showline_right(struct line_rec * line_table_ptr,
              char * text,
              int ly,
              int cx,
              int hilite_block)
{
  cmove(ly,cur_sx-vx);
  if (  (line_table_ptr->in_block == TRUE)
     && (hilite_block)
     )
    {
    enter_inv();
    win_reverse = A_REVERSE;
    }
  else
    {
    normal();
    win_reverse = 0;
    }

  if (*text == '\0')
    n_waddstr(" ",VX_CHECK,cx);
  else
    n_waddstr(text,VX_CHECK,cx);

  clreol();
}

/*************************/
void show_bot_line()
{
int linecount = 0;
int newcy = 0;
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;

line_table_ptr = line_table_top;

while (  (line_table_ptr != NULL)
      && (linecount <= LINES-2)
      )
  {
  if (linecount == cy)
    line_table_cur = line_table_ptr;
  line_table_next = line_table_ptr->next_line;
  if (linecount == (LINES-2))
    showline(line_table_ptr,line_table_ptr->text,linecount,0,HILITE);
  line_table_bot = line_table_ptr;
  line_table_ptr = line_table_next;
  newcy = linecount;
  linecount = linecount + 1;
  }

if (line_table_cur == NULL)
  {
  line_table_cur = line_table_bot;
  cy = newcy;
  }
}

/*************************/
void showblock(int hilite)
{
int linecount = 0;
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;

line_table_ptr = line_table_top;

while (  (line_table_ptr != NULL)
      && (linecount <= LINES-2)
      )
  {
  if (line_table_ptr->in_block == TRUE)
    showline(line_table_ptr,line_table_ptr->text,linecount,0,hilite);

  line_table_next = line_table_ptr->next_line;
  line_table_bot = line_table_ptr;
  line_table_ptr = line_table_next;
  linecount = linecount + 1;
  }
}

/*************************/
void recalc_screen_cur()
/*
-------------
starting at line_table_top, recalc all other ptr's and
make current line the line that ends up at row cy on the screen.
-------------
*/
{
int linecount = 0;
int newcy = 0;
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;

line_table_ptr = line_table_top;

line_table_cur = NULL;

while (  (line_table_ptr != NULL)
      && (linecount <= LINES-2)
      )
  {
  if (linecount == cy)
    line_table_cur = line_table_ptr;
  line_table_next = line_table_ptr->next_line;
  line_table_bot = line_table_ptr;
  line_table_ptr = line_table_next;
  newcy = linecount;
  linecount = linecount + 1;
  }

if (line_table_cur == NULL)
  {
  line_table_cur = line_table_bot;
  cy = newcy;
  }
}

/*************************/
void showscreen()

/*
-------------
display screenfull of text starting at line_table_top,
and make current line the line that ends up at row cy on the screen.
-------------
*/

{
int linecount = 0;
int newcy = 0;
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;

  line_table_ptr = line_table_top;

  cur_sx = calc_sx_str(line_table_cur->text,cx);

  if (cur_sx > (vx + COLS - 1))
    vx = cur_sx - COLS + (COLS/3);
  else
    if (cur_sx < vx)
      {
      vx = cur_sx - (COLS/3);
      if (vx < 0) vx = 0;
      }
  line_table_cur = NULL;

  while (  (line_table_ptr != NULL)
        && (linecount <= LINES-2)
        )
    {
    showline(line_table_ptr,line_table_ptr->text,linecount,0,HILITE);

    if (linecount == cy)
      line_table_cur = line_table_ptr;
    line_table_next = line_table_ptr->next_line;
    line_table_bot = line_table_ptr;
    line_table_ptr = line_table_next;
    newcy = linecount;
    linecount = linecount + 1;
    }

  if (line_table_cur == NULL)
    {
    line_table_cur = line_table_bot;
    cy = newcy;
    }

  while (linecount <= LINES-2)
    {
    cmove(linecount,0);
    clreol();
    linecount = linecount + 1;
    }

cur_sx = calc_sx_str(line_table_cur->text,cx);
cmove(cy,cur_sx-vx);
}

/*************************/
void showscreen_cur()

/*
-------------
display screenfull of text starting at line_table_top,
and make current line the line that ends up at row cy on the screen,
BUT only display lines from cursor position down.
-------------
*/

{
int linecount = 0;
int newcy = 0;
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;

line_table_ptr = line_table_top;

cur_sx = calc_sx_str(line_table_cur->text,cx);

if (cur_sx > (vx + COLS - 1))
  vx = cur_sx - COLS + (COLS/3);
else
  if (cur_sx < vx)
    {
    vx = cur_sx - (COLS/3);
    if (vx < 0) vx = 0;
    }
line_table_cur = NULL;

while (  (line_table_ptr != NULL)
      && (linecount <= LINES-2)
      )
  {
  if (linecount >= cy)
    showline(line_table_ptr,line_table_ptr->text,linecount,0,HILITE);

  if (linecount == cy)
    line_table_cur = line_table_ptr;
  line_table_next = line_table_ptr->next_line;
  line_table_bot = line_table_ptr;
  line_table_ptr = line_table_next;
  newcy = linecount;
  linecount = linecount + 1;
  }

if (line_table_cur == NULL)
  {
  line_table_cur = line_table_bot;
  cy = newcy;
  }

while (linecount <= LINES-2)
  {
  cmove(linecount,0);
  clreol();
  linecount = linecount + 1;
  }

cur_sx = calc_sx_str(line_table_cur->text,cx);

cmove(cy,cur_sx-vx);

}

/************************/
void recalc_screen_cy(int y)
/*
--------
 starting at line_table_top, recalc ptr's and
 make currrent line as far down from the top to row y as possible
--------
 */
{
int linecount = 0;
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;

line_table_ptr = line_table_cur;
line_table_next = line_table_ptr->prev_line;

while (  (line_table_next != NULL)
      && (linecount < y)
      )
  {
  line_table_ptr = line_table_next;
  line_table_next = line_table_ptr->prev_line;
  linecount = linecount + 1;
  }

line_table_top = line_table_ptr;
linecount = 0;

while (  (line_table_ptr != NULL)
      && (linecount <= LINES-2)
      )
  {
  if (line_table_cur == line_table_ptr)
    cy = linecount;
  line_table_next = line_table_ptr->next_line;
  line_table_bot = line_table_ptr;
  line_table_ptr = line_table_next;
  linecount = linecount + 1;
  }
}

/************************/
void showscreency(int y)
/*
--------
 show screen-full of text, with currrent line as far down from the
 top to row y as possible
--------
 */
{
int linecount = 0;
struct line_rec *line_table_ptr;
struct line_rec *line_table_next;

if (  (use_new_cy)
   && (new_cy >= 0)
   && (new_cy <= (LINES-2))
   )
  cy = new_cy;
else
  {
  line_table_ptr = line_table_cur;
  line_table_next = line_table_ptr->prev_line;

  cur_sx = calc_sx_str(line_table_cur->text,cx);

  if (cur_sx > (vx + COLS - 1))
    vx = cur_sx - COLS + (COLS/3);
  else
    if (cur_sx < vx)
      {
      vx = cur_sx - (COLS/3);
      if (vx < 0) vx = 0;
      }
  while (  (line_table_next != NULL)
        && (linecount < y)
        )
    {
    line_table_ptr = line_table_next;
    line_table_next = line_table_ptr->prev_line;
    linecount = linecount + 1;
    }

  line_table_top = line_table_ptr;
  linecount = 0;

  while (  (line_table_ptr != NULL)
        && (linecount <= LINES-2)
        )
    {
    showline(line_table_ptr,line_table_ptr->text,linecount,0,HILITE);

    if (line_table_cur == line_table_ptr)
     cy = linecount;
    line_table_next = line_table_ptr->next_line;
    line_table_bot = line_table_ptr;
    line_table_ptr = line_table_next;
    linecount = linecount + 1;
    }

  while (linecount <= LINES-2)
    {
    cmove(linecount,0);
    clreol();
    linecount = linecount + 1;
    }
  }
cur_sx = calc_sx_str(line_table_cur->text,cx);
cmove(cy,cur_sx-vx);
use_new_cy = FALSE;
new_cy = cy;
}

/************************/
void initialise()
{
int i,j;
int argerror = FALSE;
struct line_rec *line_table_new;
char temp_arg[LINEMAXLEN];
struct stat fstat_buffer;
char * p;

strcpy(errortext,"finished ok.");

if (ioctl(0,TCGETA,(struct termio *) &old_term))
  {
  fprintf(stderr,"\nUnable to get initial terminal characteristics.\n");
  exit(1);
  }

new_term = old_term;
/* new_term.c_iflag = new_term.c_iflag & ~ISTRIP & ~IXON & ~BRKINT; */
new_term.c_lflag = new_term.c_lflag & ~ECHO & ~ICANON & ~ISIG;
/*
new_term.c_cc[0] = 255;
new_term.c_cc[0] = 255;
new_term.c_cc[0] = 255;
new_term.c_cc[0] = 255;
*/
new_term.c_cc[4] = 1; /* MIN  */
new_term.c_cc[5] = 0; /* TIME */

if (ioctl(0,TCSETAF,(struct termio *) &new_term))
  {
  fprintf(stderr,"\nUnable to setup terminal - ioctl() failed.\n");
  exit(1);
  }

/* flush terminal input buffer & wait until no key is being pressed */
new_term.c_cc[4] = 0;
new_term.c_cc[5] = 3;
if (ioctl(0,TCSETAF,(struct termio *) &new_term))
  {
  fprintf(stderr,"\nUnable to setup terminal - ioctl() failed.\n");
  exit(1);
  }

/* Flush the kb buffer */
c = x_getch();
while (c != ERR) c = x_getch();

new_term.c_cc[4] = 1; /* MIN  */
new_term.c_cc[5] = 0; /* TIME */

if (ioctl(0,TCSETAF,(struct termio *) &new_term))
  {
  fprintf(stderr,"\nUnable to setup terminal - ioctl() failed.\n");
  exit(1);
  }


/* Once past here, we can use errortext & finish(n) on error conditions */
setup();
enter_ca();

scr_buffer_i = 0;
cmove(0,0);
if (scr_buffer_i == 0)
  {
  strcpy(errortext,"Terminal does not have cursor addressing capability.");
  finish(1);
  }
scr_buffer_i = 0;
insline();
if (scr_buffer_i == 0)
  {
  strcpy(errortext,"Terminal does not have 'insert line' capability.");
  finish(1);
  }
scr_buffer_i = 0;
deline();
if (scr_buffer_i == 0)
  {
  strcpy(errortext,"Terminal does not have 'delete line' capability.");
  finish(1);
  }
scr_buffer_i = 0;
clreol();
if (scr_buffer_i == 0)
  {
  strcpy(errortext,"Terminal does not have 'clear to end of line' capability.");
  finish(1);
  }
scr_buffer_i = 0;
clrscr();
if (scr_buffer_i == 0)
  {
  strcpy(errortext,"Terminal does not have 'clear screen' capability.");
  finish(1);
  }
scr_buffer_i = 0;
clrbot();
if (scr_buffer_i == 0)
  {
  strcpy(errortext,"Terminal does not have 'clear to bottom of screen' capability.");
  finish(1);
  }
scr_buffer_i = 0;

if (  (g_argc < 1)
   || (g_argc > 5)
   )
  {
  strcpy(errortext,usage);
  finish(1);
  }

strcpy(pathname,"");

p = getenv("MEDITKEYPATH");
if (p != NULL)
  {
  strncpy(pathname,p,LINEMAXLEN-1);
  i = 0;
  while (pathname[i] != '\0') i++;
  i--;
  if (pathname[i] != '/')
    strcat(pathname,"/");
  }
else
  {
  strcpy(pathname,"");
  i = 0;
  while (g_argv[0][i] != '\0')
    {
    pathname[i] = g_argv[0][i];
    i++;
    }

  pathname[i] = '\0';
  while (  (pathname[i] != '/')
        && (i >= 0)
        )
    i -= 1;
  if (i < 0)
    {
    i = 0;
    pathname[i] = '\0';
    }
  else
    if (pathname[i] == '/')
      pathname[i+1] = '\0';
  }

p = getenv("TERM");
if (p != NULL)
  strncpy(termtype,p,LINEMAXLEN-1);
else
  strcpy(termtype,"");


strcpy(filename,"");
strcpy(keymapname,"");
j = 1;
while (  (j < g_argc)
      && (!argerror)
      )
  {
  i = 0;
  while (i < 1)
    {
    temp_arg[i] = g_argv[j][i];
    i++;
    }
  temp_arg[i] = '\0';
  if (temp_arg[0] == '-')
    {
    i = 0;
    while (i < 2)
      {
      temp_arg[i] = g_argv[j][i];
      i++;
      }
    temp_arg[i] = '\0';
    upper_case(temp_arg);
    if (strcmp(temp_arg,"-T") == STR_SAME)
      {
      highlight_tabs = TRUE;
      j += 1;
      }
    else
      {
      if (strcmp(temp_arg,"-K") == STR_SAME)
        {
        j += 1;
        if (j < g_argc)
          {
          keymap = TRUE;
          strncpy(keymapname,g_argv[j],LINEMAXLEN-1);
          j += 1;
          }
        else
          {
          strcpy(errortext,"-K option had no keymap <filename>.");
          argerror = TRUE;
          }
        }
      else
        {
        sprintf(errortext,"Option %s is invalid.",temp_arg);
        argerror = TRUE;
        }
      }
    }
  else
    {
    strncpy(filename,g_argv[j],LINEMAXLEN-1);
    j += 1;
    }
  }

if (argerror)
  {
  strcat(errortext,"\n\n");
  strcat(errortext,usage);
  finish(1);
  }

/*
if (strcmp(filename,"") == STR_SAME)
  {
  strcpy(errortext,"<filename> must be specified\n\n");
  strcat(errortext,usage);
  finish(1);
  }
*/

if (!keymap)
  {
  strcpy(keymapname,pathname);
  if (strcmp(termtype,"") == STR_SAME)
    strcat(keymapname,"medit.key");
  else
    {
    strcat(keymapname,"medit.key.");
    strcat(keymapname,termtype);
    }
  }

tmpnam(tempfile);

/* build the tab stops table */
sx = 0;
i = 1;
while (i <= 80)
  {
  sx += 8;
  if (sx == 0)
    tabstop[i] = 32000;
  else
    tabstop[i] = sx;
  i = i + 1;
  }
/* end of tabstop code */

buff_table_start = buff_table_end = NULL;
line_table_block_start = line_table_block_end = NULL;

strcpy(findstr,"");
strcpy(find_text,"");
strcpy(mess_text,"");
strcpy(prev_command,"");
strcpy(line_buffer,"");

sprintf(file_text,"File: %s",filename);

message_pending = FALSE;

line_table_start = (struct line_rec *) malloc(sizeof (struct line_rec));
if (line_table_start == NULL)
  {
/*
  exit_ca();
  ioctl(0,TCSETAF,(struct termio *) &old_term);
  winptr = NULL;
*/
  strcpy(errortext,"not enough memory available");
  finish(4);
  }

file_empty = TRUE;
line_table_start->text[0] = '\0';
line_table_start->in_block = FALSE;
line_table_start->next_line = NULL;
line_table_start->prev_line = NULL;
line_table_top = line_table_start;
line_table_cur = line_table_start;
line_table_bot = line_table_start;
line_table_end = line_table_start;

load_key_map();

COLS = getCOLS();

LINES = getLINES();

clrscr();

if (strcmp(filename,"") == STR_SAME)
  {
  if (!message_pending)
    next_mess("New file.");
  showscreen();
  }
else
  {
  readfile(filename,LOAD_FILE);
  if (read_failed)
    {
    strcpy(errortext,mess_text);
    finish(1);
    }
  }
}

/******************************************************************************
   Testing and setting the file write permissions flag bit.
   --------------------------------------------------------
   The flag bits for the file are in an integer. There are predifined int
   values in sys/stat.h for each of the possible file permissions (e.g. user
   write permission is S_IWUSR), and the value equals that flag bit set on
   its own.
   To test if the S_IWUSR bit is set, the file flag set is logical ANDed with
   the S_IWUSR value, and two possible values result depending on whether that
   bit is set in the file flag set:

   The bit S_IWUSR bit is set       The S_IWUSR bit is not set
               v                            v
       1010101010101010 (File flag) 0101010101010101
   AND 0000000010000000  (S_IWUSR)  0000000010000000 AND
      ------------------           ------------------
       0000000010000000  (result)   0000000000000000

   The logical AND operation only sets a bit in the result if both the
   corresponding bits are set in the two operands. So if the bit is set on in
   the file flag bit set, then the result equals the S_IWUSR, otherwise it
   equals zero.

   In the save() subroutine a similar scheme allows us to set the bit, (if we
   want to write to a read-only file), except that we use the logical OR
   operation.

   The bit S_IWUSR bit to be set    The S_IWUSR bit to be set
               v                            v
       1111111101111111 (File flag) 0000000000000000
    OR 0000000010000000  (S_IWUSR)  0000000010000000 OR
      ------------------           ------------------
       1111111111111111  (result)   0000000010000000

   The logical OR operation sets a bit in the result if either of the two
   corresponding bits are set on in the operands, otherwise it leaves the bit
   unchanged. We OR with the S_IWUSR value which has only the required bit set
   on, therefore that bit gets set but leaves the others unchanged. (If a file
   flag bit was not set, the OR will not set it as the S_IWUSR value only has
   the one bit set on. If a file flag bit was set on then it will be left set,
   logical OR never unsets bits).

   The save() subroutine must then unset the bit after writing to a read-only
   file. To do this the stat.h S_IWUSR value is 1's complemented (0's & 1's
   are flipped to the opposite value) using the ~ operator.

        S_IWUSR = 0000000010000000
       ~S_IWUSR = 1111111101111111

   The complemented value is then ANDed with the file flag bit unset:

   The S_IWUSR bit to be unset      The S_IWUSR bit to be unset
               v                            v
       1111111111111111 (File flag) 0000000010000000
   AND 1111111101111111 (~S_IWUSR)  1111111101111111 AND
      ------------------           ------------------
       1111111101111111  (result)   0000000000000000

   The way the AND operation works ensures that only the required bit gets
   unset and all the others are left unchanged.

******************************************************************************/

/*************************/
void readfile(char readfname[],int load_or_paste)
{
struct line_rec *line_table_temp;
struct line_rec *line_table_new;
struct line_rec *line_table_new_start;
struct line_rec *line_table_prev;
char temp_str[LINEMAXLEN];
char message_str[LINEMAXLEN];
int read_only = FALSE;
struct stat fstat_buffer;

read_failed = FALSE;
text_had_nulls = FALSE;
text_had_tabs = FALSE;
total_chars = total_lines = 0;

fptr = fopen(readfname,"r");

if (fptr == NULL)
  {
  if (load_or_paste == PASTE_FILE)
    next_mess("Non-existent file.");
  else
    {
    if (!message_pending)
      next_mess("New file.");
    showscreen();
    }
  }
else
  {
  /* check that "filename" is an ordinary file */
  stat(readfname,&fstat_buffer);
  if ((fstat_buffer.st_mode & S_IFDIR) == S_IFDIR)
    {
    sprintf(temp_str,"%s is a directory.",readfname);
    next_mess(temp_str);
    read_failed = TRUE;
    goto exit;
    }
  else if ((fstat_buffer.st_mode & S_IFREG) != S_IFREG)
    {
    sprintf(temp_str,"%s is not an ordinary file.",readfname);
    next_mess(temp_str);
    read_failed = TRUE;
    goto exit;
    }

  c = getline(fptr,line_buffer,LINEMAXLEN);

  total_chars = total_chars + no_of_chars;

  line_table_new_start =
    (struct line_rec *)
    malloc(sizeof (struct line_rec) + strlen(line_buffer));

  if (line_table_new_start == NULL)
    {
    next_mess("Insufficient memory.");
    fclose(fptr);
    clrscr();
    showscreen();
    goto exit;
    }
  strcpy(line_table_new_start->text,line_buffer);
  line_table_new_start->in_block = FALSE;
  line_table_new_start->prev_line = NULL;
  line_table_new_start->next_line = NULL;

  line_table_new = line_table_new_start;

  if (c != EOF)
    {
    while (c != EOF)
      {
      c = getline(fptr,line_buffer,LINEMAXLEN);
      total_chars = total_chars + no_of_chars;

      line_table_next =
        (struct line_rec *)
        malloc(sizeof (struct line_rec) + strlen(line_buffer));

      line_table_new->next_line = line_table_next;

      if (line_table_next == NULL)
        {
        next_mess("Insufficient memory.");
        line_table_next = line_table_new;
        goto while_exit;
        }

      line_table_next->prev_line = line_table_new;
      line_table_next->next_line = NULL;
      line_table_next->in_block = FALSE;
      strcpy(line_table_next->text,line_buffer);
      line_table_new = line_table_next;
      }

    line_table_temp = line_table_new->prev_line;
    if (line_table_new != NULL)
      free(line_table_new);
    line_table_new = line_table_temp;
    line_table_new->next_line = NULL;
    }

while_exit:

  line_table_prev = line_table_cur->prev_line;

  if (file_empty)
    {
    line_table_start = line_table_top = line_table_new_start;
    free(line_table_cur);
    line_table_cur = line_table_new_start;
    line_table_bot = line_table_end = line_table_new;
    }
  else
    {
    if (line_table_prev != NULL)
      line_table_prev->next_line = line_table_new_start;
    else
      line_table_start = (line_table_top = line_table_new_start);
    line_table_new_start->prev_line = line_table_prev;
    line_table_cur->prev_line = line_table_new;
    line_table_new->next_line = line_table_cur;
    }

  if (strcmp(mess_text,"") == STR_SAME)
/* then nothing has gone wrong so far - so continue */
    {
    if (load_or_paste == PASTE_FILE)
      {
       if (  (!text_had_nulls)
         && (!text_had_tabs)
         )
        strcpy(message_str,"Pasted file ok. ");
      else
        {
        strcpy(message_str,"Pasted file had ");
        if (  (text_had_tabs)
           && (text_had_nulls)
           )
          strcat(message_str,"TABs & NULLs. ");
        else if (text_had_tabs)
          strcat(message_str,"TABs. ");
        else if (text_had_nulls)
          strcat(message_str,"NULLs. ");
        }
       sprintf(temp_str,
             "Read %d lines, %d chars.",
              total_lines,
              total_chars
             );
      strcat(message_str,temp_str);
      next_mess(message_str);
      file_needs_saving = TRUE;
      }

    else
/* This is not a paste, but an initial load of a file */
      {

      /* get the file permission flag bit set */
      /* for a discussion of this see the comment above this subroutine */

      stat(filename,&fstat_buffer);

      if (  ((fstat_buffer.st_mode & S_IWUSR) == S_IWUSR)
            /* file is not read-only */
         && (!text_had_nulls)
         )
        strcpy(message_str,"(c) 1995 Mark Taylor. ");
      else
        {
        /* test the file user write permission flag */
        /* for a discussion of this see the comment above this subroutine */

        if ((fstat_buffer.st_mode & S_IWUSR) != S_IWUSR)
          strcpy(message_str,"Read-only file");
        else
          strcpy(message_str,"File");
        if (!text_had_nulls)
          strcat(message_str,". ");
        else
          {
          strcat(message_str," had ");
          if (text_had_nulls)
            strcat(message_str,"NULLs. ");
          }
        }
      sprintf(temp_str,
             "Read %d lines, %d chars.",
              total_lines,
              total_chars
             );
      strcat(message_str,temp_str);
      if (!message_pending)
        next_mess(message_str);
      file_needs_saving = FALSE;
      }
    }

  if (total_chars > 0)
    file_empty = FALSE;
  showscreency(cy);
  fclose(fptr);
  }
exit:
  ;
}

/*************************/
void help()
{
int i=0;
struct key_desc_node * key_desc_ptr;
char temp_str[82];
char temp_str2[82];

enter_bold();
enter_inv();

cmove(i,0);
clreol();
x_waddstr(" Medit help - press any key to return to edit mode                             ",NO_VX_CHECK,0);

normal();
enter_bold();

key_desc_ptr = key_desc_table;
while (  (i < (LINES-12))
      && (key_desc_ptr != NULL)
      )
    {
    i += 1;
    sprintf(temp_str,"%s - %s | ",key_desc_ptr->key_desc,key_desc_ptr->func_desc);
    key_desc_ptr = key_desc_ptr->next_node;
    if (key_desc_ptr != NULL)
      sprintf(temp_str2,"%s - %s ",key_desc_ptr->key_desc,key_desc_ptr->func_desc);
    else
      strcpy(temp_str2,  "                                     ");
    strcat(temp_str,temp_str2);
    strcat(temp_str,"   ");
    cmove(i,0);
    clreol();
    x_waddstr(temp_str,NO_VX_CHECK,0);
    if (key_desc_ptr != NULL)
      key_desc_ptr = key_desc_ptr->next_node;
    }

i += 1;
if (i <= LINES-2)
{
enter_inv();
cmove(i,0);
clreol();
x_waddstr(" Commands available from command line (can be abbreviated to first letter)     ",NO_VX_CHECK,0);

normal();
enter_bold();
i = i + 1;

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr(" EXIT [filename] - Save and quit.        QUIT - Quit without saving.           ",NO_VX_CHECK,0);
i = i + 1;

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr(" WRITE filename  - save text or block    SAVE [filename] - save.               ",NO_VX_CHECK,0);
i = i + 1;

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr("                   to a new file.        FILE filename  - paste from a file.   ",NO_VX_CHECK,0);
i = i + 1;

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr(" 'string'        - find string.          CASE'string'   - case sensitive find. ",NO_VX_CHECK,0);
i = i + 1;

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr(" REPLACE 'abc'xyz' ALL   - Find occurrances of 1st string, replace with 2nd.   ",NO_VX_CHECK,0);
i = i + 1;

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr(" REPLACE 'abc'xyz' QUERY - Find occurrances of 1st string, query & replace.    ",NO_VX_CHECK,0);
i = i + 1;

if (i <= LINES-2)
{
  if (!blk_cmd_mapped)
    {
    cmove(i,0);
    clreol();
    x_waddstr(" COPY - copy block.        MOVE - move block.        DELETE - delete block.    ",NO_VX_CHECK,0);
    i = i + 1;
    }

if (i <= LINES-2)
{
  if (!blk_cmd_mapped)
    {
    cmove(i,0);
    clreol();
    x_waddstr(" PASTE - paste contents of buffer.                   BLOCK  - mark block.      ",NO_VX_CHECK,0);
    i = i + 1;
    }

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr(" Cursor Up   - previous command.     a number, eg 99 - go to that line number. ",NO_VX_CHECK,0);
i = i + 1;

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr(" Cursor Down - blank command         ! Unix command  - execute a Unix command. ",NO_VX_CHECK,0);
i = i + 1;

if (i <= LINES-2)
{
cmove(i,0);
clreol();
x_waddstr(" MAP         - Produce keymap file.                                            ",NO_VX_CHECK,0);
}
}
}
}
}
}
}
}
}
}
}
}

while(i < LINES-2)
  {
  i += 1;
  cmove(i,0);
  clreol();
  x_waddstr("                                                                               ",NO_VX_CHECK,0);
  }


i = decode_key();

/*
if (ccx == -1)
  restore_status(ALL_STATUS,LINES-1);
*/

showscreen();

if (ccx >= 0)
  cmove(LINES-1,ccx);
}

/*************************/
void ctrl_char_help()
{
int i=0;

enter_bold();
enter_inv();
cmove(i,0);
clreol();
x_waddstr("              Medit help - press any key to return to edit mode                        ",NO_VX_CHECK,0);

normal();
enter_bold();

i += 1;
cmove(i,0);
clreol();
x_waddstr("                                                                               ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("You can enter a control character in one of two ways:                          ",NO_VX_CHECK,0);
i += 1;
cmove(i,0);
clreol();
x_waddstr("                                                                               ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("1. Press the CTRL key and a letter simultaneously e.g. CTRL-A.                 ",NO_VX_CHECK,0);
i += 1;
cmove(i,0);
clreol();
x_waddstr("                                                                               ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("2. Type the ASCII code of the control character in decimal, e.g. 26 for CTRL-Z.",NO_VX_CHECK,0);
i += 1;
cmove(i,0);
clreol();
x_waddstr("                                                                               ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("Note that option 2 must be used in the following cases:                        ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("Ascii codes > 126; any key escape codes defined in the key mapping.            ",NO_VX_CHECK,0);
i += 1;
cmove(i,0);
clreol();
x_waddstr("                                                                               ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("The following character codes are not allowed:                                 ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("0 (null), 10 (new line), any code > 255                                        ",NO_VX_CHECK,0);
i += 1;
cmove(i,0);
clreol();
x_waddstr("                                                                               ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("To enter a newline character, press RETURN while in edit mode!                 ",NO_VX_CHECK,0);
i += 1;
cmove(i,0);
clreol();
x_waddstr("                                                                               ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("To correct typing errors, use the <Delete Line> key to backspace.              ",NO_VX_CHECK,0);
i += 1;
cmove(i,0);
clreol();
x_waddstr("                                                                               ",NO_VX_CHECK,0);
i = i + 1;
cmove(i,0);
clreol();
x_waddstr("Press RETURN to enter the control character.                                   ",NO_VX_CHECK,0);

while(i < LINES-2)
  {
  i += 1;
  cmove(i,0);
  clreol();
  x_waddstr("                                                                               ",NO_VX_CHECK,0);
  }

i = decode_key();

/*
if (ccx == -1)
  restore_status(ALL_STATUS,LINES-1);
*/

showscreen();
if (ccx >= 0)
  cmove(LINES-1,ccx);
}

/******************************/
void next_mess(char message[])
{
strncpy(mess_text,message,60);
message_pending = TRUE;
}

/******************************/
int disp_mess()
{
int result;
int prev_pending;

if (scrolling)
  {
  normal();
  cmove(cy,cur_sx-vx);
  return message_pending;
  }

if (strcmp(mess_text,"") != STR_SAME)
  result = TRUE;
else
  result = FALSE;

prev_pending = message_pending;
message_pending = FALSE;

restore_status(ALL_STATUS,LINES-1);

message_pending = prev_pending;

normal();
cmove(cy,cur_sx-vx);

strcpy(mess_text,"");
return result;
}

/*******************************/
void finish(int errcode)
{
exit_ca();
if (ioctl(0,TCSETAF,(struct termio *) &old_term))
  fprintf(stderr,"Unable to RESET initial terminal characteristics.\n");

if (fptr != NULL)
  fclose(fptr);

fprintf(stderr,"\n\nMedit: ");
fprintf(stderr,errortext);
fprintf(stderr,"\n\n");
if (save_failed)
  {
  fprintf(stderr,"An attempt to save a file has failed.\n");
  fprintf(stderr,"You may be able to recover your edited text from the file:\n");
  fprintf(stderr,"'%s'\n\n",tempfile);
  fprintf(stderr,"Please delete this temporary file after you have finished with it.\n");
  }

exit(errcode);
}

/*******************************/
/* function that executes a mv equivalent. Writes old to new, either via
   a link if files on same volume, or by a copy. Old file is deleted on
   successful termination.
   Returns 0 for success, or -1 for any other condition, sets errno if
   errno.h included */

int rename_file(char *old, char *new)
{
FILE *oldf, *newf;
int c;

/* try to use links..can't span mounted volumes */

if (link(new,old)) {
  /* if fail, copy data over by opening up files*/
  if ((oldf = fopen(old,"r")) == NULL)
    goto exit;
  if ((newf = fopen(new,"w")) == NULL)
    goto exit;
  /* then doing a char by char copy, hope you've got small files */
  while (( c = getc(oldf)) != EOF) {
    putc(c,newf);
    /* of course checking for any wierd errors */
    if (ferror(oldf) || ferror(newf)) {
      fclose(oldf);
      fclose(newf);
      goto exit;
      }
    }
  /* and then shutting down and removing the temporary file */
  fclose(oldf);
  fclose(newf);
  if (ferror(oldf) || ferror(newf))
    goto exit;
  unlink(old);
  }
else {
  unlink(old);
  }
return 0;
/* returns -1 if any problems, also sets errno */
exit:
  return -1;
}

/*******************************/
void display(char * mess, int pause)
/* This function is for debugging purposes. */
{
int c;

cmove(LINES-1,20);
clreol();

normal();
enter_blink();
x_waddstr(mess,NO_VX_CHECK,0);
normal();
enter_bold();

if (pause)
  c = decode_key();
  /* while((c = x_getch()) == ERR); */

cmove(cy,cur_sx-vx);
}

/*******************************/
char getline(FILE *stream,char s[],int line_max)
  {
  int c,i,result;

  i=0;
  no_of_chars = 0;
  tabs_in_line = 0;

  c=getc(stream);
  while  (  (i < line_max-1)
         && (c != EOF)
         && (c != '\n')
         )
    {
    if (c != '\0')
      {
      s[i] = c;
      if (c == TAB)
        tabs_in_line += 1;
      i=i+1;
      }
    else
      text_had_nulls = TRUE;
    c=getc(stream);
    }

  if (  (c != EOF)
     || (i != 0)
     )
    {
    total_lines = total_lines + 1;
    no_of_chars = i + 1;
    }

  s[i] = '\0';

  if (  (c != EOF)
     && (c != '\n')
     )
    {
    sprintf(errortext,
           "one or more lines are > %d characters long\n",
           line_max-1);
    finish(3);
    }

return c;
}

