#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#include "types.h"
#include "raster.h"

int RdWORD_LoEnd( WORD *num, FILE *f );
int RdDWORD_LoEnd( DWORD *num, FILE *f );
int RdWORD_HiEnd( WORD *num, FILE *f );
int RdDWORD_HiEnd( DWORD *num, FILE *f );


#define Rd_word(f,w) RdWORD_LoEnd(w,f);
#define Rd_dword(f,d) RdDWORD_LoEnd(d,f);
#define Wr_word(f,w) WrWORD_LoEnd(w,f);
#define Wr_dword(f,d) WrDWORD_LoEnd(d,f);

#ifndef SEEK_SET
 #define SEEK_SET 0
#endif

int loadstruct(FILE *F,char *description, ...);

char *UnitName="Rasters";

void RGB_BGR(char *Data,int size)
{
char c;
while(size-->0)
	{
	c=Data[2];
	Data[2]=*Data;
	*Data=c;
	Data+=3;
	}
}

//------------------BMP--------------------------
struct  BMPHeader
	{
	BYTE    bfType[2]	PACKED;
	DWORD   bfSize		PACKED;
	WORD    bfReserved1	PACKED;
	WORD    bfReserved2	PACKED;
	DWORD   bfOffBits	PACKED;
	};

struct	BitmapInfoHeader
	{
	DWORD  biSize		PACKED;
	DWORD  biWidth		PACKED;
	DWORD  biHeight		PACKED;
	WORD   biPlanes		PACKED;
	WORD   biBitCount	PACKED;
	DWORD  biCompression	PACKED;
	DWORD  biSizeImage	PACKED;
	DWORD  biXPelsPerMeter	PACKED;
	DWORD  biYPelsPerMeter	PACKED;
	DWORD  biClrUsed	PACKED;
	DWORD  biClrImportant	PACKED;
	};

struct	BitmapCoreInfoHeader
	{
	DWORD bcSize		PACKED;
	WORD  bcWidth		PACKED;
	WORD  bcHeight		PACKED;
	WORD  bcPlanes		PACKED;
	WORD  bcBitCount	PACKED;
	};


union BMPInfo_Coreinfo
	{
	BitmapInfoHeader BIH	PACKED;
	BitmapCoreInfoHeader BCIH PACKED;
	};

struct BMP_RGBQuad
	{
	BYTE rgbBlue;
	BYTE rgbGreen;
	BYTE rgbRed;
	BYTE rgbReserved;
	};

Image LoadPictureBMP(const char *Name)
{
FILE  *f;
WORD ldblk;
int   i;
BMPHeader  Header;
BMPInfo_Coreinfo Info;
BMP_RGBQuad PalItem;
long       w;
BYTE	   k;
int RetValue;
Raster2DAbstract *Raster=NULL;
APalette *Palette=NULL;
char Compression=0;
Image Img;


 if((f=fopen(Name,"rb"))==NULL) return(Img);

 ldblk=fread(&Header,sizeof(Header),1,f);
 if((Header.bfType[0]!='B')||(Header.bfType[1]!='M') ||
      (ldblk!=1)) {fclose(f);return(Img);}

 if(fread(&Info,4,1,f)!=1) {fclose(f);return(Img);};
 if(Info.BIH.biSize>40) goto ENDPROC;

 if(fread(&Info.BIH.biWidth,Info.BIH.biSize-4,1,f)!=1) {fclose(f);return(NULL);};

 switch(Info.BIH.biSize)
	{
	case 12:
	   {
	   if(Info.BCIH.bcPlanes!=1) goto ENDPROC;
	   Raster=CreateRaster2D(Info.BCIH.bcWidth,Info.BCIH.bcHeight,
				 Info.BCIH.bcPlanes*Info.BCIH.bcBitCount);
	   k=3;			//RGB Triple
	   break;
	   }
	case 40:
	   {
	   if(Info.BIH.biPlanes!=1) goto ENDPROC;
	   Compression=Info.BIH.biCompression;
	   if(Info.BIH.biCompression>2) goto ENDPROC;; //unsupported compression mode
	   Raster=CreateRaster2D(Info.BIH.biWidth,Info.BIH.biHeight,
				 Info.BIH.biPlanes*Info.BIH.biBitCount);
	   k=4;			//RGB Quad
	   break;
	   }
	default: goto ENDPROC;
	}

 if(Raster==NULL) goto ENDPROC;				//unsupported format
 i=Raster->GetPlanes();
 if(!((i==1)||(i==4)||(i==8)||(i==24)))
	 {
ERASEEND:delete Raster;
	 Raster=NULL;
	 goto ENDPROC;
	 }

 if (i==8||i==4) 				//load of the palette
	{
	Palette=BuildPalette(1 << i,8);
	if(Palette!=NULL)
	  {
	  for(i=0;i<(1 << Raster->GetPlanes()) /*p.palette^.Colors-1*/  ;i++)
	     {
	     fread(&PalItem,sizeof(PalItem),1,f);
	     Palette->R(i,PalItem.rgbRed);
	     Palette->G(i,PalItem.rgbGreen);
	     Palette->B(i,PalItem.rgbBlue);
	     }
	  if(GrayPalette(Palette,Raster->GetPlanes()))
		{
		delete Palette;
		Palette=NULL;
		}
	  };
	}

 fseek(f,Header.bfOffBits,SEEK_SET);
 ldblk=(Raster->GetPlanes()*Raster->Size1D+7)/8;
 k=(-ldblk) & 0x3;

 switch(Compression)
   {
   case 0:for(i=Raster->Size2D-1;i>=0;i--)
		{
		if(fread(Raster->GetRow(i),ldblk,1,f)!=1) {goto ERASEEND;}

		fread(&w,k,1,f);			//read till end of line
 //		AlineProc(i,p);
		}
	  break;
   case 1:{
	  unsigned x,y;
	  unsigned char *ptrB;

	  y=Raster->Size2D;
	  x=0;
	  i=0;
	  while( y-->0 )
	    {
	    ptrB=(unsigned char *)Raster->GetRow(y);
	    if(!ptrB) goto End_of_bitmap;
	    while(!feof(f))
		{
		Rd_word(f,(WORD*)&i);
		if((i&0xFF)==0) 		//escape code
			{
			i=i>>8;
			if(i==0) {x=0;break;}
			if(i==1) goto End_of_bitmap;
			if(i==2)
			   {
			   Rd_word(f,(WORD*)&i);
			   x+=i&0xFF;
			   y-=i>>8-1;
			   break;
			   }

			while( (i>0)&&(x<ldblk-1) )
			       {
			       Rd_word(f,(WORD*)&w);
			       *ptrB=w;
			       ptrB++;
			       x++;
			       i--;
			       if((x==ldblk-1)|| i==0 ) {i=0;break;}
			       *ptrB=w >> 8;
			       ptrB++;
			       x++;
			       i--;
			       }
			continue;
			};

		if( x>=ldblk )
			 continue;
		if((i&0xFF)>0)
			{
			k=ldblk-x;
			if(k>(i&0xFF)) k=i&0xFF;
			memset(ptrB,i>>8,k);
			ptrB+=k;
			x+=k;
			i-=k;
			}
		};
//	   if AlineProc<>nil then AlineProc^.NextLine;
	   if(feof(f)) goto End_of_bitmap;	//unexpected end
	   };
	break;
	}

   case 2:{		//BI_RLE4
	 unsigned x,y;

	 y=Raster->Size2D;
	 x=0;
	 while(y-->0)
	   {
	   while(!(feof(f)))
		{
		Rd_word(f,(WORD*)&i);
		if((i&0xFF)==0)			//escape code
			{
			i=i>>8;
			if(i==0) {x=0;break;} 	//end of line
			if(i==1) goto End_of_bitmap;
			if(i==2)
			     {
			     Rd_word(f,(WORD*)&i);
			     x+=i&0xFF;
			     y-=i>>8-1;
			     break;
			     };

			k=0;
			ldblk=0;
			while( i>0 && x<Raster->Size1D )
			    {
			    switch(k)
				{
				case 0:Rd_word(f,&ldblk);
				       Raster->SetValue2D(x,y,(ldblk >> 4)& 0xF);
				       break;
				case 1:Raster->SetValue2D(x,y,ldblk & 0xF);break;
				case 2:Raster->SetValue2D(x,y,ldblk >> 12);break;
				case 3:Raster->SetValue2D(x,y,(ldblk >> 8) & 0xF);break;
				};
			    x++;
			    i--;
			    k=(k+1) & 3;
			    }
			continue;
			}

		if((i&0xFF)>0)
			{
			ldblk=Raster->Size1D-x;
			if(ldblk>(i&0xFF)) ldblk=(i&0xFF);
			k=i>>8;
			for(i=0;i<ldblk;i++)
				{
				if(i&1) Raster->SetValue2D(x,y,k & 0xF);
				   else Raster->SetValue2D(x,y,k >> 4);
				x++;
				}
			}
		}
//	   if AlineProc<>nil then AlineProc^.NextLine;
	   if(feof(f)) goto End_of_bitmap;	//unexpected end
	   }
     break;
     }

   }

End_of_bitmap:


 if(Raster->GetPlanes()==24)
	for(i=Raster->Size2D-1;i>=0;i--)
		RGB_BGR((char *)Raster->GetRow(i),Raster->Size1D);

ENDPROC:
 fclose(f);
 if(Raster!=NULL) {Img.Raster=Raster;Img.Raster->UsageCount++;}
 if(Palette!=NULL){Img.Palette=Palette;Img.Palette->UsageCount++;}
 return(Img);
}


int SavePictureBMP(const char *Name,const Image &Img)
{
FILE *f;
unsigned  ldblk;
BMPHeader Header;
BMPInfo_Coreinfo Info;
char k;
int i;

BMP_RGBQuad PalItem;

 if(Img.Raster==NULL) return(-10);

// if not(p.planes in [1,4,8,24]) then exit;    {nepodporovany format}

 if((f=fopen(Name,"wb"))==NULL) return(-1);

 ldblk=(Img.Raster->GetPlanes()*Img.Raster->Size1D+7)/8;
 memset(&Header,0,sizeof(Header));
 memmove(&Header.bfType,"BM",2);
 Header.bfOffBits=sizeof(Header)+sizeof(Info)+ (1 << (Img.Raster->GetPlanes() + 2));
 Header.bfSize=Header.bfOffBits+ Img.Raster->Size2D*((ldblk+3) & 0xFFFC);

 memset(&Info,0,sizeof(Info));
 Info.BIH.biSize=40;
 Info.BIH.biWidth=Img.Raster->Size1D;
 Info.BIH.biHeight=Img.Raster->Size2D;
 Info.BIH.biPlanes=1;
 Info.BIH.biBitCount=Img.Raster->GetPlanes();
 Info.BIH.biCompression=0;                      //I don't allow compression
 Info.BIH.biSizeImage=Header.bfSize-Header.bfOffBits;
 if ((Img.Raster->GetPlanes()==4)||(Img.Raster->GetPlanes()==8)) Info.BIH.biClrUsed=1 << (Img.Raster->GetPlanes());

 fwrite(&Header,sizeof(Header),1,f);
 fwrite(&Info,sizeof(Info),1,f);

 if ((Img.Raster->GetPlanes()==4)||(Img.Raster->GetPlanes()==8))
       {                                        //palette
//       gray=p.palette==NULL;
//     if(gray) CreatePalette(p.palette,p.planes);
       PalItem.rgbReserved=0;


       if(Img.Palette==NULL)
	 {
	 k=1;
	 if(Img.Raster->GetPlanes()==4) k=16;
	 for(i=0;i<(1 << Img.Raster->GetPlanes()) /*p.palette^.Colors-1*/  ;i++)
	   {
	   PalItem.rgbRed= k*i;   //  p.palette^.pal[K].Red;
	   PalItem.rgbGreen=k*i;  // p.palette^.pal[K].Green;
	   PalItem.rgbBlue=k*i;   // p.palette^.pal[K].Blue;
	   fwrite(&PalItem,sizeof(PalItem),1,f);
	   }
//       if Gray then ErasePalette(p.palette);
	 }
       else
	 {
	 k=Img.Palette->GetPlanes()-8;
	 if(k<0) k=0;
	 for(i=0;i<(1 << Img.Raster->GetPlanes()) /*p.palette^.Colors-1*/  ;i++)
	   {
	   PalItem.rgbRed=Img.Palette->GetValue1D(3*i)>>k;    //  p.palette^.pal[K].Red;
	   PalItem.rgbGreen=Img.Palette->GetValue1D(3*i+1)>>k;// p.palette^.pal[K].Green;
	   PalItem.rgbBlue=Img.Palette->GetValue1D(3*i+2)>>k; // p.palette^.pal[K].Blue;
	   fwrite(&PalItem,sizeof(PalItem),1,f);
	   }
	 }
       }

 fseek(f,Header.bfOffBits,SEEK_SET);
 k=(-(int)ldblk)& 3;
 for(i=Img.Raster->Size2D-1;i>=0;i--)
	{
	if(Img.Raster->GetPlanes()==24) RGB_BGR((char *)Img.Raster->GetRow(i),Img.Raster->Size1D);
	fwrite(Img.Raster->GetRow(i),ldblk,1,f);
	if(Img.Raster->GetPlanes()==24) RGB_BGR((char *)Img.Raster->GetRow(i),Img.Raster->Size1D);
	fwrite(&Header.bfReserved1,k,1,f);      //write zeroes to end of line
//      AlineProc(i,p);
	}
FINISH:
 fclose(f);
 return(0);
}
//-------------------End of BMP routines------------------

//------------------PCX--------------------------
struct RGBColor
       {
	BYTE Red,Green,Blue;
       };


struct PCXHeader{
       BYTE Id0;
       BYTE Version;
       BYTE encoding;
       BYTE BitPerPix;
       WORD x0,y0;
       WORD x1,y1;
       WORD horizontal;
       WORD vertical;
       RGBColor ColorMap[16];
       BYTE reserved;
       BYTE NumPlanes;
       WORD BytesPerLine;
       WORD PaletteInfo;
       BYTE user[58];
       };


Image LoadPicturePCX(const char *Name)
{
const int Mask=0xC0;
//const int MaxCnt=63;
FILE *f;
WORD X,Y;
BYTE C,Cnt;
PCXHeader Header;
BYTE *adata,*data;
int Aplane;
Raster1DAbstract *pbuf=NULL;
APalette *Palette=NULL;
Raster2DAbstract *Raster=NULL;
Image Img;

  if((f=fopen(Name,"rb"))==NULL) return(Img);
  data=NULL;
// Pbuf.Init;
//  p.Erase;
  if(fread(&Header,128,1,f)!=1) {fclose(f);return(Img);};
  if(Header.Id0!=10) goto ENDPROC;	/*incorrect identifier*/

  Raster=CreateRaster2D(Header.x1-Header.x0+1,Header.y1-Header.y0+1,
			Header.BitPerPix*Header.NumPlanes);
  if(Raster==NULL) goto ENDPROC;

//  p.typ='G';
// if(p.planes=24) p.typ='C';
//  if(AlineProc!=NULL) AlineProc->InitPassing(Raster->Size2D,'Loading PCX');

  Aplane=Header.NumPlanes;

  pbuf=CreateRaster1D(Raster->Size1D,Header.BitPerPix);
  data=(BYTE *)pbuf->Data1D;

  Y=0;
  Cnt=0;
  while(Y<Raster->Size2D)
    {
    Aplane=Header.NumPlanes;
    if(Header.NumPlanes==1) data=(BYTE *)Raster->GetRow(Y);

    do {
       adata=data;
       X=0;

       if(Cnt>0) goto RestOfLast;
       while(X<Header.BytesPerLine)
	  {
	  fread(&C,1,1,f);
	  Cnt=1;
	  if((C & Mask)==Mask)	/*the prefix was added*/
		  {
		  Cnt=C & 0x3F;
		  fread(&C,1,1,f);
		  };
RestOfLast:
	  while(Cnt>0)
	     {
	     if(X>=Header.BytesPerLine) break;  /*repeater overflows over a line*/
	     *adata=C;
	     adata++;
	     X++;
	     Cnt--;
	     }
	  }
       if(X!=Header.BytesPerLine)
		{		/*load error*/
		delete Raster;
		Raster=NULL;
		goto ENDPROC;
		}

					/*adding a next bit plane*/
       if(Header.NumPlanes!=1)
	 {
	 if(Header.BitPerPix==1)
	    {
	    Raster->GetRowRaster(Y)->Join1Bit(data,Raster->GetPlanes()-Aplane);
	    }
	 if((Header.BitPerPix==8)&&(Raster->GetPlanes()==24))
	    {		/*RGB image*/
//	    if not aplane in [1,2,3]  asm int 3; };
	    Raster->GetRowRaster(Y)->Join8Bit(data,Raster->GetPlanes()-8*Aplane);
//		JoinB(p.X,addr(p.data^[Y]^.b[3-Aplane]),data,3);
	    }
	  }

      Aplane--;		/*loop for bit rows*/
      } while(Aplane>=1);

//  if(AlineProc!=NULL) AlineProc->NextLine();
  Y++;
  if(Cnt>0) 	//repeater overflows over line}
	{		/*load error*/
	delete Raster;
	Raster=NULL;
	goto ENDPROC;
	}
   }

 if(Header.BitPerPix==8)   //loading of palette
     {
     C=0;
     fread(&C,1,1,f);
     if (C!=12) goto ENDPROC;
     Palette=BuildPalette(256,8);
     if(Palette!=NULL)
	fread(Palette->Data1D,(7+Palette->GetSize()*Palette->GetPlanes())/8,1,f);
     if(GrayPalette(Palette,Raster->GetPlanes()))
		{
		delete Palette;
		Palette=NULL;
		}
     }
  else if (Raster->GetPlanes()==4 || Raster->GetPlanes()==2)  /*move palette from header*/
	{
	Palette=BuildPalette(1<<Raster->GetPlanes(),8);
	if(Palette!=NULL)
	  memcpy(Palette->Data1D,Header.ColorMap,(7+Palette->GetSize()*Palette->GetPlanes())/8);
	if(GrayPalette(Palette,Raster->GetPlanes()))
		{
		delete Palette;
		Palette=NULL;
		}
	}

ENDPROC:
  if(pbuf) delete pbuf;
  fclose(f);
  if(Raster!=NULL) {Img.Raster=Raster;Img.Raster->UsageCount++;}
  if(Palette!=NULL){Img.Palette=Palette;Img.Palette->UsageCount++;}
  return(Img);
}  //LoadPCX
//-------------------End of PCX routines------------------


//------------------EPS--------------------------
int SavePictureEPS(const char *Name,const Image &Img)
{
FILE *f;
int Y;
unsigned char *ptrb,b;
int ColorMode;
long X,BytesPerLine;
const Image *CurrImg=&Img;
float x1,y1,x2,y2;

  if(Img.Raster==NULL) return(-10);

  if((f=fopen(Name,"wt"))==NULL) return(-1);

  CurrImg=Img.Next;	//computation of bounding box
  x1=Img.x;y1=Img.y;
  x2=x1+Img.dx;y2=y1+Img.dy;
  while(CurrImg!=NULL)
    {
    if(x1>CurrImg->x) x1=CurrImg->x;
    if(y1>CurrImg->y) y1=CurrImg->y;
    if(x2<CurrImg->x+CurrImg->dx) x2=CurrImg->x+CurrImg->dx;
    if(y2<CurrImg->y+CurrImg->dy) y2=CurrImg->y+CurrImg->dy;
    CurrImg=CurrImg->Next;
    }
  if(x2<=1e-6) x2=Img.Raster->Size1D;
  if(y2<=1e-6) y2=Img.Raster->Size2D;

  CurrImg=&Img;

  fprintf(f,
"%%!PS-Adobe-2.0 EPSF-2.0\n"
"%%%%Title: %s\n"					//%title
"%%%%Creator: %s - (c) 2000 by Jaroslav Fojtik\n"	//%creator
"%%%%BoundingBox: %2.2f %2.2f %2.2f %2.2f\n"
"%%%%Pages: 1\n"
"%%%%DocumentFonts:\n"
"%%%%EndComments\n"
"%%%%EndProlog\n"
"\n"
"%%%%Page: 1 1",Name,UnitName,
  x1*71/25.4,y1*71/25.4,x2*71/25.4,y2*71/25.4);


  while(CurrImg!=NULL)
    {
    ColorMode=0;
    if(CurrImg->Palette!=NULL) ColorMode=1;
    if(CurrImg->Raster->GetPlanes()==24) ColorMode=2; //&& (p.typ='C'))

    fprintf(f,
"\n"
"%% remember original state\n"
"/origstate save def\n"
"\n"
"%% build a temporary dictionary\n"
"20 dict begin\n"
"\n"
"%% define string to hold a scanline's worth of data\n"
"/pix %d string def\n"
"\n"
"%% define space for color conversions\n"
"/grays %d string def  %% space for gray scale line\n"
"/npixls 0 def\n"
"/rgbindx 0 def\n"
"\n"
"%% lower left corner\n"
"%2.2f %2.2f translate\n"
"\n"
"%% size of image (on paper, in 1/72inch coords)\n"
"%2.2f %2.2f scale\n\n",
    ColorMode!=1?(CurrImg->Raster->Size1D*CurrImg->Raster->GetPlanes()+7)/8:CurrImg->Raster->Size1D,
    CurrImg->Raster->Size1D,
    CurrImg->x*71/25.4,CurrImg->y*71/25.4,
    (CurrImg->dx<1e-6?(float)CurrImg->Raster->Size1D:CurrImg->dx)*71/25.4,
    (CurrImg->dy<1e-6?(float)CurrImg->Raster->Size2D:CurrImg->dy)*71/25.4);

    if(ColorMode)
      fprintf(f,
       "%% define 'colorimage' if it isn't defined\n"
       "%%   ('colortogray' and 'mergeprocs' come from xwd2ps\n"
       "%%     via xgrab)\n"
       "/colorimage where   %% do we know about 'colorimage'?\n"
       "  { pop }           %% yes: pop off the 'dict' returned\n"
       "  {                 %% no:  define one\n"
       "  /colortogray {  %% define an RGB->I function\n"
       "    /rgbdata exch store    %% call input 'rgbdata'\n"
       "    rgbdata length 3 idiv\n"
       "    /npixls exch store\n"
       "    /rgbindx 0 store\n"
       "    0 1 npixls 1 sub {\n"
       "      grays exch\n"
       "      rgbdata rgbindx       get 20 mul    %% Red\n"
       "      rgbdata rgbindx 1 add get 32 mul    %% Green\n"
       "      rgbdata rgbindx 2 add get 12 mul    %% Blue\n"
       "      add add 64 idiv      %% I = .5G + .31R + .18B\n"
       "      put\n"
       "      /rgbindx rgbindx 3 add store\n"
       "    } for\n"
       "    grays 0 npixls getinterval\n"
       "  } bind def\n"
       "\n"
       "%% Utility procedure for colorimage operator.\n"
       "%% This procedure takes two procedures off the\n"
       "%% stack and merges them into a single procedure.\n"
       "\n"
       "/mergeprocs { %% def\n"
       "  dup length\n"
       "  3 -1 roll\n"
       "  dup\n"
       "  length\n"
       "  dup\n"
       "  5 1 roll\n"
       "  3 -1 roll\n"
       "  add\n"
       "  array cvx\n"
       "  dup\n"
       "  3 -1 roll\n"
       "  0 exch\n"
       "  putinterval\n"
       "  dup\n"
       "  4 2 roll\n"
       "  putinterval\n"
       "} bind def\n"
       "/colorimage { %% def\n"
       "  pop pop     %% remove ""false 3"" operands\n"
       " {colortogray} mergeprocs\n"
       "image\n"
       "} bind def\n"
       "} ifelse          %% end of 'false' case\n"
       "\n");

  fprintf(f,
  "%% dimensions of data\n"
  "%d %d %d\n"					//%SizeX %SizeY %Planes
  "\n"
  "%% mapping matrix\n"
  "[ %d 0 0 -%d 0 %d]\n"				//%SizeX %SizeY %SizeY
  "\n"
  "{currentfile pix readhexstring pop}\n"
  "%simage\n",
    CurrImg->Raster->Size1D,CurrImg->Raster->Size2D,
      ColorMode==1?(int)8
		  :ColorMode==2?(int)CurrImg->Raster->GetPlanes()/3:(int)CurrImg->Raster->GetPlanes(),
    CurrImg->Raster->Size1D,CurrImg->Raster->Size2D,CurrImg->Raster->Size2D,
    ColorMode?"false 3 color":"" );


    BytesPerLine=((long)(CurrImg->Raster->GetPlanes())*CurrImg->Raster->Size1D+7) / 8;

    for(Y=0;Y<CurrImg->Raster->Size2D;Y++)
      {
      if(ColorMode==1)
	   {
	   for(X=0;X<CurrImg->Raster->Size1D;X++)
	      {
	      b=CurrImg->Raster->GetValue2D(X,Y);
	      fprintf(f,"%.2X%.2X%.2X",
		  (int)CurrImg->Palette->R(b),
		  (int)CurrImg->Palette->G(b),
		  (int)CurrImg->Palette->B(b));
	      }
	   }
      else {
	   ptrb=(BYTE *)CurrImg->Raster->GetRow(Y);
	   for(X=0;X<BytesPerLine;X++)
	      {
	      fprintf(f,"%.2X",(unsigned)*ptrb++);
	      }
	   }
      fputc('\n',f);
  //    if AlineProc<>nil then AlineProc^.NextLine;
    }

    fprintf(f,
      "%% stop using temporary dictionary\n"
      "end\n"
      "\n"
      "%% restore original state\n"
      "origstate restore\n");

    CurrImg=CurrImg->Next;
    }

  fprintf(f,
      "\n"
      "showpage\n"
      "\n"
      "%%%%Trailer");

  fclose(f);
  return(0);
}
//-------------------End of EPS routines------------------


//------------------TIFF--------------------------
struct TIFHeader
	{
	BYTE CharId[2];
	WORD CharV;
	DWORD Pos;
	};

struct TIFDirectory
	{
	WORD Tag,Field;
	DWORD Long2,Value;
	};


Image LoadPictureTIFF(const char *Name)
{
DWORD *StripsPtrs;
FILE *f;
DWORD delka;
WORD Ldblk;
int i;
TIFHeader Header;
TIFDirectory IFD;
WORD EntryCount,SizeX,SizeY;
/*    :^StripArray;*/
DWORD  PalettePos,NextPtr;
WORD BitsPerSample,SamplesPerPixel;
WORD Compression;
DWORD StripOffsets;
WORD StripCount;
WORD RowsPerStrip,PlanarCfg;
int (*RdWORD_TIFF)( WORD *num, FILE *f );
int (*RdDWORD_TIFF)( DWORD *num, FILE *f );
char *IFD_Marshall=NULL;
Raster2DAbstract *Raster=NULL;
APalette *Palette=NULL;
Image Img;

StripsPtrs=NULL;
PalettePos=0;
BitsPerSample=1;
SamplesPerPixel=1;
Compression=1;
StripOffsets=0;
StripCount=0;
PlanarCfg=0;
RowsPerStrip=65535;

 if((f=fopen(Name,"rb"))==NULL) return(Img);
// Pbuf.Init;
//  p.Erase;
 if(fread(&Header,2,1,f)!=1) {fclose(f);return(Img);};

 RdWORD_TIFF=NULL;RdDWORD_TIFF=NULL;
 if(Header.CharId[0]=='I' && Header.CharId[1]=='I')
	{
	RdWORD_TIFF=RdWORD_LoEnd;
	RdDWORD_TIFF=RdDWORD_LoEnd;
	IFD_Marshall="wwdd";
	}
 if(Header.CharId[0]=='M' && Header.CharId[1]=='M')
	{
	RdWORD_TIFF=RdWORD_HiEnd;
	RdDWORD_TIFF=RdDWORD_HiEnd;
	IFD_Marshall="WWDD";
	}

 if(RdWORD_TIFF==NULL || RdDWORD_TIFF==NULL ) goto FINISH;
 RdWORD_TIFF(&Header.CharV,f);
 RdDWORD_TIFF(&Header.Pos,f);

 if (Header.CharV!=42) goto FINISH; /*check of version*/
 fseek(f,Header.Pos,SEEK_SET);

 RdWORD_TIFF(&EntryCount,f);
 for(i=1;i<=EntryCount;i++)
	{
	loadstruct(f,IFD_Marshall,&IFD.Tag,&IFD.Field,&IFD.Long2,&IFD.Value);

	if (IFD.Field==3 && Header.CharId[0]=='M')
		{
		if(IFD.Long2>=65536) IFD.Long2=IFD.Long2 >> 16;
		if(IFD.Value>=65536) IFD.Value=IFD.Value >> 16;
		}

	switch(IFD.Tag)
	   {
//		255:;				{SubFileType}
	   case 256:SizeX=IFD.Value;break;	//ImageWidth   $0100
	   case	257:SizeY=IFD.Value;break;	//ImageLength  $0101
	   case	258:BitsPerSample=IFD.Value;break; // $102
	   case	259:Compression=IFD.Value;break; //Compression  $0103
//		262:;				{PhotometricInterpretation}
//		266:;				{FillOrder}
	   case	273:StripOffsets=IFD.Value;	//StripOffsets $111
		    StripCount=IFD.Long2;
		    break;
//		274:;				//Orientation
	   case	277:SamplesPerPixel=IFD.Value;break;//SamplesPerPixel
	   case	278:RowsPerStrip=IFD.Value;break;//RowsPerStrip
//		280:;				{MinSampleValue
//		281:;				{MaxSampleValue
//		282:;				{XResolution}
//		283:;				{YResolution}
	   case	284:PlanarCfg=IFD.Value;break;	//PlanarConfiguration
	   case	320:PalettePos=IFD.Value;break; //Palette-color map
	   }
	}

 RdDWORD_TIFF(&NextPtr,f);
 if(StripOffsets==0 || RowsPerStrip==0) goto FINISH;
 if(((SizeY-1) / RowsPerStrip)+1 != StripCount) goto FINISH;

 if(SamplesPerPixel==3)
	{	/*it is a big hack but I really don't know what to do with wrong? BitsPerSample*/
	BitsPerSample=24;
//	asm int 3;
	}

 Raster=CreateRaster2D(SizeX,SizeY,BitsPerSample);
 if(Raster==NULL) goto FINISH;
/* if(AlineProc<>nil)AlineProc^.InitPassing(Raster->Size2D,'Loading TIF');*/

 if( (StripsPtrs=(DWORD *)calloc(StripCount,sizeof(DWORD)))==NULL ) goto FINISH;
 if(StripCount>1)		/*more strips*/
	{
	fseek(f,StripOffsets,SEEK_SET);
	for(i=0;i<StripCount-1;i++)
		RdDWORD_TIFF(&StripsPtrs[i],f);
	fseek(f,StripsPtrs[0],SEEK_SET);
	}
   else {
	if(StripsPtrs) StripsPtrs[0]=StripOffsets;
	fseek(f,StripOffsets,SEEK_SET);
	}


 Ldblk=((long)BitsPerSample*SizeX+7) / 8;
 for(i=0;i<SizeY-1;i++)
	{
	if(i % RowsPerStrip == 0)
		fseek(f,StripsPtrs[i/RowsPerStrip],SEEK_SET);
	if(fread(Raster->GetRow(i),Ldblk,1,f)!=1) {goto FINISH;}
//	if AlineProc<>nil then AlineProc^.NextLine;
	if(feof(f)) break;
	}

 if(PalettePos!=0)		/*load of the palette*/
	{
	fseek(f,PalettePos,SEEK_SET);
	Ldblk=(1 << BitsPerSample);
	Palette=BuildPalette(Ldblk,16);
	for(i=0;i<Ldblk;i++)
		{
		RdWORD_TIFF(&Compression,f);
		Palette->SetValue1D(3*i,Compression);
		}
	for(i=0;i<Ldblk;i++)
		{
		RdWORD_TIFF(&Compression,f);
		Palette->SetValue1D(3*i+1,Compression);
		}
	for(i=0;i<Ldblk;i++)
		{
		RdWORD_TIFF(&Compression,f);
		Palette->SetValue1D(3*i+2,Compression);
		}

	if(GrayPalette(Palette,Raster->GetPlanes()))
		{
		delete Palette;
		Palette=NULL;
		}
	}

FINISH:
 if(f) fclose(f);
 if(StripsPtrs!=NULL) free(StripsPtrs);

 if(Raster!=NULL) {Img.Raster=Raster;Img.Raster->UsageCount++;}
 if(Palette!=NULL){Img.Palette=Palette;Img.Palette->UsageCount++;}
 return(Img);
}  /*LoadTIF*/
//-------------------End of TIFF routines------------------


//------------------WPG--------------------------
struct WPGHeader
	{
	DWORD FileId;
	DWORD DataOffset;
	WORD ProductType;
	WORD FileType;
	BYTE MajorVersion;
	BYTE MinorVersion;
	WORD EncryptKey;
	WORD Reserved;
	};
//structures for WPG level 1
struct WPGRecord
	{
	BYTE	RecType;
	DWORD   RecordLength;
	};

struct WPGBitmapType1
	{
	WORD Width;
	WORD Height;
	WORD Depth;
	WORD HorzRes;
	WORD VertRes;
	};
struct WPGBitmapType2
	{
	WORD RotAngle;
	WORD LowLeftX;
	WORD LowLeftY;
	WORD UpRightX;
	WORD UpRightY;
	WORD Width;
	WORD Height;
	WORD Depth;
	WORD HorzRes;
	WORD VertRes;
	};
struct WPGColorMapRec
	{
	WORD StartIndex;
	WORD NumOfEntries;
	};
//structures for WPG level 2
struct WPG2Record
	{
	BYTE	Class;
	BYTE    Type;
	DWORD   Extension;
	DWORD   RecordLength;
	};

struct WPG2BitmapData
	{
	WORD Width;
	WORD Height;
	BYTE Depth;
	BYTE Compression;
	};
struct WPG2BitmapRectangle
	{
	WORD Flags;
	WORD LowLeftX;
	WORD LowLeftY;
	WORD UpRightX;
	WORD UpRightY;
	WORD HorzRes;
	WORD VertRes;
	};
struct WPG2DblBitmapRectangle
	{
	WORD Flags;
	DWORD LowLeftX;
	DWORD LowLeftY;
	DWORD UpRightX;
	DWORD UpRightY;
	DWORD HorzRes;
	DWORD VertRes;
	};
struct WPG2Start
	{
	WORD	HorizontalUnits;
	WORD	VerticalUnits;
	BYTE    PosSizePrecision;
	};

union WPG_records
	{
	WPGBitmapType1       BitmapType1;
	WPGBitmapType2       BitmapType2;
	WPGColorMapRec       ColorMapRec;

	WPG2BitmapData      _2BitmapData;
	WPG2BitmapRectangle _2BitmapRectangle;
	WPG2DblBitmapRectangle _2DblBitmapRectangle;
	};

inline long LoadWPGHeader(FILE *f,WPGHeader & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,1,sizeof(WPGHeader),f));
#else
return(loadstruct(f,"ddwwbbww",&SU.FileId,&SU.DataOffset,
	&SU.ProductType,&SU.FileType,&SU.MajorVersion,&SU.MinorVersion,&SU.EncryptKey,&SU.Reserved));
#endif
}

inline long LoadWPGBitmapType1(FILE *f,WPGBitmapType1 & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,sizeof(WPGBitmapType1),1,f));
#else
return(loadstruct(f,"wwwww",&SU.Width,&SU.Height,&SU.Depth,&SU.HorzRes,&SU.VertRes));
#endif
}

inline long LoadWPG2BitmapData(FILE *f,WPG2BitmapData & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,sizeof(WPG2BitmapData),1,f));
#else
return(loadstruct(f,"wwbb",&SU.Width,&SU.Height,&SU.Depth,&SU.Compression));
#endif
}

inline long LoadWPG2BitmapRectangle(FILE *f,WPG2BitmapRectangle & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,sizeof(WPG2BitmapRectangle),1,f));
#else
return(loadstruct(f,"wwwwwww",&SU.Flags,&SU.LowLeftX,&SU.LowLeftY,&SU.UpRightX,&SU.UpRightY,
	&SU.HorzRes,&SU.VertRes));
#endif
}

inline long LoadWPG2DblBitmapRectangle(FILE *f,WPG2DblBitmapRectangle & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,sizeof(WPG2DblBitmapRectangle),1,f));
#else
return(loadstruct(f,"wdddddd",&SU.Flags,&SU.LowLeftX,&SU.LowLeftY,&SU.UpRightX,&SU.UpRightY,
	&SU.HorzRes,&SU.VertRes));
#endif
}

inline long LoadWPGBitmapType2(FILE *f,WPGBitmapType2 & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,sizeof(WPGBitmapType2),1,f));
#else
return(loadstruct(f,"wwwwwwwwww",&SU.RotAngle,&SU.LowLeftX,&SU.LowLeftY,&SU.UpRightX,&SU.UpRightY,
	&SU.Width,&SU.Height,&SU.Depth,&SU.HorzRes,&SU.VertRes));
#endif
}

inline long LoadWPGColormap(FILE *f,WPGColorMapRec & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,1,sizeof(WPGColorMapRec),f));
#else
return(loadstruct(f,"ww",&SU.StartIndex,&SU.NumOfEntries));
#endif
}

inline long LoadWPG2Start(FILE *f,WPG2Start & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,1,sizeof(WPG2Start),f));
#else
return(loadstruct(f,"wwb",&SU.HorizontalUnits,&SU.VerticalUnits,&SU.PosSizePrecision));
#endif
}


#define InsertByte(b) \
{\
 *BObrPos = b;BObrPos++;\
 x++;\
 if(x>=ldblk)\
    {\
    x=0;y++;\
    BObrPos=(BYTE *)Raster->GetRow(y>=Raster->Size2D?0:y);\
	/*if(AlineProc) AlineProc->NextLine;*/\
    }\
}

#define InsertRByte(b) \
{\
 *BObrPos = b;BObrPos++;\
 x++;\
 if(x>=ldblk)\
    {\
    x=0;y-=2;\
    BObrPos=(BYTE *)Raster->GetRow((unsigned)y>=Raster->Size2D?0:y);\
    y++;\
	/*if(AlineProc) AlineProc->NextLine;*/\
    }\
}

static void Rd_WP_DWORD(FILE *f, DWORD *d)
{
   unsigned char b;

   fread(&b, sizeof(unsigned char), 1, f);
   *d = b;
   if(b<0xFF) return;

   fread(&b, sizeof(unsigned char), 1, f);
   *d = (DWORD)b ;
   fread(&b, sizeof(unsigned char), 1, f);
   *d += (DWORD)b * 256;
   if(*d<0x8000) return;

   *d = (*d & 0x7FFF) << 16;
   fread(&b, sizeof(unsigned char), 1, f);
   *d += (DWORD)b;
   fread(&b, sizeof(unsigned char), 1, f);
   *d += (DWORD)b * 256l;
   return;
}


int UnpackWPGRaster(Raster2DAbstract *Raster,FILE *f)
{
unsigned x,y,i;
BYTE bbuf,*BObrPos,RunCount;
long ldblk;

if(Raster==NULL) return(-1);  //Not Enough memory
//	     if AlineProc<>nil then AlineProc^.InitPassing(Raster->Size2D,'Loading WPG');

 x=0;
 y=0;
 BObrPos=(BYTE *)Raster->GetRow(0);
 ldblk=((long)Raster->GetPlanes()*Raster->Size1D+7)/8;

 while(y<Raster->Size2D)
     {
     if(fread(&bbuf,1,1,f)!=1)
	{
	delete Raster;
	Raster=NULL;
	return(-2);
	}

     RunCount=bbuf & 0x7F;
     if(bbuf & 0x80)
	{
	if(RunCount)	// repeat next byte runcount *
		{
		fread(&bbuf,1,1,f);
		for(i=0;i<RunCount;i++) InsertByte(bbuf);
		}
	   else {	//read next byte as RunCount; repeat 0xFF runcount*
		fread(&RunCount,1,1,f);
		for(i=0;i<RunCount;i++) InsertByte(0xFF);
		}
	}
   else {
	if(RunCount)   // next runcount byte are readed directly
		{
		for(i=0;i<RunCount;i++)
			{
			fread(&bbuf,1,1,f);
			InsertByte(bbuf);
			};
		}
	   else {	// repeat previous line runcount*
		fread(&RunCount,1,1,f);
		if(x) {
		      return(-3);  // missing code	  asm int 3; }; { not finished !!!!!!!!}
		      }
		for(i=0;i<RunCount;i++)
			{
			x=0;
			y++;
			if(y>=Raster->Size2D) BObrPos=(BYTE *)Raster->GetRow(0);
					      BObrPos=(BYTE *)Raster->GetRow(y);
//			if AlineProc<>nil then AlineProc^.NextLine;
				if(y<2) continue;
			if(y>Raster->Size2D) return(-4);
			memmove(Raster->GetRow(y-1),Raster->GetRow(y-2),ldblk);
			}
                }
        }
   }
 return(0);
}


int UnpackWPG2Raster(Raster2DAbstract *Raster,FILE *f)
{
unsigned x,y,i;
BYTE bbuf,*BObrPos,RunCount;
long ldblk;
unsigned char SampleBuffer[8];
char SampleSize=1;

if(Raster==NULL) return(-1);  //Not Enough memory
//	     if AlineProc<>nil then AlineProc^.InitPassing(Raster->Size2D,'Loading WPG');

 x=0;
 y=Raster->Size2D;
 BObrPos=(BYTE *)Raster->GetRow(y-1);
 ldblk=((long)Raster->GetPlanes()*Raster->Size1D+7)/8;

 while(y>0)
     {
     if(fread(&bbuf,1,1,f)!=1)
	{
	delete Raster;
	Raster=NULL;
	return(-2);
	}

     switch(bbuf)
       {
       case 0x7D:fread(&SampleSize,1,1,f); 		// DSZ
		 if(SampleSize>8) return(-2);
		 if(SampleSize<1) return(-2);
		 break;
       case 0x7E://asm int 3;				// XOR
		 break;
       case 0x7F:fread(&RunCount,1,1,f);		// BLK
		 for(i=0;i<SampleSize*((unsigned)RunCount+1);i++)
			{
			InsertRByte(0);
			}
		 break;
       case 0xFD://asm					// EXT
		 break;
       case 0xFE://asm					// RST
		 break;
       case 0xFF:fread(&RunCount,1,1,f);		// WHT
		 for(i=0;i<SampleSize*((unsigned)RunCount+1);i++)//read next byte as RunCount; repeat 0xFF runcount*
			{
			InsertRByte(0xFF);
			}
		 break;
       default:
	  RunCount=bbuf & 0x7F;

	  if(bbuf & 0x80)
		{	// repeat next byte runcount *
		for(i=0;i<SampleSize;i++)
			fread(&SampleBuffer[i],1,1,f);
		for(i=0;i<=(unsigned)RunCount;i++)
		    for(bbuf=0;bbuf<SampleSize;bbuf++)
			InsertRByte(SampleBuffer[bbuf]);
		}
	   else {
		for(i=0;i<SampleSize*((unsigned)RunCount+1);i++)
			{
			fread(&bbuf,1,1,f);
			InsertRByte(bbuf);
			}
		}
	  }
   }
 return(0);
}


Image LoadPictureWPG(const char *Name)
{
FILE  *f;
long	len,ldblk;
int i;
WPGHeader  Header;
WPGRecord  Rec;
WPG2Record  Rec2;
WPG_records WPG;
WPG2Start StartWPG;
WPGColorMapRec wpPal;
Raster2DAbstract *Raster=NULL;
APalette *Palette=NULL;
Image Img,*CurrImg;
float x,y,dx,dy;


 CurrImg=&Img;

 if((f=fopen(Name,"rb"))==NULL) return(NULL);

 len=LoadWPGHeader(f,Header);
 if(Header.FileId!=0x435057FF || len!=18 ||
    (Header.ProductType>>8)!=0x16 || Header.EncryptKey!=0 )
	 {fclose(f);return(NULL);}

 switch(Header.FileType)
   {
   case 1:
     while(!feof(f))
       {
       fseek(f,Header.DataOffset,SEEK_SET);
       if(feof(f)) break;

       Rec.RecType=fgetc(f);
       Rd_WP_DWORD(f,&Rec.RecordLength);

       Header.DataOffset=ftell(f)+Rec.RecordLength;

       switch(Rec.RecType)
	 {
	 case 0x0B:	// bitmap type 1
		 LoadWPGBitmapType1(f,WPG.BitmapType1);
		 if(Raster!=NULL) break;
		 Raster=CreateRaster2D(WPG.BitmapType1.Width,WPG.BitmapType1.Height,WPG.BitmapType1.Depth);
LoadRaster:	 if(Raster)
		   {
		   if(UnpackWPGRaster(Raster,f)<0) Raster->Erase();
		   if(Img.Raster!=NULL)
		     {
		     CurrImg->Next=new Image;
		     CurrImg=CurrImg->Next;
		     }
		   CurrImg->x=x;
		   CurrImg->y=y;
		   CurrImg->dx=dx;
		   CurrImg->dy=dy;
		   x=y=dx=dy=0;

		   CurrImg->Raster=Raster;
		   CurrImg->Raster->UsageCount++;
		   Raster=NULL;
		   if(Palette!=NULL)
			{
			if(CurrImg->Raster->GetPlanes()==1 &&
			   (Palette->GetValue1D(0)==0 && Palette->GetValue1D(1)==0 && Palette->GetValue1D(2)==0 &&
			    Palette->GetValue1D(3)==0 && Palette->GetValue1D(4)==0 && Palette->GetValue1D(5)==0))
					CurrImg->Palette=NULL;  //reject crippled monochrome palette
			else {CurrImg->Palette=Palette;CurrImg->Palette->UsageCount++;}
			}
		   }
		 break;
	 case 0x0E:	// color palette
		 LoadWPGColormap(f,wpPal);
		 if(Palette!=NULL && Palette->UsageCount==0) delete Palette;
		 Palette=BuildPalette(wpPal.NumOfEntries+wpPal.StartIndex,8);
		 if(Palette!=NULL)
		    {
		    fread((char *)Palette->Data1D + 3*wpPal.StartIndex,
			  3*wpPal.NumOfEntries,1,f);
		    if(GrayPalette(Palette,Raster==NULL?0:Raster->GetPlanes()))
			{
			delete Palette;
			Palette=NULL;
			}
		    }
		 break;

	 case 0x14:	// bitmap type 2
		 LoadWPGBitmapType2(f,WPG.BitmapType2);
		 if(Raster!=NULL) break;
		 Raster=CreateRaster2D(WPG.BitmapType2.Width,WPG.BitmapType2.Height,WPG.BitmapType2.Depth);
		 x=WPG.BitmapType2.LowLeftX/47.0;
		 y=WPG.BitmapType2.LowLeftY/47.0;
		 dx=(WPG.BitmapType2.UpRightX-WPG.BitmapType2.LowLeftX)/47.0;
		 dy=(WPG.BitmapType2.UpRightY-WPG.BitmapType2.LowLeftY)/47.0;
		 goto LoadRaster;
	 }
      }
     break;
   case 2:
     StartWPG.PosSizePrecision=0;
     while(!feof(f))
       {
       fseek(f,Header.DataOffset,SEEK_SET);
       if(feof(f)) break;

       Rec2.Class=fgetc(f);
       Rec2.Type=fgetc(f);
       Rd_WP_DWORD(f,&Rec2.Extension);
       Rd_WP_DWORD(f,&Rec2.RecordLength);

       Header.DataOffset=ftell(f)+Rec2.RecordLength;

       switch(Rec2.Type)
	 {
	 case 0x01:LoadWPG2Start(f,StartWPG);
		   break;
	 case 0x0C:	// color palette
		 LoadWPGColormap(f,wpPal);
		 if(Palette!=NULL && Palette->UsageCount==0) delete Palette;
		 Palette=BuildPalette(wpPal.NumOfEntries+wpPal.StartIndex,8);
		 if(Palette!=NULL)
		    {
		    for(i=wpPal.StartIndex;i<wpPal.NumOfEntries;i++)
			{
			Palette->SetValue1D(3*i,fgetc(f));
			Palette->SetValue1D(3*i+1,fgetc(f));
			Palette->SetValue1D(3*i+2,fgetc(f));
			fgetc(f);
			}
		    if(GrayPalette(Palette,Raster==NULL?0:Raster->GetPlanes()))
			{
			delete Palette;
			Palette=NULL;
			}
		    }
		 break;

	 case 0x0E:	//raster data
	     LoadWPG2BitmapData(f,WPG._2BitmapData);
	     if(WPG._2BitmapData.Compression>1) goto Skip_WP_BMP;
	     switch(WPG._2BitmapData.Depth)
	       {
	       case 1:i=1;break;
	       case 2:i=2;break;
	       case 3:i=4;break;
	       case 4:i=8;break;
	       case 8:i=24;break;
	       default:continue;  /*Ignore raster with unknown depth*/
	       }
	     Raster=CreateRaster2D(WPG._2BitmapData.Width,WPG._2BitmapData.Height,i);
	     if(Raster==NULL) continue;

	     if(WPG._2BitmapData.Compression==1)
		   {
		   if(UnpackWPG2Raster(Raster,f)<0)
				  Raster->Erase();
		   }
	     if(WPG._2BitmapData.Compression==0)
		{
		ldblk=(Raster->GetPlanes()*Raster->Size1D+7)/8;
		for(i=0;i<Raster->Size2D;i++)
		   {
		   if(fread(Raster->GetRow(i),ldblk,1,f)!=1) {goto FINISH;}
	//	   AlineProc(i,p);
		   }
		}
	     if(Img.Raster!=NULL)
		   {
		   CurrImg->Next=new Image;
		   CurrImg=CurrImg->Next;
		   }
	     CurrImg->x=x;
	     CurrImg->y=y;
	     CurrImg->dx=dx;
	     CurrImg->dy=dy;
	     x=y=dx=dy=0;
	     if(Raster!=NULL) {CurrImg->Raster=Raster;CurrImg->Raster->UsageCount++;Raster=NULL;}
	     if(Palette!=NULL){CurrImg->Palette=Palette;CurrImg->Palette->UsageCount++;}
	     break;
	 case 0x1B:	//bitmap rectangle
	     if(StartWPG.PosSizePrecision==0)
			{
			LoadWPG2BitmapRectangle(f,WPG._2BitmapRectangle);
			x=WPG._2BitmapRectangle.LowLeftX/47.0;
			y=WPG._2BitmapRectangle.LowLeftY/47.0;
			dx=(WPG._2BitmapRectangle.UpRightX-WPG._2BitmapRectangle.LowLeftX)/47.0;
			dy=(WPG._2BitmapRectangle.UpRightY-WPG._2BitmapRectangle.LowLeftY)/47.0;
			break;
			}
	     if(StartWPG.PosSizePrecision==1)
			{
			LoadWPG2DblBitmapRectangle(f,WPG._2DblBitmapRectangle);
			x=WPG._2DblBitmapRectangle.LowLeftX/(47.0*0xFFFF);
			y=WPG._2DblBitmapRectangle.LowLeftY/(47.0*0xFFFF);
			dx=(WPG._2DblBitmapRectangle.UpRightX-WPG._2DblBitmapRectangle.LowLeftX)/(47.0*0xFFFF);
			dy=(WPG._2DblBitmapRectangle.UpRightY-WPG._2DblBitmapRectangle.LowLeftY)/(47.0*0xFFFF);
			break;
			}
	     break;
	 }
Skip_WP_BMP: {}
       }
     break;
   }

FINISH:
 if(f) fclose(f);
 if(Palette!=NULL && Palette->UsageCount==0) delete Palette;
 return(Img);
}
//-------------------End of WPG routines------------------

//------------------GIF--------------------------
typedef struct
	{
	char SignatureMain[3];
	char SignatureRelease[3];
	WORD Width,Height;
	BYTE Flag,BackGround,Nul;
	}TypeGIFHeader;
typedef struct
	{
	WORD Left,Top,Width,Height;
	BYTE Flag;
	}TypeGIFImageDescriptor;

inline long LoadGIFHeader(FILE *f,TypeGIFHeader &SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,1,sizeof(SU),f));
#else
return(loadstruct(f,"a3a3wwbbb",
	 &SU.SignatureMain,&SU.SignatureRelease,
	 &SU.Width,&SU.Height,&SU.Flag,&SU.BackGround,&SU.Nul));
#endif
}

inline long LoadGIFImageDescriptor(FILE *f,TypeGIFImageDescriptor &SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,1,sizeof(SU),f));
#else
return(loadstruct(f,"wwwwb",
	 &SU.Left,&SU.Top,&SU.Width,&SU.Height,&SU.Flag));
#endif
}

class ABitStream
	{
public: BYTE *ptr;
	BYTE *RealBufStart,*BufEnd;
	DWORD BitBuf;
	WORD BufSize;
	BYTE BitPos;

	long FileLeft;
	FILE *f;
	BYTE Flag;

	BYTE *_BufEnd;
	BYTE m;
	BYTE n;

	void InitBitStream(void *p,WORD NewBufSize,FILE *file,long fsize);
	WORD Read_N(BYTE BitCount);
	virtual void RdFromFile(void)=0;	// Load bits from GIF file
	};

//This procedure make initialization of the bit stream
void ABitStream::InitBitStream(void *p,WORD NewBufSize,FILE *file,long fsize)
{
//----------------------}
  m=0x01;
  _BufEnd=(BYTE *)p+BufSize-1;
  n=0;
//---------------------}
  FileLeft=fsize;
  BufSize=NewBufSize & ~(WORD)1;
  RealBufStart=(BYTE *)p;
  ptr=(BYTE *)p;
  *RealBufStart=0;
  BufEnd=RealBufStart;
  BitPos=32;
  Flag=0;

  f=file;
}

/*This procedure read n bits from the bit stream*/
WORD ABitStream::Read_N(BYTE BitCount)
{
static const WORD bmsk[17]={0x0,0x1,0x3,0x7,0xF,0x1F,0x3F,0x7F,0xFF,
                     0x1FF,0x3FF,0x7FF,0xFFF,0x1FFF,0x3FFF,0x7FFF,0xFFFF};
WORD u;

 if(BitPos>=16)
	{
	BitBuf=BitBuf >> 16;
	BitPos-=16;
	if(ptr>=BufEnd) RdFromFile();
	BitBuf=BitBuf | ( (long)(*(WORD *)ptr) << 16 );
	ptr+=2;
	}
 u=BitBuf >> BitPos;
 BitPos+=BitCount;
 return(u & bmsk[BitCount]);
}

class GIFBitStream:public ABitStream
	{
public: virtual void RdFromFile(void);	// Load bits from GIF file
	};

void GIFBitStream::RdFromFile(void)		// Load bits from GIF file
{
BYTE len;
WORD ReadedB;

ReadedB=0;
if(Flag==1) *(RealBufStart)=*(BufEnd);

if(FileLeft>0)
	{
	ReadedB=fread(&len,1,1,f);
	if(len==0) len=255;
	if(ReadedB!=1) len=0;

	ptr=RealBufStart;
	if(Flag==1) ptr++;
	ReadedB=fread(ptr,1,len,f);
	FileLeft-=ReadedB;
	}

 if(ReadedB==0) FileLeft=0;		//EOF
 ptr=RealBufStart;
 BufEnd=RealBufStart;
 if(Flag==1) ReadedB++;
 if(ReadedB&1) {Flag=1;ReadedB--;}
	  else Flag=0;
 BufEnd+=ReadedB;

// if IOResult<>0 then FileLeft:=-1; {error}
}


void Add2Pixel(Raster2DAbstract *Raster,WORD Value,WORD &x,WORD &y,int &Interlace)
{
 if(!Raster) return;
 Raster->SetValue2D(x,y,Value);
 x++;
 if(x<Raster->Size1D) return;
 x=0;
// if AlineProc!=NULL   AlineProc^.NextLine;

 if(Interlace<0)
	{
	y++;
	return;
	}

 switch(Interlace)
     {
     case 0:y+=8;
	    if(y>=Raster->Size2D)
	      {
	      Interlace++;
	      y=4;
	      }
	    break;
     case 1:y+=8;
	   if(y>=Raster->Size2D)
	      {
	      Interlace++;
	      y=2;
	      }
	   break;
     case 2:y+=4;
	    if(y>=Raster->Size2D)
	      {
	      Interlace++;
	      y=1;
	      }
	    break;
     case 3:y+=2;
	    break;
     }
}   //..Add2Pixel..}


Image LoadPictureGIF(const char *Name)
{
static const BYTE CodeMask[9]={0,1,3,7,15,31,63,127,255};
BYTE *OutCode=NULL;
WORD *Prefix=NULL,*Suffix=NULL;
WORD Nacteno,F;
FILE *f;
Image Img,*CurrImg;
WORD InCode,OldCode,CurCode,Code,ClearCode,OutCount,FirstFree,FreeCode,EOFCode,MaxCode;
BYTE FinChar,BitMask,CodeSize,InitCodeSize;
WORD X,Y;
int Interlace;
TypeGIFHeader GIFHeader;
TypeGIFImageDescriptor GIFImageDescriptor;
GIFBitStream bit;
void *CompBuffer;
Raster2DAbstract *Raster=NULL;
APalette *Palette=NULL,*LocalPalette=NULL;
const int GIFBufSize=512;
int I;


  CompBuffer=NULL;
  CurrImg=&Img;

  if((f=fopen(Name,"rb"))==NULL) return(NULL);

  if((X=LoadGIFHeader(f,GIFHeader))<13) goto FINISH;
  if(strncmp(GIFHeader.SignatureMain,"GIF",3)) goto FINISH;

  if(GIFHeader.Width==0||GIFHeader.Height==0) goto FINISH;
// if AlineProc!=NULL AlineProc^.InitPassing(Obr.y,'Loading GIF');

//  seek(soubor,Sizeof(GIFHeader));

  if((GIFHeader.Flag & 0x80)>0)  	/*Global Color Map*/
	{
	X=(unsigned)2 << (GIFHeader.Flag & 7);
	Palette=BuildPalette(X,8);
	if(Palette==NULL) goto FINISH;
	if(Palette->Data1D==NULL) goto FINISH;
	X*=3;
	Nacteno=fread(Palette->Data1D,1,X,f);
	if(Nacteno<X) goto FINISH;
	}
  /*?-Global Color Map*/

  if( (OutCode=(BYTE *)malloc(4096))==NULL) goto FINISH;
  if( (CompBuffer=(BYTE *)malloc(GIFBufSize))==NULL) goto FINISH;
  if( (Prefix=(WORD *)calloc(4096,sizeof(WORD)))==NULL) goto FINISH;
  if( (Suffix=(WORD *)calloc(4096,sizeof(WORD)))==NULL) goto FINISH;

  while(!feof(f))
    {
    I=fgetc(f);
//fpos_t pos;fgetpos(f,&pos);printf("%lX,%X\n",pos,I);
    switch(I)
       {
       case EOF:goto FINISH;	//Next Byte

       case 0x21:			   	/*?-Extended Bloks*/
	  if(fread(OutCode,1,2,f)!=2) goto FINISH;
	   //      WriteLn('Extension block type: ',Outcode^[0]);}
	  CodeSize=OutCode[1];
	  while(CodeSize>0)
	     {
	     Nacteno=fread(OutCode,1,CodeSize+1,f);
	     if(Nacteno!=CodeSize+1) goto FINISH;
	     CodeSize=OutCode[CodeSize];
	    }
	  break;

       case 0x2C:
	  if(LoadGIFImageDescriptor(f,GIFImageDescriptor)!=9) goto FINISH;
	  Interlace=-1;
	  if((GIFImageDescriptor.Flag & 64)>0) Interlace=0;
	  /*?-Image Descriptor*/

	  Raster=CreateRaster2D(GIFImageDescriptor.Width,GIFImageDescriptor.Height,
		NearAvailPlanes((((GIFImageDescriptor.Flag & 0x80)?GIFImageDescriptor.Flag:GIFHeader.Flag) & 7) + 1));
	  if(Raster==NULL) goto FINISH;

	  if(GIFImageDescriptor.Flag & 0x80) /*Local Color Map*/
		 {
		 X=(unsigned)2 << (GIFImageDescriptor.Flag & 7);
		 LocalPalette=BuildPalette(X,8);
		 if(LocalPalette==NULL) goto FINISH;
		 if(LocalPalette->Data1D==NULL) goto FINISH;
		 X*=3;
		 Nacteno=fread(LocalPalette->Data1D,1,X,f);
		 if(Nacteno<X) goto FINISH;
		 }

	  X=Y=0;
	  CodeSize=fgetc(f);
	  ClearCode=1 << CodeSize;
	  if(CodeSize==0) goto FINISH;			//GIF error

	  EOFCode=ClearCode+1;
	  FirstFree=ClearCode+2;
	  FreeCode=FirstFree;

	  CodeSize++;
	  InitCodeSize=CodeSize;

	  MaxCode=1 << CodeSize;
	  BitMask=CodeMask[(GIFHeader.Flag & 7)+1];
	  OutCount=0;

	  bit.InitBitStream(CompBuffer,GIFBufSize,f,MAXLONG);

	  do {				/*decompress GIF raster*/
	     if(bit.FileLeft<0)
		 {			//load error
		 goto FINISH;
		 }
	     Code=bit.Read_N(CodeSize);

	     if(Code!=EOFCode)
		{
		if(Code==ClearCode)
		   {
		   CodeSize=InitCodeSize;
		   MaxCode=1 << CodeSize;
		   FreeCode=FirstFree;

		   Code=bit.Read_N(CodeSize);

		   CurCode=Code;
		   OldCode=Code;
		   FinChar=Code & BitMask;
		   Add2Pixel (Raster,FinChar,X,Y,Interlace);
		   }
	       else
		   {
		   CurCode=Code;
		   InCode=Code;
		   if(Code>=FreeCode)
			{
			 CurCode=OldCode;
			 OutCode[OutCount]=FinChar;
			 OutCount++;
			 if(Code>FreeCode)
			     {
			     break;		//Compression Error
			     }
			 }
		   while(CurCode>BitMask)
		       {
			OutCode[OutCount]=Suffix[CurCode];
			OutCount++;
			CurCode=Prefix[CurCode];
		       }
		   FinChar=CurCode & BitMask;
		   OutCode[OutCount]=FinChar;
		   OutCount++;
		   for(I=OutCount-1;I>=0;I--)
			       Add2Pixel (Raster,OutCode[I],X,Y,Interlace);
		   OutCount=0;
		   Prefix[FreeCode]=OldCode;
		   Suffix[FreeCode]=FinChar;
		   OldCode=InCode;
		   FreeCode++;
		   if(FreeCode>=MaxCode)
			{
			 if(CodeSize<12)
			    {
			     CodeSize++;
			     MaxCode=MaxCode*2;
			    }
			    else FreeCode=FirstFree;
			 }
			}
		  }
	   } while(Code!=EOFCode);

	if(Img.Raster!=NULL)
		{
		CurrImg->Next=new Image;
		CurrImg=CurrImg->Next;
		}
	CurrImg->x=GIFImageDescriptor.Left/5.0;
//	CurrImg->y=(GIFHeader.Height-GIFImageDescriptor.Top-GIFImageDescriptor.Height)/5.0;
	CurrImg->y=(GIFImageDescriptor.Height)/5.0;
	CurrImg->dx=GIFImageDescriptor.Width/5.0;
	CurrImg->dy=GIFImageDescriptor.Height/5.0;
	if(Raster!=NULL) {CurrImg->Raster=Raster;CurrImg->Raster->UsageCount++;Raster=NULL;}
	if(LocalPalette) {CurrImg->Palette=LocalPalette;CurrImg->Palette->UsageCount++;LocalPalette=NULL;}
		    else if(Palette!=NULL){CurrImg->Palette=Palette;CurrImg->Palette->UsageCount++;}

	I=fgetc(f);
	if(I==EOF) goto FINISH;
	if(I!=0) ungetc(I,f);
	break;

     default:goto FINISH;
     }
  }

FINISH:
 if(f) fclose(f);

 if(CompBuffer!=NULL) free(CompBuffer);
 if(OutCode!=NULL) free(OutCode);
 if(Prefix!=NULL) free(Prefix);
 if(Suffix!=NULL) free(Suffix);

 if(Raster!=NULL && Raster->UsageCount==0) delete Raster;
 if(Palette!=NULL && Palette->UsageCount==0) delete Palette;
 if(LocalPalette!=NULL && LocalPalette->UsageCount==0) delete LocalPalette;
 return(Img);
} //LoadGIF
//-------------------End of GIF routines------------------


Image LoadPicture(const char *Name)
{
Image Img;
  Img=LoadPictureBMP(Name);
  if(Img.Raster!=NULL) return(Img);
  Img=LoadPictureGIF(Name);
  if(Img.Raster!=NULL) return(Img);
  Img=LoadPicturePCX(Name);
  if(Img.Raster!=NULL) return(Img);
  Img=LoadPictureTIFF(Name);
  if(Img.Raster!=NULL) return(Img);
  Img=LoadPictureWPG(Name);

return(Img);
}


#ifdef RasImg

#include <alloc.h>
void main(void)
{
int x;
Raster1DAbstract AR;
//Raster1D_8Bit AR8,*par8;
//Raster2D_8Bit R2D_8;
Raster2DAbstract R2Da;
Image I,I1,I2;
APalette *pal;


pal=BuildPalette(255);


I1.Create(10,10,8);
I2=CreateRaster2D(10,10,8);

/*for(x=0;x<20;x++) //Memory test
	{
	I=LoadPictureBMP("c:\\temp\\x_duha.bmp");
	SavePictureBMP("c:\\temp\\mRGB.bmp",I.Raster);
	SavePictureBMP("c:\\temp\\m.bmp",I);
//	I.Erase();
	printf("Coreleft:%ld\n",coreleft());
	}*/
printf("%d\n",4/*/*/ */* */-4);

//I=LoadPicturePCX("c:\\temp\\c_duha.pcx");
//I=LoadPictureBMP("c:\\temp\\testimg.bmp");
//I=LoadPictureBMP("c:\\temp\\garfield.bmp");
//I=LoadPictureGIF("c:\\temp\\boris39.gif");
I=LoadPictureGIF("c:\\temp\\mix.gif");

SavePictureBMP("c:\\temp\\mRGB.bmp",I.Raster);
SavePictureBMP("c:\\temp\\m.bmp",I);
SavePictureEPS("c:\\temp\\m.eps",I);

Image I3(I);
//LoadImage(I);

I.SetPixel(5,5,100);
I.SetPixel(8,8,200);
x=I.GetPixel(5,5);
x=I.GetPixel(8,8);
x++;
//if(I.GetPixel(5,5)!=100) abort();

//if((unsigned)I.GetPixel(8,8)!=200) abort();

if(I.Raster)
  printf("Hello Word %d %x",I.Raster->GetPlanes(),255);
}
#endif RasImg
