void AbstractError(void);
unsigned NearAvailPlanes(unsigned n);

class Raster1DAbstract
	{
public: short UsageCount;
	int Size1D;
	void *Data1D;
	char Planes;

	Raster1DAbstract(void) {UsageCount=Planes=Size1D=0;Data1D=NULL;};
	virtual ~Raster1DAbstract() 		{Erase1DStub();};

	virtual int GetPlanes(void) {return(0);};
	virtual int GetSize(void) {return(Size1D);};
	virtual void Allocate1D(int NewSize1D)	{Erase1DStub();};
	virtual void Erase(void)		{Erase1DStub();};
	virtual void Erase1D(void)		{Erase1DStub();};
	void Erase1DStub(void);

	virtual long GetValue1D(int x) {return(0);};
	virtual void SetValue1D(int x,long NewValue) {AbstractError();};

/*This set of functions copy internal data in the proper format to the given buffer.*/
	void Get(Raster1DAbstract &R1);
	virtual void Get1Bit(void *Buffer1Bit);
	virtual void Get2Bit(void *Buffer2Bit);
	virtual void Get4Bit(void *Buffer4Bit);
	virtual void Get8Bit(void *Buffer8Bit);
	virtual void Get16Bit(void *Buffer16Bit);
	virtual void Get24Bit(void *Buffer24Bit) {AbstractError();}
	virtual void Get32Bit(void *Buffer32Bit);

/*This set of functions receives given data and stores them to the internal structure.*/
	void Set(Raster1DAbstract &R1);
	virtual void Set1Bit(void *Buffer1Bit) {AbstractError();}
	virtual void Set2Bit(void *Buffer2Bit) {AbstractError();}
	virtual void Set4Bit(void *Buffer4Bit) {AbstractError();}
	virtual void Set8Bit(void *Buffer8Bit) {AbstractError();}
	virtual void Set16Bit(void *Buffer16Bit) {AbstractError();}
	virtual void Set24Bit(void *Buffer24Bit) {AbstractError();}
	virtual void Set32Bit(void *Buffer32Bit) {AbstractError();}

/*Bitwise operation with raster buffer*/
	virtual void Peel1Bit(void *Buffer1Bit,char plane);
	virtual void Join1Bit(void *Buffer1Bit,char plane);
	virtual void Peel8Bit(void *Buffer8Bit,char plane8) {AbstractError();}
	virtual void Join8Bit(void *Buffer8Bit,char plane8);

	virtual Raster1DAbstract *GetPalette(void) {return(NULL);}
	virtual void SetPalette(Raster1DAbstract *NewPalette) {AbstractError();}
	friend Raster1DAbstract *CreateRaster1D(int SizeX, int Planes);
	};

class Raster2DAbstract:virtual public Raster1DAbstract
	{
public: int Size2D;
	void **Data2D;

	Raster2DAbstract(void) {Size2D=0;Data2D=NULL;}
	virtual ~Raster2DAbstract() 			{Erase2DStub();}

	virtual void Allocate2D(int NewSize1D, int NewSize2D);
	virtual void Erase(void)			{Erase2DStub();}
	virtual void Erase2D(void)			{Erase2DStub();}

	long GetValue2D(int Offset1D,int Offset2D);
	void SetValue2D(int Offset1D,int Offset2D,long x);
	void *GetRow(int Offset2D)   {if(Data2D==NULL || Offset2D>=Size2D) return(NULL);return(Data2D[Offset2D]);}
	Raster1DAbstract *GetRowRaster(int Offset2D);

	void Erase2DStub(void);

	friend Raster2DAbstract *CreateRaster2D(int SizeX, int SizeY, int Planes);
	};

class Raster2DColorAbstract:virtual public Raster1DAbstract
	{
public: Raster1DAbstract *palette;

	virtual Raster1DAbstract *GetPalette(void) {return(palette);}
	virtual void SetPalette(Raster1DAbstract *NewPalette) {if(palette!=NULL) delete(palette); palette=NewPalette;}
	};

class Raster3DAbstract:virtual public Raster2DAbstract
	{
public: int Size3D;
	void ***Data3D;

	Raster3DAbstract(void) {Size3D=0;Data3D=NULL;}
	virtual ~Raster3DAbstract() 			{Erase3DStub();}

	virtual void Allocate3D(int NewSize1D, int NewSize2D, int NewSize3D);
	virtual void Erase(void)			{Erase3DStub();}
	virtual void Erase3D(void)			{Erase3DStub();}

	long GetValue3D(int Offset1D,int Offset2D,int Offset3D);
	void SetValue3D(int Offset1D,int Offset2D,int Offset3D,long x);
	void *GetRow(int Offset3D)   {if(Data3D==NULL || Offset3D>=Size3D) return(NULL);return(Data3D[Offset3D]);}
	Raster2DAbstract *GetRowRaster(int Offset3D);

	void Erase3DStub(void);

	friend Raster3DAbstract *CreateRaster3D(int SizeX, int SizeY, int SizeZ, int Planes);
	};

#ifndef __Rasters_h
#define __Rasters_h



/* ---------- Global class Image --------- */
struct RGBQuad
	{
	unsigned R,G,B,O;
	};

class APalette:virtual public Raster1DAbstract
	{
public:	unsigned R(unsigned index) {return(GetValue1D(3*index));}
	unsigned G(unsigned index) {return(GetValue1D(3*index+1));}
	unsigned B(unsigned index) {return(GetValue1D(3*index+2));}

	unsigned R(unsigned index,unsigned newR) {SetValue1D(3*index,newR);return(newR);};
	unsigned G(unsigned index,unsigned newG) {SetValue1D(3*index+1,newG);return(newG);};
	unsigned B(unsigned index,unsigned newB) {SetValue1D(3*index+2,newB);return(newB);};

	friend APalette *BuildPalette(int Indices, char type=0);
	};


class Image
	{
public: Raster2DAbstract *Raster;
	APalette *Palette;
	Image *Next;
	float x,y,dx,dy;

	Image(void)  {x=y=dx=dy=0;Raster=NULL;Next=NULL;Palette=NULL;};
	Image(const Image & I);
	Image(Raster2DAbstract *NewRaster) {x=y=dx=dy=0;Next=NULL;Palette=NULL;if((Raster=NewRaster)!=NULL) Raster->UsageCount++;};
	~Image(void) {Erase();};

	Raster2DAbstract *operator=(Raster2DAbstract *NewRaster);
	Image &operator=(const Image & I);

	long GetPixel(int SizeX, int SizeY) {return(Raster==NULL?0:Raster->GetValue2D(SizeX,SizeY));};
	void SetPixel(int SizeX, int SizeY, long x) {if(Raster) Raster->SetValue2D(SizeX,SizeY,x);};

	void Create(int SizeX, int SizeY, int Planes);
	void Erase(void);
	};

int GrayPalette(Raster1DAbstract *Palette,int Planes=0);

#endif __Rasters_h