/******************************************************************************
 * program:     wp2latex                                                      *
 * function:    convert Macintosh WordPerfect 3.x files into LaTeX  	      *
 * modul:       pass1.cc                                                      *
 * description: This modul contains functions for first pass. In the first    *
 *              pass information of the WP binary file will splitted in two   *
 *              parts:                                                        *
 *              1) A text file containing the whole text. The special WP      *
 *                 characters are translated.                                 *
 *              2) A binary file which contains information about             *
 *                 environments, tabbings, line endings                       *
 ******************************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <ctype.h>

#include <sets.h>
#include <lists.h>
#include <strings.h>
#include "wp2latex.h"

#ifdef __UNIX__
 #include <unistd.h>
#endif
#ifdef __BORLANDC__
 #include "cp_lib/cptran.h"
#else
 #include "cptran.h"
#endif

#ifndef FINAL
void CrackObject(TconvertedPass1 *cq, DWORD end_of_code);
#endif

/*Register translators here*/
void Convert_pass1_WP3(FILE *FileIn,FILE  *table, FILE *StripOut,FILE *LogFile,FILE *ErrorFile);
FFormatTranslator FormatWP3("WP3.x",&Convert_pass1_WP3);


extern list UserLists;

extern set filter[4];
int set0_3[]={0 ___ 0xFF};		//Everything
int set1_3[]={0 ___ 0xFF}; //Header, Footer, Minipage Text
int set2_3[]={0 ___ 0x7F};  //Characters Only


/*(3,8,10,2,2,4,5,6,0,0,0,0,0,0,0,0);*/
static unsigned char SizesC0[0x10] = {
  5, 8, 7, 4, 0, 0, 0, 0, 0, 0, 0, 0, 7, 9, 7, 0
  };

static unsigned char SizesF0[0x10] = {
  11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


class TconvertedPass1_WP3:public TconvertedPass1
     {
public:CpTranslator *ConvertWP3;
     };


static void Attribute3(TconvertedPass1 *cq)
{
WORD w;

 RdWORD_LoEnd(&w,cq->wpd);

 if(w == 0x8000) Attr_OFF(cq,12); /* End boldface */
 if(w == 0x8001) Attr_OFF(cq,8);  /* End italic */
 if(w == 0x8002) Attr_OFF(cq,14); /* End underline */
 if(w == 0x8012) Attr_OFF(cq,15); /* End SmCaps */

 if(w == 0x0100) Attr_ON(cq,12);  /* Begin boldface */
 if(w == 0x0101) Attr_ON(cq,8);   /* Begin italic */
 if(w == 0x0102) Attr_ON(cq,14);  /* Begin underline */
 if(w == 0x0112) Attr_ON(cq,15);  /* Begin SmCaps */
}


static void Column3(TconvertedPass1 *cq)
{
#ifdef DEBUG
  fprintf(cq->log,"\n#Column3() ");fflush(cq->log);
#endif
BYTE NoColumns;

fseek(cq->wpd, 2L, SEEK_CUR);	//skip <flags>,[non del size]
NoColumns=getc(cq->wpd);

Column(cq,NoColumns);

sprintf(cq->ObjType, "Col Def:%d",(int)NoColumns);
}



static void ExtendedCharacter3(TconvertedPass1 *cq)
{
#ifdef DEBUG
  fprintf(cq->log,"\n#ExtendedCharacter3() ");fflush(cq->log);
#endif
  unsigned char chr_code, chr_set;
  char *ch;

  fseek(cq->wpd,1l,SEEK_CUR);
  fread(&chr_set, 1, 1, cq->wpd);
  fread(&chr_code, 1, 1, cq->wpd);

  sprintf(cq->ObjType, "%d,%d",chr_set,chr_code);
  ch=Ext_chr_str(chr_set, chr_code, cq);
  CharacterStr(cq,ch);
}



static boolean CheckConzistency3(TconvertedPass1_WP3 *cq, long NewPos)
{
#ifdef DEBUG
  fprintf(cq->log,"\n#CheckConzistency() ");fflush(cq->log);
#endif

  boolean Result = true;
  unsigned char TestedBy,TestedSubBy;
  long Pos;

  Pos = ftell(cq->wpd);

  TestedSubBy=cq->subby;
  if (cq->by>=0xd0 && cq->by<0xF0)
		 {                              //subby will be also tested
                 fseek(cq->wpd, NewPos-2 , 0);
		 fread(&TestedSubBy, 1, 1, cq->wpd);
		 }
	    else fseek(cq->wpd, NewPos-1 , 0);
  fread(&TestedBy, 1, 1, cq->wpd);
  if ((TestedBy != cq->by)||(TestedSubBy != cq->subby))
  	{
        if (cq->err != NULL)
          {
//        if(CorruptedObjects<19)
	    cq->perc.Hide();
            fprintf(cq->err,
	      _("\nError: Object %lX:%X consistency check failed. Trying to ignore."),cq->ActualPos,(int)cq->by);
/*        if(CorruptedObjects==19)
	    fprintf(cq->err,
	      _("\nError: Overwhelmed by too many currupted objects, switching to the silent mode.);*/
          }
        CorruptedObjects++;
        Result = false;
	}

  fseek(cq->wpd, Pos, 0);
  return Result;
}



/*This is main procedure for processing one key. It is recursivelly called.*/
void ProcessKeyWP3(TconvertedPass1_WP3 *cq)
{
#ifdef DEBUG
  fprintf(cq->log,"\n#ProcessKey() ");fflush(cq->log);
#endif
  WORD w;
  DWORD NewPos = 0;
  unsigned char by, subby;

  if (cq->by == 0)
    fread(&cq->by, 1, 1, cq->wpd);

  *cq->ObjType = '\0';
  w=1;
  cq->subby=0;


  /*Guessing end position of the object*/

  if (cq->by > 0xbf)
      {
      if (cq->by >= 0xC0 && cq->by <= 0xCF)
	  {	/*these functions has fixed size - see table SizesC0*/
	  if (SizesC0[cq->by - 0xc0] != 0)
		{
		w=SizesC0[cq->by- 0xc0];
		NewPos = w + cq->ActualPos;
		}
//	  else printf("!!%d",cq->by - 0xc0);
	  cq->subby = 0xff;   /*ignored subcommand*/
	  }
      if (cq->by >= 0xF0)
	  {	/*these functions has fixed size - see table SizesF0*/
	  if (SizesF0[cq->by - 0xF0] != 0)
		{
		w=SizesF0[cq->by - 0xF0];
		NewPos = w + cq->ActualPos;
		}
	  cq->subby = 0xff;   /*ignored subcommand*/
	  }
      else if (cq->by >= 0xd0)
	 {
	 fread(&cq->subby, 1, 1, cq->wpd);
	 RdWORD_HiEnd(&w,cq->wpd);
	 NewPos = cq->ActualPos + w + 4;
	 cq->Linebegin = false;
	 }
      }	/*other functions has size only 1 byte - all OK*/
      
  by = cq->by;
  subby = cq->subby;

  if (ExtendedCheck && NewPos != 0)
    if (!CheckConzistency3(cq, NewPos))
      {
      NewPos = cq->ActualPos+1;
      strcpy(cq->ObjType, "Corrupted!!");
      goto _LObjectError;
      }

/*  if((by==0xd2)&&(subby==0xb))
               fprintf(cq->strip, " ****Start Of Table:%d****",cq->flag);*//**/

  if( filter[cq->flag][cq->by] )
     {
     switch (by)
	{
	case 0x0a:HardReturn(cq);	break;		// Hard return
	case 0x0b:Terminate_Line(cq,'p');strcpy(cq->ObjType, "SRt SoftPg");break;// Soft page
	case 0x0c:HardPage(cq);		break;	// Hard page
	case 0x0d:SoftReturn(cq);	break;		// Soft return
	case 0x20:putc(' ', cq->strip);   break;	/*soft space ' '*/

	case 0x96:HardHyphen(cq);break;			/*Hyphen*/

	case 0xA8:CenterPage(cq);break;

	case 0xC0:ExtendedCharacter3(cq);break;
	case 0xC1:Flush_right(cq,0);break;

	case 0xC3:Attribute3(cq);break;

	case 0xD0:if(subby == 6) Justification(cq,3);
		  break;
	case 0xD1:if(subby == 0) Color(cq,3);
		  break;

	case 0xD2:if(subby == 1) Column3(cq);
		  break;

	case 0xD7:if(subby == 0) StartSection(cq,3);
		  if(subby == 1) EndSection(cq,3);
		  break;

	case 0xDC:if(subby==0) SoftReturn(cq);
		  if(subby==2 || subby==3) HardReturn(cq);
		  if(subby==0x16) CellTable(cq);
		  if(subby==0x18) RowTable(cq);
		  if(subby==0x1A) EndTable(cq);
		  break;

	default:if (by >= 0x03 && by <= 0x7f)
		     {
		     cq->RequiredFont=1;
		     CharacterStr(cq,lat[cq->by]);   /* Normal_char  */
		     }
		break;
	}
  }

_LObjectError:
  if (cq->log != NULL)
    {   /**/
    if (by==0xC0) fprintf(cq->log, " [ExtChr %s] ", cq->ObjType);
    else if ((by > 0x80)||(*cq->ObjType != '\0'))
	{
	fprintf(cq->log, _("\n%*sObject type:%3X subtype:%3X length:%4u"),
		  cq->recursion * 2, "", by, subby, w);
	if (*cq->ObjType != '\0')
		  fprintf(cq->log, " [%s] ", cq->ObjType);
	     else fprintf(cq->log, "    ");
	if(*cq->ObjType==0) UnknownObjects++;
	}
    else if (by >= ' ' && by <= 'z')
	         putc(by, cq->log);
    }

  if (NewPos == 0)
	{
	if(by<0xC0) cq->ActualPos++;	//Only one byte read - simple guess of new position
	       else cq->ActualPos = ftell(cq->wpd);
	return;
	}
  cq->ActualPos = ftell(cq->wpd);
  if (NewPos == cq->ActualPos) return;
  fseek(cq->wpd, NewPos, 0);
  cq->ActualPos = NewPos;
  NewPos = 0;
}

static void InitFilter3(void)
{
 filter[0]=set(set0_3,sizeof(set0_3)/sizeof(int));
 filter[1]=set(set1_3,sizeof(set1_3)/sizeof(int));
 filter[2]=set(set2_3,sizeof(set2_3)/sizeof(int));
 filter[3]=set();
}


/*******************************************************************/
/* This procedure provides all needed processing for the first pass*/
void Convert_pass1_WP3(FILE *FileIn,FILE  *table, FILE *StripOut,FILE *LogFile,FILE *ErrorFile)
{
#ifdef DEBUG
  fprintf(LogFile,"\n#Convert_pass1_WP5() ");fflush(LogFile);
#endif
  DWORD fsize;
  TconvertedPass1_WP3 cq;


  InitFilter3();
  Table_Init(lat);

  InitAttr(cq.attr);
  cq.ConvertWP3=NULL;
//  if(WPcharset == 2) cq.ConvertWP3 = GetTranslator("wp5_czTOwp5");

  cq.wpd   = FileIn;
  cq.table = table;
  cq.strip = StripOut;
  cq.log   = LogFile;
  cq.err   = ErrorFile;

  WP_Default(&cq);
//  cq.DefColumns=2;

  printf(_("WARNING: converting from WP 3.x from Macintosh is highly experimetnal!\n"));

  cq.DocumentStart=ftell(cq.wpd);

  fsize=filesize(cq.wpd);
  cq.perc.Init(ftell(cq.wpd), fsize,_("First pass WP 3.x:") );


  if(cq.DocumentStart>0x10)	//This can be omitted
    {
    fseek(cq.wpd,cq.DocumentStart,SEEK_SET);
    }


  cq.envir = ' ';
  cq.nomore_valid_tabs = false;
  cq.rownum = 0;

  Make_tableentry_attr(&cq);   /* attribuut instelling */
  Make_tableentry_tabset(&cq);   /* Geef the defaulttabinstelling door */
  /* tied the 2e pass */

  cq.ActualPos = ftell(cq.wpd);
  while (cq.ActualPos < fsize)
      {
      if(Verbosing >= 1)		//actualise a procentage counter
	      cq.perc.Actualise(cq.ActualPos);

      cq.by = 0;
      ProcessKeyWP3(&cq);
      }
  Close_All_Attr(cq.attr,cq.strip);

  cq.line_term = 's';			   /* Soft return */
  Make_tableentry_envir_extra_end(&cq);
  if(cq.CentrePage)    /* finish page centering */
  	{
	fprintf(cq.strip, "\\vfil ");
        cq.CentrePage=false;
	}

  num_of_lines_stripfile = cq.rownum;

  if(Verbosing >= 1)		//finishing a procentage counter
     {
     cq.perc.SetPercent(100);
     putchar('\n');
     }
}


/*--------------------End of PASS1_WP3--------------------*/
