/******************************************************************************
 * program:     wp2latex                                                      *
 * function:    convert WordPerfect 3.x,4.x,5.x and 6.x files into LaTeX      *
 * modul:       wp2l_lib.cc                                                   *
 * description: Library with commonly used procedures.			      *
 ******************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//Non standard include files
#include <common.h>
#include "wp2latex.h"
#include <strings.h>
#include <lists.h>

#ifdef __UNIX__
 #include <unistd.h>
 #include <signal.h>
#endif

#ifdef __Disable_NULL_printf
 #define chk(x) x==NULL?"NULL":x
#else
 #define chk(x) x
#endif


/*           Version 1.0  Date 27-1-1990
	     Version 2.0  Date 7-12-1996
		     2.1  Date 19-7-1997
		     2.15 Date 17-11-1997
		     2.19 Date 1-3-1998
		     2.20 Date 14-3-1998
		     2.84 Date 31.6.2000
*/

/*
Flags:'h';       Hard return
      's';       Soft return      !!!!!!! 'r'
      'P';       Hard page
      'p';       Soft page

      'B'        Table
      'C','c'    Center (only one line)
      'I','i'    Identing (both)
      'L'        Left
      'M' [cols] Multicolumn enviroment
      'R'        Right
      'Q','q'    Equation (array of equations)
      'T'        Tabbing

      'l' size			Line Spacing
      'S' number,[tab1,tab2]	Tabset

      '%'		        Empty line that might be removed
      '!'			Ignore change Enviroment before (push)
      '^'			Ignore change Enviroment after (pop)
      #FF			end of line
*/

/*WP2LaTex switches: -1:disabled; 0:automatically enabled; 1: forced to enable*/
SBYTE   Cyrilic=false,
	colors=false,			/* Only for LaTex 2.0e */
	textcomp=-1,
	EndNotes=false,
	FancyHdr=-1,
	Columns=1,	/*-1 disabled; 0 auto two; 1 auto multi; 2 forced two; >2 forced multi*/
	InputPS=0,	/*-1 disabled; 0 auto InputPS.sty; 1 InputPS.sty; 2 auto graphicx.sty; 3 raphicx.sty; 4 auto epsfig.sty; 5 epsfig.sty;*/
	PostScript=false,
	MakeIdx=-1,
	Wasy=-1,	/*Totaly Disabled*/
	Amssymb=-1,	/*Totaly Disabled*/
	LaTeXsym=0,
	Bbm=-1,
	fontenc=-1,
	Rsfs=-1,
	Ulem=-1,
	LongTable=false,
	Czech=false,
	German=false;
boolean Index=false,
	NFSS=false,
	Tabs = true,
	twoside=false,
	SaveWPG=false,
	ExactFnNumbers=false,
	CPStyles=false;

int WPVersion=0x501,
    LaTeX_Version=0x300; /* 0x209=LaTeX 2.09; 0x300=LaTex 3.0 (or 2e) */

unsigned short CodePage, WPcharset;


void (*Convert_first_pass)(FILE *FileIn,FILE  *table, FILE *StripOut,FILE *LogFile,FILE *ErrorFile);
char *DocumentType="Unknown type";
//char *DocumentShortType="";

string & fGets2(FILE *f, string & pstr);

boolean OptimizeSection;
unsigned char Verbosing;
boolean TexChars;
boolean ExtendedCheck,FixSpaces;
boolean EraseStrip=true;
char *InputDir="./",*OutputDir="./",*RelativeFigDir=".";
list ImgInputDirs="./";
list Counters;
list FontTable;
list UserLists;

long  num_of_lines_stripfile,
      UnknownCharacters,UnknownObjects,CorruptedObjects;
LatTable lat;
char** UserWPCharSet=NULL;

string wpd_filename, strip_filename, table_filename, latex_filename,
       log_filename;
string wpd_password;
FILE *wpd, *table;
FILE *strip, *logg, *err;
FILE *latex;

char *Open_com[26] = {
  "{\\LARGE ", "{\\Large ", "{\\large ", "{\\small ", "{\\footnotesize ",
  "$^{{\\rm ", "$_{{\\rm ", "{\\fontshape{ol}\\selectfont ", "{\\it ", "", "\\textcolor{red}{", "\\uuline{",
  "{\\bf ", "\\sout{", "\\uline{", "{\\sc ",
  "\\part{", "\\chapter{", "\\section{", "\\subsection{", "\\subsubsection{",
  "\\paragraph{", "\\subparagraph{","{","{","\\iffalse "
};

char *Close_com[26] = {
  "}", "}", "}", "}", "}", "}}$", "}}$", "}", "\\/}", "", "}", "}", "}", "}",
  "}", "}", "}", "}", "}", "}", "}", "}", "}", "}", "}","\\fi "
};

int OutputStyle=First_com_section + 2;

#if defined(__BORLANDC__) && defined(__MSDOS__)
 #include <alloc.h>
 #define CORELEFT() coreleft()
#endif


void UpdateFigDir(void)
{
char ch;
int i;

if(wpd_filename=="") return;

if((i=GetPath(wpd_filename())) > 0)
   {
   ch=wpd_filename[++i];
   wpd_filename[i]=0;
   InputDir=strdup(wpd_filename);
   ImgInputDirs|=InputDir;
   wpd_filename[i]=ch;
   }

if(latex_filename == '-') return;

if((i=GetPath(latex_filename())) > 0)
	{
	ch=latex_filename[++i];
	latex_filename[i]=0;
	if(!strcmp(latex_filename,InputDir)) OutputDir=InputDir;
			   else	OutputDir=strdup(latex_filename);
	latex_filename[i]=ch;
	}

}


void Sys_Variable()
{
string var;
char **argv;
int argc,i;
char typ,oldtyp;

argc=0;

var=getenv("wp2latex");
if(var=="") var=getenv("WP2LATEX");
if(var=="") return;

oldtyp=0;
for(i=0;i<length(var);i++)
	{
        if((var[i]==' ')||(var[i]==9))
			{
			typ=0;
			var[i]=0;
                        }
        	   else typ=1;

        if(typ==1 && oldtyp==0) argc++;
        oldtyp=typ;
        }

if((argv=(char **)calloc(argc,sizeof(char**))) == NULL) return;
argc=0;
if(var[0]!=0) argv[argc++]=var();
for(i=1;i<length(var);i++)
	{
	if((var[i-1]==0)&&(var[i]!=0))
			 argv[argc++]=&var[i];
	}
CommandLine(argc, argv);
free(argv);
return;
}


/*This procedure is aimed to parse a configuration file*/
void LoadConfigFile(const char *filename)
{
FILE *F;
string s;
int Section=0,i;

if( (F=fopen(filename,"r"))==NULL ) return;
while(!feof(F))
   {
   fGets2(F,s);
   s=cutspaces(s);
   if(s=="") continue;
   if(s=="[WPUserChars]")
	{
	Section=1;
	continue;
	}
   if(s=="[Fonts]")
	{
	Section=2;
	continue;
	}

   switch(Section)
	{
	case 0:break;
	case 1:if(sscanf(s(),"%d",&i)!=1) break;
	       if(i<0 || i>255) break;
	       while(isdigit(s[0]))
		   {
		   s=copy(s,1,length(s)-1);
		   if(s=="") break;
		   }
	       s=cutspaces(s);
	       if(UserWPCharSet==NULL)
		   {
		   UserWPCharSet=(char **)calloc(255,sizeof(char *));
		   if(UserWPCharSet==NULL) break;
		   memset(UserWPCharSet,0,255*sizeof(char *));
		   }

	       if(UserWPCharSet[i]!=NULL) free(UserWPCharSet[i]);
	       UserWPCharSet[i]=strdup(s());
	       break;
	case 2:FontTable+=s;
	       break;
	}
   }

fclose(F);
}


//This is a state machine for detecting HTML files
#define MaxChar 0x200   //this can be used to tune a quality of autodetection
#define MaxChar 0x200
static int DetectHtml(FILE *text)
{
int Status,CharNo=0,chr;

  if(text==NULL) return(0);

  Status=0;
  fseek(text,0,SEEK_SET);
  while(CharNo<MaxChar)
	{
	if( (chr=fgetc(text)) == EOF) goto NoHTML;
        chr=toupper(chr);
	CharNo++;
	switch(Status)
		{
		case 0:if(isspace(chr)) break;
		       if(chr=='<') {Status=1;break;}
		       goto NoHTML;
		case 1:if(chr=='!') {Status=6;break;}
		       if(chr=='H') {Status=2;break;}
                       if(chr=='M') {Status=7;break;}
		       goto NoHTML;
		case 2:if(chr=='T') {Status=3;break;}
		       goto NoHTML;
		case 3:if(chr=='M') {Status=4;break;}
		       goto NoHTML;
		case 4:if(chr=='L') {Status=5;break;}
		       goto NoHTML;
		case 5:if(chr=='>')
			    {
			    fseek(text,0,SEEK_SET);
			    return(1);		//HTML succesfully detected
			    }
		       goto NoHTML;
		case 6:if(chr=='>') Status=0;
		       break;
		case 7:if(chr=='E') {Status=8;break;}
		       goto NoHTML;
		case 8:if(chr=='T') {Status=9;break;}
		       goto NoHTML;
		case 9:if(chr=='A') {Status=10;break;}
		       goto NoHTML;
                case 10:if(chr==' '){Status=6;break;}
                       if(chr=='>') {Status=0;break;}
		       goto NoHTML;
		default:printf("Chyba: Status automat failed at state %d\n",Status);
			Status=0;
		}

	}

NoHTML:
  fseek(text,0,SEEK_SET);
  return(0);
}


/*This procedure provides autodetection of RTF format*/
static int DetectRTF(FILE *text)
{
long OldFilePos;
char buffer[6];

  if(text==NULL) return(0);

  OldFilePos=ftell(text);
  fseek(text, 0L, SEEK_SET);

  if(fread(buffer,1,5,text)!=5) goto NoRTF;
  if(!strncmp(buffer,"{\\rtf",5))		//RTF startup detected
	{
	return(1);
	}

NoRTF:
  fseek(text,OldFilePos,SEEK_SET);
  return(0);
}


/*This procedure provides autodetection of RTF format*/
static int DetectT602(FILE *text)
{
long OldFilePos;
string Command;
int commands=0;

  if(text==NULL) return(0);
  OldFilePos=ftell(text);
  fseek(text, 0L, SEEK_SET);


  while(!feof(text))
    {
    fGets2(text, Command);
    if(Command[0]!='@')
	{
	if(commands<3) goto NoT602;
	fseek(text,0,SEEK_SET);
	return(1);
	}

    commands++;
    }


NoT602:
  fseek(text,OldFilePos,SEEK_SET);
  return(0);
}




void Convert_pass1_WP3(FILE *FileIn,FILE  *table, FILE *StripOut,FILE *LogFile,FILE *ErrorFile);
/* Checks whether the given document is a valid  WP5.x or WP6.x - document.
 * return: int: 0 if OK, else error code  */
int Wpd_check(FILE *wpd,FILE *err)
{
  WORD id1, id2, filetype, ver, encryption, dmp3;
  DWORD StartDoc;

  if(wpd==NULL) return(0);

  Rd_word(wpd, &id1);
  Rd_word(wpd, &id2);

//*****Strong tests*****
  if (id1 == 0x57ff && id2 == 0x4350)
    {
    Rd_dword(wpd,&StartDoc);
    Rd_word(wpd, &filetype);
    Rd_word(wpd, &ver);
    Rd_word(wpd, &encryption);
    Rd_word(wpd, &dmp3);

    switch(filetype>>8)
	{
	case  0xA:DocumentType="WordPerfect Document";break;
	case  0xB:DocumentType="Dictionary file";break;
	case  0xC:DocumentType="Thesaurus file";break;
	case  0xD:DocumentType="Block";
		  RunError(0x15);
		  break;
	case  0xE:DocumentType="Rectangular Block";
		  RunError(0x15);
		  break;
	case  0xF:DocumentType="Column Block";
		  RunError(0x15);
		  break;
	case 0x10:DocumentType="Printer Resource file (.PRS)";break;
	case 0x11:DocumentType="Setup file";break;
	case 0x12:DocumentType="Reserved";break;
	case 0x13:DocumentType="Printer Resource file (.ALL)";break;
	case 0x14:DocumentType="Display Resource file (.DRS)";break;
	case 0x15:DocumentType="Overlay file";break;
	case 0x16:DocumentType="WordPerfect Graphics file";break;
	case 0x17:DocumentType="Hypenation code module";break;
	case 0x18:DocumentType="Hypenation data module";break;
	case 0x19:DocumentType="Macro Resource File";break;
	case 0x1A:DocumentType="Graphics screen Driver";break;
	case 0x1B:DocumentType="Hypenation lex module";break;
	case 0x1C:DocumentType="Printer Q codes (VAX/DG)";
		  RunError(0x15);
		  break;
	case 0x1D:DocumentType="Spell code module word list";break;
	case 0x1E:DocumentType="WP5.1 equation resource file";break;
	case 0x1F:DocumentType="Reserved";break;
	case 0x20:DocumentType="VAX.SET";break;
	case 0x21:DocumentType="Spell code module rules";break;
	case 0x22:DocumentType="Dictionary rules";break;
	case 0x23:DocumentType="Reserved";break;
	case 0x24:DocumentType=".WPD files";
		  RunError(0x15);
		  break;
	case 0x25:DocumentType="Rhymer word file";break;
	case 0x26:DocumentType="Rhymer word file";break;
	case 0x27:
	case 0x28:DocumentType="Reserved";break;
	case 0x29:DocumentType="WP5.1 ins file";
		  RunError(0x15);
		  break;
	case 0x2A:DocumentType="Mouse driver for WP5.1";break;
	case 0x2B:DocumentType="Unix setup for WP5.0";break;
	case 0x2C:DocumentType="Macintosh WP2.0 document";break;
	case 0x2D:DocumentType="VAX file WP4.0 document";
		  RunError(0x15);
		  break;
	case 0x2E:DocumentType="External spell code";break;
	case 0x2F:DocumentType="External spell dictionary";break;
	case 0x30:DocumentType="Macintosh soft graphics";break;

	case 0x38:DocumentType="WPWin5.1 App Resource library";break;
	}


    if(filetype>>8 == 0x2C)
	{
	if((ver & 0xFF) == 3 )	/*   version == 3 -> WP 3.x  */
	   {
	   Convert_first_pass=GetConvertor("WP3.x");//Convert_pass1_WP3;
	   fseek(wpd, StartDoc, SEEK_SET);

	   WPVersion=256*(ver&0xFF) | ver>>8;
	   if(encryption!=0) fprintf(err,_("\nWarning: File seems to be encrypted. Conversion may not be possible!!"));
	   return(3);
	   }
	}

    if (filetype == 0xa01)
	   /*   version == 0x0000 -> WP 5.0
	    *   version == 0x0100 -> WP 5.1
	    *   version == 0x0002 -> WP 6.0
	    *   version == 0x0102 -> WP 6.1
	    *   version == 0x0202 -> WP 7.0  */
       {
       fseek(wpd, StartDoc, SEEK_SET);
       switch(ver & 0xFF)  // major version
	   {
	   case 0:WPVersion=0x500 + (WORD)(ver>>8);
		  if(encryption==0) Convert_first_pass=GetConvertor("WP5.x");
			       else Convert_first_pass=GetConvertor("WP5.x encrypted");
		  break;
	   case 2:WPVersion=0x600 + (WORD)(ver>>8);
		  if(WPVersion>=0x602) WPVersion+=0xFE;
		  if(encryption==0) Convert_first_pass=GetConvertor("WP6.x");
			       else Convert_first_pass=GetConvertor("WP6.x encrypted");
		  break;
	   default:WPVersion=0x500l + (WORD)(ver>>8) + 256*(ver & 0xFF);
		   Convert_first_pass=GetConvertor("WP6.x");
		   if(Convert_first_pass==NULL) Convert_first_pass=GetConvertor("WP5.x");

		   fprintf(err,_("\nWarning: Unknown WP version 0x%x. Guessing WP%d.%d. EXPECT ERRORS!!!"),ver,WPVersion>>8,WPVersion & 0xFF);
		   encryption=0;
	   }

	if(encryption!=0) fprintf(err,_("\nWarning: File seems to be encrypted. Conversion may not be possible!!"));
	}
    return 5;     /* it is a valid WP [5,6].x - document*/
    }

  if (id1 == 0xFFFE && id2 == 0x6161)	//This is a WP4.1 encrypted document
	{
	WPVersion=0x401;
	Convert_first_pass=GetConvertor("WP4.x encrypted");//Convert_pass1_WP4_encrypted;
	return 4;
	}


  if (id1 == 0xFEFF)	//This is an unicode text document
	{
	fseek(wpd, 2L, SEEK_SET);
	Convert_first_pass=GetConvertor("UNICODE");
	return 130;
	}


//++++Weak Tests++++

  if (id1 == 0x2125 && id2 == 0x5350)	//This is a Postscript file
	{
	fseek(wpd, 0L, SEEK_SET);
	Convert_first_pass=GetConvertor("Postscript");
	DocumentType="Postscript file";
	return(100);
	}

  if(DetectHtml(wpd))
	{
	fseek(wpd, 0L, SEEK_SET);
	DocumentType="HTML file";
	Convert_first_pass=GetConvertor("HTML");
	return(110);
	}

  if(DetectRTF(wpd))
	{
	fseek(wpd, 0L, SEEK_SET);
	DocumentType="RTF file";
	Convert_first_pass=GetConvertor("RTF");
	return(120);
	}

  if(DetectT602(wpd))
	{
	fseek(wpd, 0L, SEEK_SET);
	DocumentType="T602 file";
	Convert_first_pass=GetConvertor("T602");
	return(120);
	}


//-----Forced only-----
#ifdef WP4x
  if(WPVersion==0x400)
	{
	fprintf(err,_("\nWarning: WP v 4.0 is forced. No autodetection is possible. ERRORS may occur!!!"));
	Convert_first_pass=GetConvertor("WP4.x");
	fseek(wpd, 0L, SEEK_SET);
	ExtendedCheck = true;	/* It would be better to switch safe mode on. */
	return(4);
	}
#endif

  RunError(0x201);     /* it may be a valid WP 4.x document - but error is reported*/
  return 0;
}


/*This function computes checksum for the given password*/
WORD CheckSum(char *PassWord)
{
WORD csum;

csum=0;
if(PassWord==NULL) return(0);

while(*PassWord!=0)
	{
	if( *PassWord>='a' && *PassWord<='z')    /* convert to upper case */
			*PassWord -= 'a'-'A';

	csum = ( (csum >> 1) | ( csum << 15) ) ^ ( ((WORD)*PassWord)<<8 );
	PassWord++;
	}
return(csum);
}


void DecryptWP5(FILE *Encrypted,FILE *Decrypted,const string & PassWord,FILE *ErrorFile)
{
char *buf;
WORD idx;                     /* index into buffer */
WORD pidx;                    /* index down password */
BYTE xmask;                   /* XOR mask */
DWORD len;
DWORD pos=ftell(Encrypted);
DWORD fsize=filesize(Encrypted);

 if(ErrorFile==NULL) ErrorFile=stderr;
 if((buf=(char *)malloc(4096))==NULL) return;


 pidx = 0;                       /* start of password */
 xmask = length(wpd_password)+1; /* start at password length+1 */

 if (Verbosing >= 1) printf(_("\nDecrypting ...\n\n"));

 while( pos<fsize )
	 {
	 if( fsize-pos > 4096 ) len = 4096;
			   else len = fsize-pos;

	 if( fread( buf, 1, len, Encrypted ) !=len )
		 {
		 fprintf(ErrorFile,_("\nError: Unexpected end of the input file."));
		 break;
		 }

	  for( idx=0; idx<len; idx++ )	/* now decrypt */
		 {
		 buf[idx] ^= PassWord()[pidx++] ^ xmask++;
		 if( pidx == length(PassWord) ) pidx = 0;
		 }

	   if( Decrypted )
		 {
		 if( fwrite( buf, 1, len, Decrypted ) !=len )
			 {
			 fprintf(ErrorFile,_("\nError:Couldn't write %lu bytes to the temporary file.\n"), (long unsigned)len);
			 break;
			 }
		 }
		 pos += len;
	 } /* while */

 if(buf!=NULL) free(buf);
}



#define L1_5cm          709   /*length of 1.5 cm*/
#define L1_27cm         600   /*length of 1.27 cm*/

void WP_Default(TconvertedPass1 *cq)
{
#ifdef DEBUG
  fprintf(cq->log,"\n#WP_Default() ");fflush(cq->log);
#endif
  long j;

  cq->RequiredFont=cq->Font = 1;
  cq->FirstSection = 120;
  cq->recursion = 0;
  cq->flag = NormalText;

  cq->latex_tabpos = 0;
  cq->tab_type   = 0;
  cq->CentrePage = false;

  cq->indenting = false;
  cq->indent_end = false;
  cq->ind_text1 = false;
  cq->char_on_line = false;              //!!!!!!
  cq->ind_leftmargin = 0;
  cq->ind_rightmargin = 0;
  cq->Columns = Columns;


// Default sizes of the document
  cq->WP_sidemargin = 1200;
  cq->Lmar=1890;
  cq->Rmar=1200;

// Default tab sets
  cq->tabpos[0] = L1_27cm;   /* 1e WP-tab is kantlijn --> */
  for (j = 2; j <= 10; j++)   /* Size of tabs is 1,27  cm  */
	cq->tabpos[j-1] = cq->tabpos[j-2] + L1_27cm;
  for (j = 10; j <= 39; j++)   /* ($02c5 wpu) verder        */
	cq->tabpos[j] = 0xffffL;

  cq->num_of_tabs = 10;
  cq->envir=' ';
}

#undef L1_5cm
#undef L1_27cm

void Heading(void);
void ShowHelp(void);


/* This procedure handles parametters and the filenames*/
int CommandLine(int argc, char *argv[])
{
  string StrName;
  char *name;
  long p;
  char firstch;

  wpd_filename = "";
  latex_filename = "";

  for (p = 1; p < argc; p++)
    {
    StrName=argv[p];
    if(StrName=="") continue;
    name=StrName();
    firstch = name[0];

    if (firstch == '/') name[0] = '-';
    if (name[0] == '-')
      {
      if (!strcmp(name, "-amssymb")||!strcmp(name, "-Amssymb"))
		{
		Amssymb = false;
		continue;
		}
      if (!strcmp(name, "-bbm")||!strcmp(name, "-Bbm"))
		{
		Bbm = false;
		continue;
		}
      if (!strcmp(name, "-configfile")||!strcmp(name, "-@"))
		{
		if(++p<argc)
		    {
		    LoadConfigFile(argv[p]);
		    }
		continue;
		}
      if ((!strcmp(name, "-copyright"))||(!strcmp(name, "--copyright")))
	   {
	   Heading();
	   puts(_("\nThis program and related files are copyright (c) 1996-2001 Fojtik Jaroslav."
		  "\nThis code can be redistributed and/or modified under the terms of the GNU"
		  "\nGeneral Public License available from: http://www.gnu.org/copyleft/gpl.html."
		  "\nWith one aditional requirement:"
		  "\n         If you decide to modify, you should somehow sign a modified"
		  "\nplace in the code and you must append your name at the end of this message"
		  "\nin this manner:"
		  "\n \"Modified by Mr Xxxxx Yyyy month.day.year\""
		  "\n        If you decide to modify this code, try to ask me first. Your change may"
		  "\nbe userfull for others and it will be good to include it into further releases."
		  "\n        If you didn't contribute to WP2LaTeX in any way, you must not say that"
		  "\nyou are author of it."
		  "\n"
		  "\nThanks to:Rafael Rodriguez Pappa  for several notes and bugfixes"
		  "\n          Damil Cavar             for several bugreports"
		  "\n          Hilmar Preusse          for translating several WP extended chars"
		  "\n                                  and translating messages to the German"
		  "\n          and of course to all authors of previous releases of WP2LaTeX"
		  "\n"
		  "\nIf you find this program userful please send me a *POSTCARD* with"
		  "\ncharacteristic image of your country."
		  "\nMy address is: Jaroslav Fojtik; Dvouramenna 13; Prague 4; Czech Republic"
		  "\n                14000"
		  "\n"
		  "\nAuthor is not liable for any damage of your system or loose of your profit"
		  "\ncaused by using or not using this program."
		  "\n"));
	   exit(EXIT_SUCCESS);
	   }

      if (!strncmp(name, "-cp",3))
	    {
	    name+=3;
	    if(*name=='-') name++;
	    if (!strcmp(name, "852"))
		  {
		  CodePage = 852;
		  continue;
		  }
	    if (!strcmp(name, "895"))
		  {
		  CodePage = 895;
		  continue;
		  }
	    if (!strcmp(name, "1250"))
		  {
		  CodePage = 1250;
		  continue;
		  }
	    if (!strcmp(name, "ISO8859-1") || !strcmp(name, "ISO-8859-1"))
		  {
		  CodePage = ISO8859_1;
		  continue;
		  }
	    if (!strcmp(name, "ISO8859-2") || !strcmp(name, "ISO-8859-2"))
		  {
		  CodePage = ISO8859_2;
		  continue;
		  }
	    if (!strcmp(name, "Styles") || !strcmp(name, "styles"))
		  {
		  CPStyles=true;
		  continue;
		  }
	    CodePage = 0;
	    Heading();
	    printf(_("Warning: Unknown output codepage \"%s\"!"),chk(name));
	    continue;
	    }
      if (!strcmp(name, "-endnotes")||!strcmp(name, "-Endnotes"))
		{
		EndNotes = 0;
		continue;
		}
      if (!strcmp(name, "-epsfig"))
		{
		InputPS = 4;
		continue;
		}
      if (!strcmp(name, "-extractWPG")||!strcmp(name, "-extract-WPG")||!strcmp(name, "-extract-wpg")||!strcmp(name, "-Extract-WPG"))
	    {
	    SaveWPG = true;
	    continue;
	    }
      if (!strcmp(name, "-fancyhdr")||!strcmp(name, "-Fancyhdr"))
		{
		FancyHdr = false;
		continue;
		}
      if (!strcmp(name, "-fix-spaces")||!strcmp(name, "-fixspaces"))
      		{
		FixSpaces = true;
		continue;
		}
      if (!strcmp(name, "-fontenc-T1")||!strcmp(name, "-FontEnc-T1"))
      		{
		fontenc = false;
		continue;
		}
      if (!strcmp(name, "-force-WP4")||!strcmp(name, "-forceWP4"))
		{
		WPVersion=0x400;
		continue;
		}
      if (!strcmp(name, "-graphicx"))
		{
		InputPS = 2;
		continue;
		}
      if (!strcmp(name, "-charset1"))
		{
		WPcharset = 1;
		continue;
		}
      if (!strcmp(name, "-charsetCZ") || !strcmp(name, "-charsetCS"))
		{
		WPcharset = 2;
		continue;
		}

      if (!strcmp(name, "-ignore-tabs") || !strcmp(name, "-ignoretabs"))
		{
		Tabs = false;
		continue;
		}
      if (!strcmp(name, "-image-path")||!strcmp(name, "-imagepath"))
		{
		p++;
		ImgInputDirs |= argv[p];
		continue;
		}
      if (!strcmp(name, "-input-PS") || !strcmp(name, "-inputPS"))
		{
		InputPS = 0;
		continue;
		}

      if (!strcmp(name, "-L"))
		{
		if(++p<argc)
		    {
		    StrName=EnvLanguage;
		    StrName+=argv[p];
		    putenv(StrName);
		    }
		continue;
		}

      if (!strcmp(name, "-LaTeX2") || !strcmp(name, "-latex2") || !strcmp(name, "-latex209"))
		{
		LaTeX_Version=0x209;
		Amssymb=-1;
		if(Columns==1) Columns=0;
		continue;
		}
      if (!strcmp(name, "-LaTeX3") || !strcmp(name, "-latex3"))
		{
		LaTeX_Version=0x300;
		continue;
		}
      if (!strcmp(name, "-LaTeXsym")||!strcmp(name, "-latexsym"))
		{
		LaTeXsym = false;
		continue;
		}
      if (!strcmp(name, "-longtable") || !strcmp(name, "-LongTable"))
		{
		LongTable = true;
		continue;
		}
      if (!strcmp(name, "-optimize-section")||!strcmp(name, "-optimizesection"))
		{
		OptimizeSection = true;
		continue;
		}
      if (!strcmp(name, "-makeidx")||!strcmp(name, "-MakeIdx"))
		{
		MakeIdx = false;
		continue;
		}
      if (!strcmp(name, "-NFSS")||!strcmp(name, "-nfss"))
		{
		NFSS=true;
		continue;
		}

      if (!strncmp(name, "-no",3))
	    {
	    name+=3;
	    if(*name=='-') name++;
	    if (!strcmp(name, "amssymb")||!strcmp(name, "Amssymb")) /*-no-amssymb*/
		      {
		      Amssymb = -1;
		      continue;
		      }
	    if (!strcmp(name, "bbm")||!strcmp(name, "Bbm")) /*-no-bbm*/
		      {
		      Bbm = -1;
		      continue;
		      }
	    if (!strcmp(name, "endnotes")||!strcmp(name, "Endnotes")) /*-no-endnotes*/
		      {
		      EndNotes = -1;
		      continue;
		      }
	    if (!strcmp(name, "erase-strip")||!strcmp(name, "erasestrip")) /*-no-erase-strip*/
		      {
		      EraseStrip = false;
		      continue;
		      }
	    if (!strcmp(name, "extract-WPG")||!strcmp(name, "extractWPG")||!strcmp(name, "extractwpg")) /*-no-extract-WPG*/
		  {
		  SaveWPG = false;
		  continue;
		  }
	    if (!strcmp(name, "fancyhdr")||!strcmp(name, "FancyHdr")||!strcmp(name, "Fancyhdr")) /*-no-fancyhdr*/
		      {
		      FancyHdr = -1;
		      continue;
		      }
	    if (!strcmp(name, "fix-spaces")||!strcmp(name, "fixspaces")) /*-no-fix-spaces*/
		      {
		      FixSpaces = false;
		      continue;
		      }
	    if (!strcmp(name, "fontenc-T1")||!strcmp(name, "FontEnc-T1"))
		      {
		      fontenc = -1;
		      continue;
		      }
	    if (!strcmp(name, "columns")||!strcmp(name, "Columns")) /*-no-columns*/
		      {
		      Columns = -1;
		      continue;
		      }
	    if (!strcmp(name, "inputPS")||!strcmp(name, "InputPS")) /*-no-inputPS*/
		      {
		      InputPS = -1;
		      continue;
		      }
	    if (!strcmp(name, "latexsym")||!strcmp(name, "LaTeXsym")) /*-no-latexsym*/
		      {
		      LaTeXsym = -1;
		      continue;
		      }
	    if (!strcmp(name, "longtable") || !strcmp(name, "LongTable")) /*-no-longtable*/
		      {
		      LongTable = -1;
		      continue;
		      }
	    if (!strcmp(name, "NFSS")||!strcmp(name, "nfss")) /*-no-NFSS*/
		      {
		      NFSS=false;
		      continue;
		      }
	    if (!strcmp(name, "makeidx")||!strcmp(name, "MakeIdx"))
		      {
		      MakeIdx = -1;
		      continue;
		      }
	    if (!strcmp(name, "optimize-section")||!strcmp(name, "optimizesection")||!strcmp(name, "Optimize-Section"))
		      {
		      OptimizeSection = false;
		      continue;
		      }
	    if (!strcmp(name, "rsfs") || !strcmp(name, "Rsfs")) /*-no-rsfs*/
		      {
		      Rsfs = -1;
		      continue;
		      }
	    if (!strcmp(name, "safemode")||!strcmp(name, "safe-mode")) /*-no-safemode*/
		      {
		      ExtendedCheck = false;
		      continue;
		      }
	    if (!strcmp(name, "texchars")||!strcmp(name, "TeXchars")) /*-no-texchars*/
		      {
		      TexChars = false;
		      continue;
		      }
	    if (!strcmp(name, "textcomp")) /*-no-textcomp*/
		      {
		      textcomp = -1;
		      continue;
		      }
	    if (!strcmp(name, "ulem")) /*-no-ulem*/
		      {
		      Ulem = -1;
		      continue;
		      }
	    if (!strcmp(name, "wasy")||!strcmp(name, "Wasy")) /*-no-wasy*/
		      {
		      Wasy = -1;
		      continue;
		      }
	    name=StrName();
	    }

      if (!strcmp(name, "-rsfs"))
		{
		Rsfs = true;
		continue;
		}
      if (!strcmp(name, "-texchars"))
		{
		TexChars = true;
		continue;
		}
      if (!strcmp(name, "-textcomp"))
		{
		textcomp = true;
		continue;
		}
      if (!strcmp(name, "-safemode"))
		{
		ExtendedCheck = true;
		continue;
                }
      if (!strcmp(name, "-s"))
      		{
		if(++p<argc)
		    {
	  	    wpd_password = argv[p];
		    }
		continue;
		}
      if (!strcmp(name, "-S") || !strcmp(name, "-Silent") || !strcmp(name, "-silent"))
      		{
		Verbosing = 0;
		continue;
		}
      if (!strcmp(name, "-ulem"))
		{
		Ulem = 0;
		continue;
		}
      if (!strcmp(name, "-use-all") || !strcmp(name, "-useall"))
		{
		Amssymb = false;
		Bbm = false;
		Columns = 1;	/*-1 disabled; 0 auto two; 1 auto multi; 2 forced two; >2 forced multi*/
		EndNotes = false;
		InputPS = false,
		LaTeXsym = false;
		PostScript = false;
		Rsfs = false;
		textcomp = false;
		Ulem = false;
		Wasy = false;
		continue;
		}
      if (!strcmp(name, "-v"))
		{
		printf("wp2latex %s   %s\n  modules:(",version,versiondate);
		FFormatTranslator *Fx;
		Fx=FFormatTranslator::first();
		while(Fx!=NULL)
		  {
		  printf("%s%c",Fx->shortkey(),Fx->next()==NULL?0:',');
		  Fx=Fx->next();
		  }
		printf(")\n");
#ifdef CORELEFT
		printf("  free memory %ld\n",CORELEFT());
#endif
		exit(EXIT_SUCCESS);
		}
      if (!strcmp(name, "-wasy")||!strcmp(name, "-Wasy"))
		{
		Wasy = 0;
		continue;
		}


      if (!strcmp(name, "--help") || !strcmp(name, "-?") || !strcmp(name, "-help"))
	   {
	   Heading();
	   ShowHelp();
	   exit(EXIT_SUCCESS);
	   }

      if (!strcmp(name, "-i"))
	  {
	  p++;
	  wpd_filename = argv[p];
	  continue;
	  }
      if (!strcmp(name, "-o"))
	  {
	  p++;
	  latex_filename = argv[p];
	  if(latex_filename == '-') Verbosing = 0; /*disable messages to stdout*/
	  continue;
	  }
      if (!strcmp(name, "-l"))
	  {
	  p++;
	  log_filename = argv[p];
	  continue;
	  }


      name[0]=firstch;
      if(firstch=='-') goto UnknownArg;
      }


    if (wpd_filename == "")
        {
        wpd_filename = name;
        continue;
        }
    if (latex_filename == "")
        {
        latex_filename = name;
        continue;
	}

UnknownArg:
    Heading();
    printf(_("Warning: Unknown argument \"%s\"!"),chk(name));
  }

  if(latex_filename == '-') Verbosing=0;

  if(Amssymb>=false && LaTeX_Version<0x300)
	fprintf(err,_("\nWarning: Package \"amssymb\" should not be used in LaTeX 2.09!"));

return(0);
}


/*This class is a container for convertor modules*/
FFormatTranslator *FFormatTranslator::First=NULL;
FFormatTranslator::FFormatTranslator(const char *NewShortKey,Convert_Pass1_XXX NewConvertor)
{
 ShortKey=NewShortKey;
 Convertor=NewConvertor;

 Next=First;
 Previous=NULL;
 First=this;
 if(Next!=NULL) Next->Previous=this;
}

FFormatTranslator::~FFormatTranslator(void)
{
 if(First==this)
	{
	First=Next;
	if(Next!=NULL) Next->Previous=NULL;
	return;
	}
 if(Next!=NULL) Next->Previous=Previous;
 if(Previous==NULL) return; //Error
 Previous->Next=Next;
}

Convert_Pass1_XXX GetConvertor(const char *ShortKey)
{
FFormatTranslator *Fx;

if(ShortKey==NULL) return(NULL);
Fx=FFormatTranslator::First;
while(Fx!=NULL)
  {
  if(!strcmp(Fx->ShortKey,ShortKey)) return(Fx->Convertor);
  Fx=Fx->Next;
  }
return(NULL);
}


/* End. */
