/******************************************************************************
 * program:     wp2latex                                                      *
 * function:    convert WordPerfect 3.x,4.x,5.x and 6.x files into LaTeX      *
 * modul:       wp2latex.cc                                                   *
 * description: The main program that communicates with OS.                   *
 ******************************************************************************/

///To do list:///
// PODIVAT SE NA sTRsTR
// u minipage se musi nacist vertical position a v pripade full nastavit \vsize
// podivat se na pismo wsuipa
// sjednotit LineSpacing pro vsechny verze WP
// vnorene dokumenty
// udelat titulni stranku i pro WP6.x
// pokusit se zkonvertovat invisible characters
// nechodi center tab a right tab
// konverze prikazu alignr, alignl, alignc v matematickem modu
// pridat parametr -force-HTML

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//Non standard include files
#include <common.h>
#include "wp2latex.h"
#include <strings.h>
#include <lists.h>

#ifdef __UNIX__
 #include <unistd.h>
 #include <signal.h>
#endif

//#include <alloc.h>

void Sys_Variable(void);
void UpdateFigDir(void);
WORD CheckSum(char *PassWord);

extern FILE *wpd, *table;
extern FILE *strip, *logg, *err;
extern FILE *latex;

extern void (*Convert_first_pass)(FILE *FileIn,FILE  *table, FILE *StripOut,FILE *LogFile,FILE *ErrorFile);

extern string wpd_filename, strip_filename, table_filename, latex_filename,
       log_filename;
extern string wpd_password;

extern int WPVersion;

extern list ImgInputDirs;


//Definition of methods of the Percentor class
void Percentor::Actualise(long NewValue)
{
int NewPercent;

 if(EndValue==StartValue) NewPercent=100;
	else NewPercent = (int)floor(
	     (float)(NewValue - StartValue) / (EndValue - StartValue) * 100 + 0.5);

 if(NewPercent!=CurrentPercent)
		{
		CurrentPercent=NewPercent;
		Show();
		}
}


void Percentor::Hide(void)
{
 if(displayed) printf("\b\b\b\b");
 displayed=false;
}


void Percentor::Init(long Start, long End, const char *text)
{
 StartValue=Start;
 EndValue=End;
 if (Verbosing >= 1)
	{
	if(text) printf(text);
	printf(_("\nConverting-percentage: "));
	}
}


Percentor::Percentor(void)
{
 displayed=false;
 CurrentPercent=-1;
}


Percentor::~Percentor(void)
{
 return;
}

void Percentor::SetPercent(int NewPercent)
{
 if(NewPercent==CurrentPercent && displayed) return;
 CurrentPercent=NewPercent;
 Show();
}


void Percentor::Show(void)
{
 printf(displayed?"\b\b\b\b%3d%%":"%3d%%",CurrentPercent);
 fflush(stdout);
 displayed=true;
}


/*This function reads string from standard input*/
void StrGets(string & s)
{
char c;

  c=0;
  s="";
  while(c!='\n')
      {
      c=fgetc(stdin);
      switch(c)
	 {
	 case '\n':break;
	 default:s+=c;
	 }
      }
}


/*This function handles error situation*/
void RunError(int ExitCode, const char *s)
{
  putchar('\n');
  switch (ExitCode)
    {
    case 0:    fprintf(err,_("Conversion completed.\n\n"));
	       break;
    case 0x2:  fprintf(err,_("Error: File %s not found.\n"),chk(s));
	       break;
    case 0x3:  fprintf(err,_("Error: Path not found.\n"));
	       break;
    case 0x6:  fprintf(err,_("Error: Temporary file %s cannot be created!\n"),s);
	       break;
    case 0x10: fprintf(err,_("Info: Program was aborted.\n"));
	       break;
    case 0x15: fprintf(err,_("Info: Please report this situation to enable further developping WP2LaTeX.\n"));
	       break;
    case 0x20: fprintf(err,_("Error: Output device is full.\n"));
    	       break;           
    case 0x8B: fprintf(err,_("Fatal: Program Error. Write full bug report.\n"));
               break;           
    case 0x200:fprintf(err,_("Error: No filename entered!\n"));
               break;
    case 0x201:fprintf(err,_("Error: File \"%s\" is not a WordPerfect document!\n"),wpd_filename());
	       break;
    case 0x202:fprintf(err,_("Error: Unsupported format (%s)!\n"),s);
	       break;
    case 0x203:fprintf(err,_("Error: Sorry, I have no idea how to decrypt this file!\n"));
	       break;
    }


  if (wpd != NULL)   fclose(wpd);     	  wpd = NULL;
  if (strip != NULL)
  	{
	fclose(strip);			  strip = NULL;
#ifndef DEBUG
	if(EraseStrip) unlink(strip_filename);
#endif
	}
  if (table != NULL)
	{
	fclose(table);			 table = NULL;
#ifndef DEBUG
	if(EraseStrip) unlink(table_filename);
#endif
	}
  if (latex_filename != '-')
      if (latex != NULL) fclose(latex);   latex = NULL;
  if (logg != NULL)
        {
        if(!ExitCode)
	  fprintf(logg,_("\n!!!! Fatal Exit No %d!!!!\n"),ExitCode);
	fclose(logg);		          logg = NULL;
        }

  exit(ExitCode);
}


int ObtainPassword(string & password, WORD encryption)
{
char c;

TryAgain:

  printf(_("\nEnter Password: "));
  StrGets(password);

if(CheckSum(password())!=encryption)
     {
     printf( _("\nWarning! Password checksum does not match file checksum! Again Y/n/i?"));

NextChar:c=getchar();
//	 putchar('\n');
	 switch(c)
	       {
	       case 'A':
	       case 'a':
	       case 'Y':
	       case 'y':getchar();
	       case '\n':goto TryAgain;
	       case 'n':
	       case 'N':getchar();return(false);
	       case 'I':
	       case 'i':getchar();
			ExtendedCheck=true;
			return(true);		// ignore=run anyway
	       default:goto NextChar;
	       }
	   }

return(true);
}


/*This function prints text console WP2LaTeX heading*/
void Heading(void)
{
static boolean printed=false;
 if(printed) return;
 printed=true;

 if(Verbosing >= 1)
     printf(_("\n>>>WP2LaTeX<<< Conversion program: From Wordperfect to LaTeX Version %s \n\n"
	      "  (c) TUE-Eindhoven ---- Made by R.C.Houtepen ---- Date : 24 Jan 1990\n"
	      "  (c) CVUT Prague   ---- Made by J.Fojtik     ---- Date : 1996-2001\n"),
	    version);
}


/*This function prints wp2latex help to the text console*/
void ShowHelp()
{
 printf(_("\nSomebody at Eindhoven Univ of Techn wrote a WordPerfect 5.0 to LaTeX"
	  "\nconversion program (only running on MSDOS of course) :=)"
	  "\n           This program was completely rewritten by Jaroslav Fojtik"
	  "\nand many new functions was added (eg conversion of math equations)"
	  "\nWrite any suggestions, reports or bugs to address FOJTIK@htc.honeywell.cz"
	  "\n     the address FOJTIK@CMP.FELK.CVUT.CZ may not be valid after 2000."
	  "\n"
	  "\nImportant note: Conversions performed by wp2latex are only partial."
	  "\nPrint the resulting document and compare carefully with original."
	  "\nIn all but the simplest documents you will have to do further editing."
	  "\nOr you may wisch to edit the WordPerfect document to remove unwanted"
	  "\ncodes (such as Tab, Tab Set), then run it through wp2latex again."
	  "\n"));
 printf(_(
	  "\nParametters: WP2LATEX [switches] [input_file [output_file]] [switches]"
	  "\n   -?  --help   Display this help page."
	  "\n   -amssymb     %sUse aditional package AMSsymb with special characters."
	  "\n   -bbm         %sUse aditional package Bbm with special characters."
	  "\n   -copyright   Display informations about Copyright of WP2LaTeX."
	  "\n   -cp-codepage Use \"codepage\" for output text. Supported code pages are"
	  "\n         \"852\"(ISO8), \"895\"(Kamenicky), \"1250\"(Win), \"ISO8859-2\", \"ISO8859_1\"."
	  "\n   -cp-styles   Turn on special styles for codepages."
	  "\n   -epsfig      Use style epsfig.sty for Postscript images."
	  "\n   -extract-WPG %sImages will be extracted into separate files (only WP 5.x)"
	  "\n   -fancyhdr    Use package fancyhdr.sty for headers and footers"
	  "\n   -fix-spaces  Try to fix multiple spaces by ~."
	  "\n   -fontenc-T1  Use package fontenc.sty [T1] with additional characters"
	  "\n   -forceWP4    Allow to convert documents from WP 4.x"
	  "\n   -graphicx    Use style graphicx.sty for Postscript images."
	  "\n   -configfile -@ filename   Load an external configuration file."
	  "\n   -charset1    %sSwitch internal WordPerfect char set to set 1."
	  "\n   -charsetCZ   %sSwitch internal WordPerfect char set to keybcs."
	  "\n   -i; -o; -l   Explicitely define an imput; output; log file"
	  "\n   -ignore-tabs %sRemove all [TAB] symbols from WP document."
	  "\n   -image-path PATH/  Add one item to searching path for images."
	  "\n   -input-PS    Use Postscript features in the LaTeX (mainly images)."
	  "\n   -L LANGUAGE  Switch message translations to language LANGUAGE."
	  "\n   -LaTeX2      %sOptimize output for LaTeX 2.09."
	  "\n   -LaTeX3      %sOptimize output for LaTeX 2e or LaTeX 3.0"
	  "\n   -latexsym    %sUse aditional package latexsym with special characters."
	  "\n   -longtable   Use aditional package longtable for better tables."
	  "\n   -makeidx     %sUse aditional package makeidx for placing index."
	  "\n   -no-amssymb  %sDisable aditional package AMSsymb with special characters."
	  "\n   -no-bbm      %sDisable aditional package Bbm with special characters."
	  "\n   -no-columns  Do not translate multicolumn text to more columns."
	  "\n   -no-extractWPG Turn an all extracing images off."
	  "\n   -no-fancyhdr Do not use package fancyhdr.sty."
	  "\n   -no-input-PS Disable using anything from Postscript in the LaTeX."
	  "\n   -no-latexsym Disable aditional package latexsym."
	  "\n   -no-safemode Disable checking features - faster conversion."
	  "\n   -no-texchars %sStandard conversion of all characters to proper TeX sequences."
	  "\n   -no-wasy     %sDisable aditional package Wasy2 with special characters."
	  "\n   -rsfs        %sUse aditional package mathrsfs with special characters."
	  "\n   -safemode    %sCheck safely consistency of each object and try to fix all"
	  "\n                incorrect things. Good for corrupted WP files."
	  "\n   -s \"password\" Password for decoding WP document (WP4.x & WP5.x only)."
	  "\n   -S -silent   %sSwitch to the silent mode."
	  "\n   -texchars    %sDo not interpret tex macros (all chars 32-128 will not be"
	  "\n                converted to TeX sequences)."
	  "\n   -ulem        %sUse aditional package ulem for underlining and striking out."
	  "\n   -v           Get info about version."
	  "\n   -wasy        %sUse aditional package Wasy2 with special characters."
	  "\n"),
	  Amssymb>=false?"*":"",
	  Bbm>=false?"*":"",
	  SaveWPG>=true?"":"*",		//bold the default options
	  WPcharset==1?"*":"",
	  WPcharset==2?"*":"",
	  Tabs<=false?"*":"",
	  (LaTeX_Version & 0xFF00) == 0x200?"*":"",
	  (LaTeX_Version & 0xFF00) == 0x300?"*":"",
	  LaTeXsym>=false?"*":"",
	  MakeIdx>=false?"*":"",
	  Amssymb>=false?"":"*",
	  Bbm>=false?"":"*",
	  TexChars>=true?"":"*",
	  Wasy>=false?"":"*",
	  Rsfs>=false?"*":"",
	  ExtendedCheck==true?"*":"",
	  Verbosing>0?"":"*",
	  TexChars<=false?"":"*",
	  Ulem>=false?"*":"",
	  Wasy>=false?"*":""
	  );
#ifndef Registered
   puts(_("\nPlease send me a postcard. Run \"wp2latex -copyright\" for more details."));
#endif
}


static int Filenames(void)
{
string StrName;
long p;

  Heading();

  if (latex_filename == "")
      {
      if (Verbosing >= 1 || wpd_filename == "")
         {
	 printf(_("\nWordPerfect-filename [%s] : "), chk(wpd_filename()));
	 StrGets(StrName);
	 if (StrName != "") wpd_filename=StrName;
         }

    if (wpd_filename == "")
        { 			  /* No filename entered */
	RunError(0x200);
        }

    p=length(wpd_filename);
    while(p>0)
	 {
	 if(wpd_filename[p]=='.') break;
	 if((wpd_filename[p]=='/')||(wpd_filename[p]=='\\'))
		{
		p=length(wpd_filename);
		break;
		}
	 p--;
	 }
    if(p<=0) p=length(wpd_filename); //no extension


    if (latex_filename == "")
          {
	  latex_filename=copy(wpd_filename,0,p) + ".tex";
	  }

    if (Verbosing >= 1 || latex_filename == "")
       {
       printf(_("LaTeX-filename [%s] : "), chk(latex_filename()));
       StrGets(StrName);
       if (StrName != "") latex_filename = StrName;
       }
  }

  p=length(wpd_filename);
    while(p>0)
    	 {
         if(wpd_filename[p]=='.') break;
         if((wpd_filename[p]=='/')||(wpd_filename[p]=='\\'))
	 	{
                p=length(wpd_filename);
                break;
		}
         p--;
	 }

  strip_filename = copy(wpd_filename,0,p) + ".STR";
  table_filename = copy(wpd_filename,0,p) + ".TBL";

  return(0);
}


static void CheckOptions(void)
{
if(LaTeX_Version<0x300)		//disable colors for LaTeX 2.09
	{
	if(colors>true)
		fprintf(err,_("Warning: Color style is available only for LaTeX 2e !\n"));
	colors=-1;
	}
}


#if defined(__UNIX__) & !defined(DEBUG)
/* This is a main signal handler */
void sig_int(int signo)
{
static int sigcount=0;

 if(sigcount++>2) abort();
 switch(signo)
    {
    case SIGINT:RunError( 0x10, "");   break;
    case SIGSEGV:RunError(0x8B, "");   break;
    }
}

static void HookSignals(void)
{
 signal(SIGINT,sig_int);
 signal(SIGSEGV,sig_int);
}
#endif


/*This procedure checks sizes of standard variables*/
static void CheckCompiler(void)
{
if((sizeof(__u8)!=1)||(sizeof(__u16)!=2)||(sizeof(__u32)!=4)||
   (sizeof(__s8)!=1)||(sizeof(__s16)!=2)||(sizeof(__s32)!=4))
	{
	Heading();
	fprintf(err,_("\nFatal: Code was badly compiled. Check types.h !!!!!"));
	}
}


/*---begin of PROGRAM---*/
int main(int argc , char *argv[])
{
  latex = NULL;		/* The output Latex file */
  strip = NULL;		/* The temporary file rough conversion */
  table = NULL;		/* The temporary file with line informations */
  wpd   = NULL;		/* The input WP file */
  logg  = NULL;		/* The output log file */
  err   = stderr;	/* The output for error messages */

/* Default initialization */
  Verbosing = 1;
  TexChars = false;
  CodePage = 0;		/*unknown - no 8 bit code generated*/
  WPcharset = 1;	/*common char set 1*/
  ExtendedCheck = true;	/*true*/
  FixSpaces = false;
  OptimizeSection = true;
  UnknownCharacters=UnknownObjects=CorruptedObjects=0;
  Convert_first_pass = NULL;

#if defined(__UNIX__) & !defined(DEBUG)
  HookSignals();
#endif

#ifdef DEBUG
  log_filename="default.log";
#endif
//  clrscr();

#ifdef __gettext__
  InitGettext();
#endif /* __gettext__ */

//printf("%ld",coreleft());


  CheckCompiler();
  Sys_Variable();
  CommandLine(argc,argv);
  Filenames();
  UpdateFigDir();
  CheckOptions();

  if ( (wpd = fopen(wpd_filename, "rb"))   == NULL )
	RunError(2,wpd_filename);
  Wpd_check(wpd,err);

  if ( (strip = fopen(strip_filename, "w"))   == NULL )
	RunError(2,strip_filename);
  if ( (table = fopen(table_filename, "w+b")) == NULL )
	RunError(2,table_filename);
  logg=NULL;
  if (log_filename != "")
      {
      if((logg = fopen(log_filename, "w")) == NULL )
      	{
	fprintf(err,_("\nWarning: Cannot open log file %s ..."),log_filename());
        }
	else fprintf(logg,_("%% log file for converted file %s, WP ver %d.%d\n"
			    "%% conversion program WP2LaTeX v. %s from %s.\n"),
			   wpd_filename(),WPVersion >> 8,WPVersion & 0xFF,version,versiondate);
      }


  if(Convert_first_pass!=NULL)
		{
		if(Verbosing >= 1)
			printf(_("\nConverting ...\n"));
		Convert_first_pass(wpd, table, strip, logg, err);
		}
	   else RunError(0x202,DocumentType);  /* unsupported WP version */


  if (wpd != NULL)
          {
          fclose(wpd);
          wpd = NULL;
          }
  if (strip != NULL)
          {
          if(fflush(strip)) RunError(0x20,""); //The RunError procedure will erase strip in this case
	  if(fclose(strip)) {strip = NULL;RunError(0x20,"");}
          strip = NULL;
          }
  if (table != NULL)
         {
         if(fflush(table)) RunError(0x20,""); //The RunError procedure will erase table in this case
         if(fclose(table)) {table = NULL;RunError(0x20,"");}
         table = NULL;
         }


  if(logg!=NULL) fprintf(logg,_("\n%% pass1 finished going to pass2\n\n"));

  if ( (strip = fopen(strip_filename, "r"))   == NULL )
	RunError(2,strip_filename);
  if (latex_filename == '-') latex=stdout;
      else if ( (latex = fopen(latex_filename, "w"))   == NULL )
		 RunError(2,latex_filename);
  if ( (table = fopen(table_filename, "r+b")) == NULL )
	RunError(2,table_filename);


  Convert_second_pass(strip, table, latex, logg, err);


  if (strip != NULL) fclose(strip);   strip = NULL;
  if (table != NULL) fclose(table);   table = NULL;
  if (latex_filename != '-')
       if (latex != NULL)
                 {
                 if(fclose(latex)) {latex = NULL;RunError(0x20,"");}
                 }
  latex = NULL;
  if (logg != NULL)
  	{
	fprintf(logg,_("\n%% pass2 finished\n"
	              "%%  There were %ld unconverted characters.\n"
		      "%%  There were %ld unrecognized/ignored objects.\n"),UnknownCharacters,UnknownObjects);
	if(CorruptedObjects>0)
	    fprintf(logg,_("%%  There were %ld corrupted objects.\n"),CorruptedObjects);
	if(fclose(logg))
	       fprintf(err,_("\nError: Log file is not completely written."));
        }
  logg = NULL;


/* Delete auxillary files */
  if(EraseStrip)
     {
#ifndef DEBUG
     unlink(strip_filename);
     unlink(table_filename);
#endif
     }

//printf("%ld",coreleft());
  return(0);
}
