# makefile for wp2latex for EMX (gcc)
# runs under DOS, linux, solaris and perhaps OS/2
#
TARGET_NAME = wp2latex
TARGET_DIR = ./execute/
GCC=gcc
FLAGS=-O2 -I./atoms/include/ -DFINAL -fno-rtti -fno-exceptions  #-Wall
OBJ=.o


# Uncomment next line for compile a debugging version of WP2LaTeX
#DEBUG=-DDEBUG -g

# Uncomment next line for allowing a gettext multilanguage translation features
# This feature is not required for succesfull run of WP2LaTeX.
GETTEXT=-D__gettext__

# to get an error list under OS/2
# ERROR=2>&1 | tee >> errlist



#### OS autodetection part of makefile ####

ifeq ($(GETTEXT),-D__gettext__)
ifeq ($(OSTYPE),)		#-lintl is no longer needed under linux
LIBS=-lintl
endif 
IGETTEXTOBJ=igettext$(OBJ)
endif

# for Linux - debian
ifeq ($(OSTYPE),linux-gnu)
TARGET_DIR = ../linux/
GCC = g++
endif

# for Linux - slakware
ifeq ($(OSTYPE),linux)
TARGET_DIR = ../linux/
LIBS+= -lstdc++
endif

# for Linux - Red Hat
ifeq ($(OSTYPE),Linux)
TARGET_DIR = ../linux/
LIBS+= -lstdc++
endif
#end of Linux

# for sun
ifeq ($(OSTYPE),solaris)
TARGET_DIR = ../sun/
LIBS=-lm
endif

ifeq ($(OSTYPE),sunos4)
TARGET_DIR = ../sun/
LIBS=-lm -lstdc++
endif
#end of sun

#for OS/2
ifneq "$(OS2_SHELL)" ""
TARGET_DIR = ../os2/
TARGET_NAME = wp2latex.exe
OSTYPE = os2
LIBS=-s -lstdcpp
endif

#Nothing is an indentifier for DOS? Too bad.
ifeq ($(OSTYPE),)
ifeq ($(HOSTTYPE),iris4d)
OSTYPE = iris4d
LIBS=-lm
TARGET_DIR = ../iris/
else
TARGET_DIR = ../dos/
TARGET_NAME = wp2latex.exe
OSTYPE = dos
endif
endif

# for aix
ifeq ($(OSTYPE),aix)
TARGET_DIR = ../aix/
LIBS=-lm
endif
#end of aix


#library components
OBJECTS=atoms/sets$(OBJ) atoms/struct$(OBJ) atoms/lists$(OBJ) atoms/dbllist$(OBJ) images/raster$(OBJ) atoms/strings$(OBJ) cptran$(OBJ)
#mandatory components
OBJECTS+=wp2latex$(OBJ) wp2l_lib$(OBJ) pass1$(OBJ) pass2$(OBJ) wp2lfuti$(OBJ) formulas$(OBJ) images$(OBJ) charactr$(OBJ) $(IGETTEXTOBJ)

OBJECTS+=pass1_3$(OBJ)
OBJECTS+=pass1_4$(OBJ)
OBJECTS+=pass1_5$(OBJ)
OBJECTS+=pass1_6$(OBJ)

## codepage library for HTML & RTF ##
OBJECTS+=cp-trn$(OBJ)
# HTML parser ##
OBJECTS+=pass1_ht$(OBJ)
# RTF parser ##
OBJECTS+=pass1rtf$(OBJ)
# T602 parser ##
OBJECTS+=pass1602$(OBJ)


######### End of definitions starting compile rules ###########

.SUFFIXES: .cc $(OBJ) .exe

.cc$(OBJ):
	$(GCC) $(FLAGS) $(DEBUG) -c $*.cc $(ERROR) -o $*$(OBJ) $(GETTEXT)

$(OBJ).exe:
	$(GCC) $*$(OBJ) $(LIBS) $(ERROR)

$(TARGET_NAME): $(OBJECTS)
	$(GCC) $(GETTEXT) -o $(TARGET_NAME) $(OBJECTS) $(LIBS) $(ERROR)
	mkdir -p $(TARGET_DIR)
	cp ./$(TARGET_NAME) $(TARGET_DIR)
	strip $(TARGET_DIR)$(TARGET_NAME)


wp2latex$(OBJ): wp2latex.cc wp2latex.h
wp2l_lib$(OBJ): wp2l_lib.cc wp2latex.h
pass1$(OBJ):    pass1.cc    wp2latex.h
pass1_3$(OBJ):  pass1_3.cc  wp2latex.h
pass1_4$(OBJ):  pass1_4.cc  wp2latex.h
pass1_5$(OBJ):  pass1.cc    wp2latex.h
pass1_6$(OBJ):  pass1_6.cc  wp2latex.h
pass1_ht$(OBJ): pass1_ht.cc wp2latex.h cp_lib/html.trn 
pass1rtf$(OBJ): pass1rtf.cc wp2latex.h
pass1602$(OBJ): pass1602.cc wp2latex.h
pass2$(OBJ):    pass2.cc    wp2latex.h
wp2lfuti$(OBJ): wp2lfuti.cc wp2latex.h
formulas$(OBJ): formulas.cc wp2latex.h
images$(OBJ):   images.cc   wp2latex.h images/raster.h images/ras_img.cc
charactr$(OBJ): charactr.cc wp2latex.h
igettext$(OBJ): igettext.cc wp2latex.h

cptran$(OBJ):   cptran.cc cptran.h cp_lib/trn.trn 
cp-trn$(OBJ):   cp-trn.cc wp2latex.h cp_lib/cpg.trn

images/raster$(OBJ): images/raster.cc images/raster.h

atoms/lists$(OBJ):   atoms/lists.cc
atoms/dbllist$(OBJ): atoms/dbllist.cc
atoms/strings$(OBJ): atoms/strings.cc
atoms/struct$(OBJ):  atoms/struct.cc
atoms/sets$(OBJ):    atoms/sets.cc


cp_lib/trn.trn:
	make -C cp_lib trn.trn

cp_lib/html.trn:
	make -C cp_lib html.trn

cp_lib/cpg.trn:
	make -C cp_lib cpg.trn


all:     $(TARGET_NAME)
#	$(TARGET_NAME) -i test/charactr.wp

default: $(TARGET_NAME)


clean:
	rm -f wp2latex *.exe *$(OBJ) *.obj *.bak core *.~??
	rm -f atoms/*$(OBJ) atoms/*.bak
	rm -f test/*.tex
	rm -f images/*$(OBJ) images/*.bak
	make -C cp_lib clean


#This work correctly only in Linux
install: $(TARGET_NAME)
	@echo "Installing binary file"        
	cp $(TARGET_DIR)$(TARGET_NAME) /usr/local/bin
	@echo "Installing manual page"        
	cp ../doc/wp2latex.1 /usr/local/man/man1/
	@echo "Attempting to install wp2latex.sty into /usr/share/texmf/tex/latex/"
	cp ../styles.tex/wp2latex.sty  /usr/share/texmf/tex/latex/ 
	texhash
	@echo "I am a dumb script, you should complete installation by copying all stuff"
	@echo "from 'styles.tex/' direcrory to a proper place and rehash tex files."

uninstall: 
	rm -f /usr/local/bin/$(TARGET_NAME)	 
	rm -f /usr/local/man/man1/wp2latex.1
	rm -f /usr/local/man/cat1/wp2latex.1.gz
#	rm -f /usr/share/texmf/tex/latex/wp2latex.sty

